/*
 * linux/arch/arm/mach-omap/dsp/dsp_core.c
 *
 * OMAP DSP driver
 *
 * Copyright (C) 2002-2004 Nokia Corporation
 *
 * Written by Toshihiro Kobayashi <toshihiro.kobayashi@nokia.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: dsp_core.c
 * $Revision: 3.0.1
 * $Date: 2004/06/29
 *
 */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/proc_fs.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/signal.h>
#include <asm/delay.h>
#include <asm/irq.h>
#include <asm/arch/dsp.h>
#include "hardware_dsp.h"
#include "dsp.h"
#include "ipbuf.h"
#include "proclist.h"


MODULE_AUTHOR("Toshihiro Kobayashi <toshihiro.kobayashi@nokia.com>");
MODULE_DESCRIPTION("OMAP DSP driver module");
MODULE_LICENSE("GPL");

static struct semaphore mbsend_sem;

static unsigned short mbseq_send = 0;
static unsigned short mbseq_expect = 0;
static int mbseq_chklevel = MBSEQ_CHECKLEVEL;
struct sync_seq *sync_seq;

/*
 * DSP module user process list
 */
static struct list_head cur_users_list;

void omap_dsp_cur_users_add(struct task_struct *tsk)
{
	proc_list_add(&cur_users_list, tsk);
}

void omap_dsp_cur_users_del(struct task_struct *tsk)
{
	proc_list_del(&cur_users_list, tsk);
}

#ifdef CONFIG_PROC_FS
static int omap_dsp_proc_cur_users_read(char *page, char **start,
					off_t off, int count,
					int *eof, void *data)
{
	char *out;
	int len;
	struct list_head *ptr;

	out = page;
	list_for_each(ptr, &cur_users_list) {
		struct proc_list *pl = (struct proc_list *)ptr;
		out += sprintf(out, "%d\n", pl->tsk->pid);
	}

	len = out - page;
	len -= off;
	if (len < count) {
		*eof = 1;
		if (len <= 0)
			return 0;
	} else {
		len = count;
	}
	*start = page + off;

	return len;
}

static void omap_dsp_create_cur_users_proc(void)
{
	struct proc_dir_entry *ent;

	ent = create_proc_read_entry("cur_users", 0, procdir_dsp,
				     omap_dsp_proc_cur_users_read, NULL);
	if (ent == NULL) {
		printk(KERN_ERR
		       "omapdsp: failed to register proc device: "
		       "cur_users\n");
	}
}

static void omap_dsp_remove_cur_users_proc(void)
{
	remove_proc_entry("cur_users", procdir_dsp);
}
#endif /* CONFIG_PROC_FS */

static void __init omap_dsp_cur_users_init(void)
{
	INIT_LIST_HEAD(&cur_users_list);
#ifdef CONFIG_PROC_FS
	omap_dsp_create_cur_users_proc();
#endif
}

static void omap_dsp_cur_users_exit(void)
{
	proc_list_flush(&cur_users_list);
#ifdef CONFIG_PROC_FS
	omap_dsp_remove_cur_users_proc();
#endif
}

void omap_dsp_cur_users_map_update(void)
{
	struct list_head *ptr;

	preempt_disable();
	list_for_each(ptr, &cur_users_list) {
		struct proc_list *pl = (struct proc_list *)ptr;
		struct task_struct *tsk = pl->tsk;

		/*
		 * excluding ZOMBIE, STOPPED processes
		 */
		if ((tsk->state == TASK_RUNNING) ||
		    (tsk->state == TASK_INTERRUPTIBLE) ||
		    (tsk->state == TASK_UNINTERRUPTIBLE)) {
			omap_dsp_map_update(tsk);
		}
	}
	preempt_enable();
}

int sync_with_dsp(unsigned short *syncwd, unsigned short tid, int try_cnt)
{
	int cnt;

	if (*(volatile unsigned short *)syncwd == tid)
		return 0;

	for (cnt = 0; cnt < try_cnt; cnt++) {
		udelay(1);
		if (*(volatile unsigned short *)syncwd == tid)	/* success! */
			return 0;
	}

	/* fail! */
	return -1;
}

static __inline__ int mbsync(int try_cnt)
{
	int cnt;

	if (omap_readw(MAILBOX_ARM2DSP1_Flag) == 0)
		return 0;
	/*
	 * mailbox is busy. wait for some usecs...
	 */
	for (cnt = 0; cnt < try_cnt; cnt++) {
		udelay(1);
		if (omap_readw(MAILBOX_ARM2DSP1_Flag) == 0)	/* success! */
			return 0;
	}

	/* fail! */
	return -1;
}

#ifdef CONFIG_OMAP_DSP_MBCMD_VERBOSE
#define print_mb_busy_abort(cmd_h) \
	printk(KERN_DEBUG \
	       "mailbox: mailbox is busy. %s is aborting.\n", cmd_name(cmd_h))
#define print_mb_lock_abort(cmd_h) \
	printk(KERN_DEBUG \
	       "mailbox: mailbox is locked. %s is aborting.\n", cmd_name(cmd_h))
#else /* CONFIG_OMAP_DSP_MBCMD_VERBOSE */
#define print_mb_busy_abort(cmd_h)	do {} while(0)
#define print_mb_lock_abort(cmd_h)	do {} while(0)
#endif /* !CONFIG_OMAP_DSP_MBCMD_VERBOSE */

static int __mbsend(unsigned char cmd_h, unsigned char cmd_l,
		    unsigned short data)
{
	unsigned short cmd;

	if (mbsync(10) < 0) {
		print_mb_busy_abort(cmd_h);
		return -1;
	}

	cmd = (mbseq_send << 15) | (cmd_h << 8) | cmd_l;
	sync_seq->ad_arm = ++mbseq_send;
	omap_dsp_mblog_add(cmd, data, MBLOG_DIR_AD);
	omap_dsp_mblog_printcmd(cmd, data, MBLOG_DIR_AD);

	omap_writew(data, MAILBOX_ARM2DSP1);
	omap_writew(cmd, MAILBOX_ARM2DSP1b);
	return 0;
}

/*
 * omap_dsp_mbsend(): mailbox dispatcher
 * interrupt routine should call this function with
 * sendtype = MBSENDTYPE_NOWAIT
 */
int __omap_dsp_mbsend(unsigned char cmd_h, unsigned char cmd_l,
		      unsigned short data, unsigned char tid,
		      int argc, unsigned short *argv, int sendtype)
{
	int ret = 0;

	/*
	 * while MMU fault is set,
	 * only recovery command can be executed
	 */
	if (omap_dsp_err_mmu_isset() && (sendtype != MBSENDTYPE_RECOVERY)) {
		print_mb_busy_abort(cmd_h);
		return -1;
	}

	if (sendtype == MBSENDTYPE_NOWAIT) {
		if (down_trylock(&mbsend_sem)) {
			print_mb_lock_abort(cmd_h);
			return -1;
		}
	} else {
		if (down_interruptible(&mbsend_sem) < 0)
			return -1;
	}

	if (argc) {	/* we have extra argument */
		int dspmem_access;
		int i;

		if ((dspmem_access = is_dsp_internal_mem(ipbuf_sys_ad)) != 0)
			if (__mbsend(MBCMD(PM), OMAP_DSP_MBCMD_PM_ENABLE,
				     DSPREG_ICR_DMA_IDLE_DOMAIN) < 0)
				goto out;
		if (sync_with_dsp(&ipbuf_sys_ad->s, OMAP_DSP_TID_FREE, 10) < 0) {
			printk(KERN_ERR "omapdsp: ipbuf_sys_ad is busy.\n");
			ret = -EBUSY;
			goto out;
		}
		for (i = 0; i < argc; i++) {
			ipbuf_sys_ad->d[i] = argv[i];
		}
		ipbuf_sys_ad->s = tid;
		if (dspmem_access)
			if (__mbsend(MBCMD(PM), OMAP_DSP_MBCMD_PM_DISABLE,
				     DSPREG_ICR_DMA_IDLE_DOMAIN) < 0)
				goto out;
	}

	if(__mbsend(cmd_h, cmd_l, data) < 0)
		print_mb_busy_abort(MBCMD(PM));

out:
	up(&mbsend_sem);
	return ret;
}

int omap_dsp_mbsend_and_wait_exarg(unsigned char cmd_h, unsigned char cmd_l,
				   unsigned short data, unsigned char tid,
				   int argc, unsigned short *argv,
				   wait_queue_head_t *q)
{
	long current_state;
	DECLARE_WAITQUEUE(wait, current);

	add_wait_queue(q, &wait);
	current_state = current->state;
	set_current_state(TASK_INTERRUPTIBLE);
	if (omap_dsp_mbsend_exarg(cmd_h, cmd_l, data, tid, argc, argv) < 0) {
		set_current_state(current_state);
		remove_wait_queue(q, &wait);
		return -1;
	}
	schedule();
	set_current_state(current_state);
	remove_wait_queue(q, &wait);

	return 0;
}

void omap_dsp_mb_config(void)
{
	static struct sync_seq dummy;

	mbseq_send = 0;
	mbseq_expect = 0;
	sync_seq = &dummy;	/* temporarily used before dspcfg */
}

void omap_dsp_mb_unconfig(void)
{
}

static void __init omap_dsp_mb_init(void)
{
	init_MUTEX(&mbsend_sem);
}

/*
 * mailbox interrupt handlers
 */
extern void mailbox1_dspcfg(unsigned char cmd_l, unsigned short data);
extern void mailbox1_regrw(unsigned char cmd_l, unsigned short data);
extern void mailbox1_getvar(unsigned char varid, unsigned short data);
extern void mailbox1_wdsnd(unsigned char tid, unsigned short data);
extern void mailbox1_wdreq(unsigned char tid);
extern void mailbox1_bksnd(unsigned char tid, unsigned short bid);
extern void mailbox1_bkreq(unsigned char tid, unsigned short cnt);
extern void mailbox1_bkyld(unsigned short bid);
extern void mailbox1_bksndp(unsigned char tid);
extern void mailbox1_bkreqp(unsigned char tid);
extern void mailbox1_tctl(unsigned char tid, unsigned short data);
extern void mailbox1_tcfg(unsigned char tid, unsigned short data);
extern void mailbox1_tadd(unsigned char tid);
extern void mailbox1_tdel(unsigned char tid);
extern void mailbox1_tdel(unsigned char tid);
extern void mailbox1_err_ipbfull(void);
extern void mailbox1_err_fatal(unsigned char tid);
extern void mailbox1_dbg(unsigned char tid);
extern void mailbox1_mapreq(unsigned char tid);
extern void mailbox1_wdt(unsigned short data);

static void mailbox1_err(unsigned char eid, unsigned short data)
{
	char *eidnm = subcmd_name(MBCMD(ERR), eid);
	unsigned char tid;

	if (eidnm) {
		printk(KERN_WARNING
		       "mailbox: ERR from DSP (%s): 0x%04x\n", eidnm, data);
	} else {
		printk(KERN_WARNING
		       "mailbox: ERR from DSP (unknown EID=0x%02x): 0x%04x\n",
		       eid, data);
	}

	switch (eid) {
	case OMAP_DSP_EID_IPBFULL:
		mailbox1_err_ipbfull();
		break;

	case OMAP_DSP_EID_FATAL:
		tid = data & 0x00ff;
		mailbox1_err_fatal(tid);
		break;
	}
}

static irqreturn_t mailbox1_interrupt(int irq, void *dev_id,
				      struct pt_regs *regs)
{
	unsigned short cmd, data;
	unsigned char cmd_h, cmd_l;
	unsigned char seq;

#if (INT_D2A_MB1 == INT_DSP_MAILBOX1)
	data = omap_readw(MAILBOX_DSP2ARM1);
	cmd  = omap_readw(MAILBOX_DSP2ARM1b);
#elif (INT_D2A_MB1 == INT_DSP_MAILBOX2)
	data = omap_readw(MAILBOX_DSP2ARM2);
	cmd  = omap_readw(MAILBOX_DSP2ARM2b);
#endif
	seq   = mb_sqbit(cmd);
	cmd_h = mb_cmd_h(cmd);
	cmd_l = mb_cmd_l(cmd);

	if (seq != (mbseq_expect & 1)) {
		switch (mbseq_chklevel) {
		case 0:	/* no check */
			break;
		case 1:	/* notify */
			printk(KERN_INFO
			       "mailbox: illegal seq bit!!!"
			       "  ignoring this command."
			       " (%04x:%04x)\n", cmd, data);
			return IRQ_HANDLED;
		case 2:	/* discard */
			return IRQ_HANDLED;
		}
	}

	omap_dsp_mblog_add(cmd, data, MBLOG_DIR_DA);
	omap_dsp_mblog_printcmd(cmd, data, MBLOG_DIR_DA);

	switch (cmd_h) {
	case MBCMD(WDSND):
		mailbox1_wdsnd(cmd_l, data);
		break;

	case MBCMD(WDREQ):
		mailbox1_wdreq(cmd_l);
		break;

	case MBCMD(BKSND):
		mailbox1_bksnd(cmd_l, data);
		break;

	case MBCMD(BKREQ):
		mailbox1_bkreq(cmd_l, data);
		break;

	case MBCMD(BKYLD):
		mailbox1_bkyld(data);
		break;

	case MBCMD(BKSNDP):
		mailbox1_bksndp(cmd_l);
		break;

	case MBCMD(BKREQP):
		mailbox1_bkreqp(cmd_l);
		break;

	case MBCMD(TCTL):
		mailbox1_tctl(cmd_l, data);
		break;

	case MBCMD(WDT):
		mailbox1_wdt(data);
		break;

	case MBCMD(TCFG):
		mailbox1_tcfg(cmd_l, data);
		break;

	case MBCMD(TADD):
		mailbox1_tadd(cmd_l);
		break;

	case MBCMD(TDEL):
		mailbox1_tdel(cmd_l);
		break;

	case MBCMD(DSPCFG):
		mailbox1_dspcfg(cmd_l, data);
		break;

	case MBCMD(REGRW):
		mailbox1_regrw(cmd_l, data);
		break;

	case MBCMD(GETVAR):
		mailbox1_getvar(cmd_l, data);
		break;

	case MBCMD(ERR):
		mailbox1_err(cmd_l, data);
		break;

	case MBCMD(DBG):
		mailbox1_dbg(cmd_l);
		break;

	case MBCMD(RUNLEVEL):
	case MBCMD(PM):
	case MBCMD(SETVAR):
		printk(KERN_ERR
		       "mailbox: %s is not allowed from DSP.\n",
		       cmd_name(cmd_h));

	default:
		printk(KERN_ERR
		       "mailbox: Unrecognized command: "
		       "cmd=0x%04x, data=0x%04x\n", cmd, data);
	}

	sync_seq->da_arm = ++mbseq_expect;
	return IRQ_HANDLED;
}

static irqreturn_t mailbox2_interrupt(int irq, void *dev_id,
				      struct pt_regs *regs)
{
	unsigned short cmd, data;

#if (INT_D2A_MB1 == INT_DSP_MAILBOX1)
	data = omap_readw(MAILBOX_DSP2ARM2);
	cmd  = omap_readw(MAILBOX_DSP2ARM2b);
#elif (INT_D2A_MB1 == INT_DSP_MAILBOX2)
	data = omap_readw(MAILBOX_DSP2ARM1);
	cmd  = omap_readw(MAILBOX_DSP2ARM1b);
#endif
	printk(KERN_DEBUG
	       "mailbox2 interrupt!  cmd=%04x, data=%04x\n", cmd, data);

	return IRQ_HANDLED;
}

#if 0
static void mpuio_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	printk(KERN_INFO "MPUIO interrupt!\n");
}
#endif

#ifdef CONFIG_PROC_FS
/*
 * proc file entry
 */
struct proc_dir_entry *procdir_dsp = NULL;

static void omap_dsp_create_procdir_dsp(void)
{
	procdir_dsp = proc_mkdir("dsp", 0);
	if (procdir_dsp == NULL) {
		printk(KERN_ERR
		       "omapdsp: failed to register proc directory: dsp\n");
	}
}

static void omap_dsp_remove_procdir_dsp(void)
{
	procdir_dsp = NULL;
	remove_proc_entry("dsp", 0);
}

#endif /* CONFIG_PROC_FS */

static char dev_id_mailbox1;	/* unique pointer for request_irq */
static char dev_id_mailbox2;	/* unique pointer for request_irq */
static char dev_id_dspmmu;	/* unique pointer for request_irq */
/*
static char dev_id_mpuio;
*/

extern irqreturn_t dspmmu_interrupt(int irq, void *dev_id,
				    struct pt_regs *regs);

extern int  dsp_ctl_core_init(void);
extern void dsp_ctl_core_exit(void);
extern void omap_dsp_ctl_init(void);
extern void omap_dsp_ctl_exit(void);
extern int  omap_dsp_mem_init(void);
extern void omap_dsp_mem_exit(void);
extern void omap_dsp_twch_init(void);
extern void omap_dsp_twch_exit(void);
extern void omap_dsp_err_init(void);
extern void omap_dsp_err_exit(void);
extern void omap_dsp_mblog_init(void);
extern void omap_dsp_mblog_exit(void);
extern int  omap_dsp_taskmod_init(void);
extern void omap_dsp_taskmod_exit(void);

static int __init omap_dsp_mod_init(void)
{
	int ret;

	printk(KERN_INFO "OMAP DSP driver initialization\n");

#ifdef CONFIG_PROC_FS
	omap_dsp_create_procdir_dsp();
#endif /* CONFIG_PROC_FS */

	if ((ret = dsp_ctl_core_init()) < 0)
		goto fail1;
	if ((ret = omap_dsp_mem_init()) < 0)
		goto fail2;
	omap_dsp_mb_init();
	omap_dsp_cur_users_init();
	omap_dsp_ctl_init();
	omap_dsp_twch_init();
	omap_dsp_err_init();
	omap_dsp_mblog_init();
	if ((ret = omap_dsp_taskmod_init()) < 0)
		goto fail3;

	/*
	 * mailbox interrupt handlers registration
	 */
#if (INT_D2A_MB1 == INT_DSP_MAILBOX1)
#	define INT_D2A_MB2 INT_DSP_MAILBOX2
#elif(INT_D2A_MB1 == INT_DSP_MAILBOX2)	/* swap MB1 and MB2 */
#	define INT_D2A_MB2 INT_DSP_MAILBOX1
#endif
	ret = request_irq(INT_D2A_MB1, mailbox1_interrupt, SA_INTERRUPT,
			  "mailbox1", (void *)&dev_id_mailbox1);
	if (ret) {
		printk(KERN_ERR
		       "omapdsp: failed to register mailbox1 interrupt: %d\n",
		       ret);
		goto fail4;
	}

	ret = request_irq(INT_D2A_MB2, mailbox2_interrupt, SA_INTERRUPT,
			  "mailbox2", (void *)&dev_id_mailbox2);
	if (ret) {
		printk(KERN_ERR
		       "omapdsp: failed to register mailbox2 interrupt: %d\n",
		       ret);
		goto fail5;
	}

	ret = request_irq(INT_DSP_MMU, dspmmu_interrupt, SA_INTERRUPT,
			  "dspmmu", (void *)&dev_id_dspmmu);
	if (ret) {
		printk(KERN_ERR
		       "omapdsp: failed to register DSP MMU interrupt: %d\n",
		       ret);
		goto fail6;
	}

#if 0
	ret = request_irq(INT_MPUIO, mpuio_interrupt, SA_INTERRUPT,
			  "mpuio", (void *)&dev_id_mpuio);
	if (ret) {
		printk(KERN_ERR
		       "omapdsp: failed to register "
		       "MPUIO interrupt: %d\n", ret);
		goto fail6;
	}
#endif

	return 0;

fail6:
	free_irq(INT_D2A_MB2, (void *)&dev_id_mailbox2);
fail5:
	free_irq(INT_D2A_MB1, (void *)&dev_id_mailbox1);
fail4:
	omap_dsp_taskmod_exit();
fail3:
	omap_dsp_mblog_exit();
	omap_dsp_err_exit();
	omap_dsp_twch_exit();
	omap_dsp_ctl_exit();
	omap_dsp_cur_users_exit();
	omap_dsp_mem_exit();
fail2:
	dsp_ctl_core_exit();
fail1:
#ifdef CONFIG_PROC_FS
	omap_dsp_remove_procdir_dsp();
#endif /* CONFIG_PROC_FS */

	return ret;
}

static void __exit omap_dsp_mod_exit(void)
{
	__dsp_reset();

#if 0
	free_irq(INT_MPUIO, (void *)&dev_id_mpuio);
#endif
	free_irq(INT_DSP_MMU, (void *)&dev_id_dspmmu);
	free_irq(INT_D2A_MB2, (void *)&dev_id_mailbox2);
	free_irq(INT_D2A_MB1, (void *)&dev_id_mailbox1);

	omap_dsp_dspuncfg();
	omap_dsp_taskmod_exit();
	omap_dsp_mblog_exit();
	omap_dsp_err_exit();
	omap_dsp_twch_exit();
	omap_dsp_ctl_exit();
	omap_dsp_cur_users_exit();
	omap_dsp_mem_exit();

	dsp_ctl_core_exit();
#ifdef CONFIG_PROC_FS
	omap_dsp_remove_procdir_dsp();
#endif /* CONFIG_PROC_FS */
}

module_init(omap_dsp_mod_init);
module_exit(omap_dsp_mod_exit);
