/*
 * linux/arch/arm/mach-omap/dsp/error.c
 *
 * OMAP DSP error detection I/F device driver
 *
 * Copyright (C) 2002-2004 Nokia Corporation
 *
 * Written by Toshihiro Kobayashi <toshihiro.kobayashi@nokia.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: error.c
 * $Revision: 3.0.1
 * $Date: 2004/10/04
 *
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/major.h>
#include <linux/fs.h>
#include <linux/poll.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/ioctls.h>
#include <asm/arch/dsp.h>
#include "dsp.h"

static wait_queue_head_t err_wait_q;
static unsigned long errcode = 0;
static int errcnt = 0;
static unsigned short wdtval;	/* FIXME: read through ioctl */
static unsigned long mmu_fadr;	/* FIXME: read through ioctl */

/*
 * DSP error detection device file operations
 */
static ssize_t omap_dsp_err_read(struct file *file, char *buf,
				 size_t count, loff_t *ppos)
{
	unsigned long flags;
	int status;

	if (count < 4)
		return 0;

	if (errcnt == 0) {
		long current_state;
		DECLARE_WAITQUEUE(wait, current);

		add_wait_queue(&err_wait_q, &wait);
		current_state = current->state;
		set_current_state(TASK_INTERRUPTIBLE);
		if (errcnt == 0)	/* last check */
			schedule();
		set_current_state(current_state);
		remove_wait_queue(&err_wait_q, &wait);
		if (signal_pending(current))
			return -EINTR;
	}

	save_flags(flags); cli();
	status = copy_to_user(buf, &errcode, 4);
	if (status) {
		restore_flags(flags);
		return -EFAULT;
	}
	errcnt = 0;
	restore_flags(flags);

	return 4;
}

static unsigned int omap_dsp_err_poll(struct file *file, poll_table *wait)
{
	unsigned int mask = 0;

	poll_wait(file, &err_wait_q, wait);
	if (errcnt != 0)
		mask |= POLLIN | POLLRDNORM;

	return mask;
}

struct file_operations omap_dsp_err_fops = {
	.owner = THIS_MODULE,
	.poll  = omap_dsp_err_poll,
	.read  = omap_dsp_err_read,
};

/*
 * DSP MMU
 */
void omap_dsp_err_mmu_set(unsigned long adr)
{
	disable_irq(INT_DSP_MMU);
	errcode |= OMAP_DSP_ERRDT_MMU;
	errcnt++;
	mmu_fadr = adr;
	wake_up_interruptible(&err_wait_q);
}

void omap_dsp_err_mmu_clear(void)
{
	errcode &= ~OMAP_DSP_ERRDT_MMU;
	enable_irq(INT_DSP_MMU);
}

int omap_dsp_err_mmu_isset(void)
{
	return (errcode & OMAP_DSP_ERRDT_MMU) ? 1 : 0;
}

/*
 * WDT
 */
void mailbox1_wdt(unsigned short data)
{
	errcode |= OMAP_DSP_ERRDT_WDT;
	errcnt++;
	wdtval = data;
	wake_up_interruptible(&err_wait_q);
}

void omap_dsp_err_wdt_clear(void)
{
	errcode &= ~OMAP_DSP_ERRDT_WDT;
}

int omap_dsp_err_wdt_isset(void)
{
	return (errcode & OMAP_DSP_ERRDT_WDT) ? 1 : 0;
}

/*
 *
 */
void omap_dsp_err_config(void)
{
	errcnt = 0;
	if (omap_dsp_err_wdt_isset())
		omap_dsp_err_wdt_clear();
	if (omap_dsp_err_mmu_isset())
		omap_dsp_err_mmu_clear();
}

void omap_dsp_err_unconfig(void)
{
	wake_up_interruptible(&err_wait_q);
}

void __init omap_dsp_err_init(void)
{
	errcode = 0;
	errcnt = 0;
	init_waitqueue_head(&err_wait_q);
}

void omap_dsp_err_exit(void)
{
}
