/*
 * linux/arch/arm/mach-omap/dsp/task.c
 *
 * OMAP DSP task device driver
 *
 * Copyright (C) 2002-2004 Nokia Corporation
 *
 * Written by Toshihiro Kobayashi <toshihiro.kobayashi@nokia.com>
 * mmap function by Hiroo Ishikawa <ext-hiroo.ishikawa@nokia.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * $Id: task.c
 * $Revision: 3.0.1
 * $Date: 2004/10/04
 *
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/major.h>
#include <linux/fs.h>
#include <linux/poll.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/slab.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/proc_fs.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/signal.h>
#include <asm/irq.h>
#include <asm/ioctls.h>
#include <asm/arch/dsp.h>
#include "uaccess_dsp.h"
#include "dsp.h"
#include "ipbuf.h"
#include "fifo.h"
#include "proclist.h"

struct taskdev {
	int enable;
	int openstat;
	unsigned int usecount;
	char name[OMAP_DSP_TNM_LEN];
	struct file_operations fops;
	wait_queue_head_t open_wait_q;
	struct semaphore dld_sem;
	struct dsptask *task;
	struct proc_dir_entry *procent_dir;
};

struct rcvdt_bk_struct {
	struct ipblink link;
	unsigned int rp;
	struct ipbuf_p *ipbuf_pvt_r;
};

struct dsptask {
	int cfgstat;
	struct list_head proc_list;
	unsigned char tid;
	char name[OMAP_DSP_TNM_LEN];
	unsigned short ttyp;

	/* read stuff */
	union {
		struct fifo_struct fifo;	/* for active word */
		struct rcvdt_bk_struct bk;
	} rcvdt;
	wait_queue_head_t read_wait_q;
#ifdef CONFIG_OMAP_DSP_TASK_MULTIOPEN
	struct semaphore read_sem;
#endif

	/* write stuff */
	size_t wsz;
	struct ipbuf_p *ipbuf_pvt_w;	/* for private block */
	wait_queue_head_t write_wait_q;	/* for active */
#ifdef CONFIG_OMAP_DSP_TASK_MULTIOPEN
	struct semaphore write_sem;
#endif

	/* ioctl stuff */
	int tctl_stat;
	wait_queue_head_t ioctl_wait_q;
#ifdef CONFIG_OMAP_DSP_TASK_MULTIOPEN
	struct semaphore ioctl_sem;
#endif

	/* mmap stuff */
	void *map_base;
	size_t map_length;

#ifdef CONFIG_OMAP_DSP_TASK_MULTIOPEN
	struct semaphore task_sem;
	pid_t lock_pid;
#endif
};

#define sndtyp_acv(ttyp)	((ttyp) & OMAP_DSP_TTYP_ASND)
#define sndtyp_psv(ttyp)	(!((ttyp) & OMAP_DSP_TTYP_ASND))
#define sndtyp_bk(ttyp)		((ttyp) & OMAP_DSP_TTYP_BKDM)
#define sndtyp_wd(ttyp)		(!((ttyp) & OMAP_DSP_TTYP_BKDM))
#define sndtyp_pvt(ttyp)	((ttyp) & OMAP_DSP_TTYP_PVDM)
#define sndtyp_gbl(ttyp)	(!((ttyp) & OMAP_DSP_TTYP_PVDM))
#define rcvtyp_acv(ttyp)	((ttyp) & OMAP_DSP_TTYP_ARCV)
#define rcvtyp_psv(ttyp)	(!((ttyp) & OMAP_DSP_TTYP_ARCV))
#define rcvtyp_bk(ttyp)		((ttyp) & OMAP_DSP_TTYP_BKMD)
#define rcvtyp_wd(ttyp)		(!((ttyp) & OMAP_DSP_TTYP_BKMD))
#define rcvtyp_pvt(ttyp)	((ttyp) & OMAP_DSP_TTYP_PVMD)
#define rcvtyp_gbl(ttyp)	(!((ttyp) & OMAP_DSP_TTYP_PVMD))

static __inline__ int down_tasksem_interruptible(struct dsptask *task,
						 struct semaphore *sem)
{
	int sem_st;

	if (task->lock_pid == current->pid) {
		/* this process has lock */
		sem_st = down_interruptible(sem);
	} else {
		if ((sem_st = down_interruptible(&task->task_sem)) < 0)
			return sem_st;
		sem_st = down_interruptible(sem);
		up(&task->task_sem);
	}
	return sem_st;
}

static int omap_dsp_rmdev_minor(unsigned char minor);
static int omap_dsp_taskdev_init(struct taskdev *dev, char *name,
				 unsigned char minor);
static void omap_dsp_taskdev_delete(unsigned char minor);
static void omap_dsp_taskdev_attach_task(struct taskdev *dev,
					 struct dsptask *task);
static void omap_dsp_taskdev_detach_task(struct taskdev *dev);

#ifdef CONFIG_PROC_FS
static void omap_dsp_taskdev_create_proc(struct taskdev *dev);
static void omap_dsp_taskdev_remove_proc(struct taskdev *dev);
#endif

static struct taskdev *taskdev[TASKDEV_MAX];
static struct dsptask *dsptask[TASKDEV_MAX];
/*
 * tcfg_wait_q is used by tcfg, tadd and tdel (i.e. in ctl device
 * context).  It is reasonable because it is guaranteed that the ctl
 * device is not opened by two or more processes at any moment.
 */
static wait_queue_head_t tcfg_wait_q;
static unsigned short tcfg_wait_cmd = 0;
static int taskmod_cfgstat = CFGSTAT_ERR;
static unsigned char tadd_tid = OMAP_DSP_TID_ANON;
static unsigned char tdel_tid = OMAP_DSP_TID_ANON;
static unsigned char n_task = 0;
static void *heap = NULL;
#ifdef CONFIG_PROC_FS
struct proc_dir_entry *procdir_dsptask = NULL;
#endif

static int omap_dsp_task_flush_buf(struct dsptask *task)
{
	unsigned short ttyp = task->ttyp;

	preempt_disable();
	if (rcvtyp_wd(ttyp)) {
		/* word receiving */
		flush_fifo(&task->rcvdt.fifo);
	} else {
		/* block receiving */
		struct rcvdt_bk_struct *rcvdt = &task->rcvdt.bk;

		if (rcvtyp_gbl(ttyp)) {
			/* global IPBUF */
			while (!ipblink_empty(&rcvdt->link)) {
				unsigned short bid = rcvdt->link.top;
				disable_irq(INT_D2A_MB1);
				ipblink_del_top(&rcvdt->link, ipbuf);
				enable_irq(INT_D2A_MB1);
				unuse_ipbuf(bid, MBSENDTYPE_NORMAL);
			}
		} else {
			/* private IPBUF */
			if (!ipblink_empty(&rcvdt->link)) {
				ipblink_del_pvt(&rcvdt->link);
				release_ipbuf_pvt(rcvdt->ipbuf_pvt_r);
			}
		}
	}
	preempt_enable();

	return 0;
}

static int omap_dsp_task_set_fifosz(struct dsptask *task, unsigned long sz)
{
	unsigned short ttyp = task->ttyp;
	struct fifo_struct *fifo = &task->rcvdt.fifo;
	int stat;

	if (!(sndtyp_wd(ttyp) && sndtyp_acv(ttyp))) {
		printk(KERN_ERR
		       "omapdsp: buffer size can be changed only for "
		       "active word sending task.\n");
		return -EINVAL;
	}
	if (sz == 0) {
		printk(KERN_ERR "omapdsp: buffer size shouldn't be zero!\n");
		return -EINVAL;
	}
	preempt_disable();
	if (!fifo_empty(fifo)) {
		printk(KERN_ERR "omapdsp: buffer is not empty!\n");
		preempt_enable();
		return -EINVAL;
	}
	if (sz & 0x1) {
		/* force even value */
		sz++;
	}

	free_fifo(fifo);
	stat = alloc_fifo(fifo, sz);
	preempt_enable();
	if (stat < 0) {
		printk(KERN_ERR
		       "omapdsp: unable to change receive buffer size. "
		       "(%ld bytes for %s)\n", sz, task->name);
		return -ENOMEM;
	}

	return 0;
}

static int omap_dsp_task_lock(struct dsptask *task)
{
#ifdef CONFIG_OMAP_DSP_TASK_MULTIOPEN
	int sem_st;
	if ((sem_st = down_interruptible(&task->task_sem)) < 0)
		return sem_st;
	task->lock_pid = current->pid;
#endif
	return 0;
}

static int omap_dsp_task_unlock(struct dsptask *task)
{
#ifdef CONFIG_OMAP_DSP_TASK_MULTIOPEN
	if (task->lock_pid != current->pid) {
		printk(KERN_ERR
		       "omapdsp: an illegal process attempted to "
		       "unlock the dsptask lock!\n");
		return -EINVAL;
	}
	task->lock_pid = 0;
	up(&task->task_sem);
#endif
	return 0;
}

static int omap_dsp_task_config(struct dsptask *task, unsigned char tid)
{
	unsigned short ttyp;

	dsptask[tid] = task;
	task->tid = tid;

	INIT_LIST_HEAD(&task->proc_list);
	init_waitqueue_head(&task->read_wait_q);
	init_waitqueue_head(&task->write_wait_q);
	init_waitqueue_head(&task->ioctl_wait_q);
#ifdef CONFIG_OMAP_DSP_TASK_MULTIOPEN
	init_MUTEX(&task->read_sem);
	init_MUTEX(&task->write_sem);
	init_MUTEX(&task->ioctl_sem);
	init_MUTEX(&task->task_sem);
	task->lock_pid = 0;
#endif

	/* TCFG request */
	task->cfgstat = CFGSTAT_GOING;
	tcfg_wait_cmd = MBCMD(TCFG);
	omap_dsp_mbsend_and_wait(MBCMD(TCFG), tid, 0, &tcfg_wait_q);

	if (task->cfgstat != CFGSTAT_DONE) {
		printk(KERN_ERR "omapdsp: task %d configuration error!\n", tid);
		return -EINVAL;
	}

	if (strlen(task->name) <= 1)
		sprintf(task->name, "%d", tid);
	printk(KERN_INFO "omapdsp: task %d: name %s\n", tid, task->name);

	ttyp = task->ttyp;

	/* task type check */
	if (rcvtyp_psv(ttyp) && rcvtyp_pvt(ttyp)) {
		printk(KERN_ERR
		       "mailbox: illegal task type(0x%04x), tid=%d\n",
		       tid, ttyp);
	}

	/* private buffer address check */
	if (sndtyp_pvt(ttyp)) {
		void *p = task->rcvdt.bk.ipbuf_pvt_r;

		if ((unsigned long)p & 0x1) {
			printk(KERN_ERR
			       "mailbox: private ipbuf (DSP->ARM) "
			       "address (0x%p) is odd number!\n", p);
			return -EINVAL;
		}
	}

	if (rcvtyp_pvt(ttyp)) {
		void *p = task->ipbuf_pvt_w;

		if ((unsigned long)p & 0x1) {
			printk(KERN_ERR
			       "mailbox: private ipbuf (ARM->DSP) "
			       "address (0x%p) is odd number!\n", p);
			return -EINVAL;
		}
	}

	/* read initialization */
	if (sndtyp_wd(ttyp)) {
		/* word */
		size_t fifosz;

		fifosz = sndtyp_psv(ttyp) ? 2 :	/* passive */
					    32;	/* active */
		if (alloc_fifo(&task->rcvdt.fifo, fifosz) < 0) {
			printk(KERN_ERR
			       "omapdsp: unable to allocate "
			       "receive buffer. (%d bytes for %s)\n",
			       fifosz, task->name);
			return -ENOMEM;
		}
	} else {
		/* block */
		INIT_IPBLINK(&task->rcvdt.bk.link);
		task->rcvdt.bk.rp = 0;
	}

	/* write initialization */
	task->wsz = rcvtyp_acv(ttyp) ? 0 :		/* active */
		    rcvtyp_wd(ttyp)  ? 2 :		/* passive word */
		    		       ipbcfg.lsz*2;	/* passive block */

	omap_dsp_mbsend(MBCMD(TCTL), tid, OMAP_DSP_MBCMD_TCTL_TINIT);

	return 0;
}

int omap_dsp_task_config_all(unsigned char n)
{
	int i, ret;
	struct taskdev *devheap;
	struct dsptask *taskheap;
	size_t devheapsz, taskheapsz;

	memset(taskdev, 0, sizeof(void *) * TASKDEV_MAX);
	memset(dsptask, 0, sizeof(void *) * TASKDEV_MAX);

	n_task = n;
	printk(KERN_INFO "omapdsp: found %d task(s)\n", n_task);
	if (n_task == 0) {
		taskmod_cfgstat = CFGSTAT_DONE;
		return 0;
	}

	/*
	 * reducing kmalloc!
	 */
	devheapsz  = sizeof(struct taskdev) * n_task;
	taskheapsz = sizeof(struct dsptask) * n_task;
	heap = kmalloc(devheapsz + taskheapsz, GFP_KERNEL);
	if (heap == NULL) {
		n_task = 0;
		return -ENOMEM;
	}
	memset(heap, 0, devheapsz + taskheapsz);
	devheap  = heap;
	taskheap = heap + devheapsz;

	for (i = 0; i < n_task; i++) {
		struct taskdev *dev  = &devheap[i];
		struct dsptask *task = &taskheap[i];

		if ((ret = omap_dsp_task_config(task, i)) < 0)
			return ret;
		if ((ret = omap_dsp_taskdev_init(dev, task->name, i)) < 0)
			return ret;
		omap_dsp_taskdev_attach_task(dev, task);
	}

	taskmod_cfgstat = CFGSTAT_DONE;
	return 0;
}

static void omap_dsp_task_unconfig(struct dsptask *task)
{
	unsigned char tid = task->tid;

	preempt_disable();
	proc_list_flush(&task->proc_list);
	if (sndtyp_wd(task->ttyp) && (task->cfgstat == CFGSTAT_DONE))
		free_fifo(&task->rcvdt.fifo);
	dsptask[tid] = NULL;
	preempt_enable();
}

void omap_dsp_task_unconfig_all(void)
{
	unsigned char minor;
	unsigned char tid;
	struct dsptask *task;

	for (minor = 0; minor < n_task; minor++) {
		/*
		 * taskdev[minor] can be NULL in case of
		 * configuration failure
		 */
		if (taskdev[minor])
			omap_dsp_taskdev_delete(minor);
	}
	for (; minor < TASKDEV_MAX; minor++) {
		if (taskdev[minor])
			omap_dsp_rmdev_minor(minor);
	}

	for (tid = 0; tid < n_task; tid++) {
		/*
		 * dsptask[tid] can be NULL in case of
		 * configuration failure
		 */
		task = dsptask[tid];
		if (task)
			omap_dsp_task_unconfig(task);
	}
	for (; tid < TASKDEV_MAX; tid++) {
		task = dsptask[tid];
		if (task) {
			/*
			 * on-demand tasks should be deleted in
			 * rmdev_minor(), but just in case.
			 */
			omap_dsp_task_unconfig(task);
			kfree(task);
		}
	}

	if (heap) {
		kfree(heap);
		heap = NULL;
	}

	n_task = 0;
	taskmod_cfgstat = CFGSTAT_ERR;
}

unsigned char dsp_task_count(void)
{
	return n_task;
}

int omap_dsp_taskmod_busy(void)
{
	struct taskdev *dev;
	unsigned char minor;

	for (minor = 0; minor < TASKDEV_MAX; minor++) {
		dev = taskdev[minor];
		if (dev &&
		    ((dev->usecount > 0) ||
		     waitqueue_active(&dev->open_wait_q)))
			return 1;
	}
	return 0;
}

/*
 * DSP task device file operations
 */
static ssize_t omap_dsp_task_read_wd_acv(struct file *file, char *buf,
					 size_t count, loff_t *ppos)
{
	unsigned int minor = MINOR(file->f_dentry->d_inode->i_rdev);
	struct dsptask *task = taskdev[minor]->task;
	struct fifo_struct *fifo = &task->rcvdt.fifo;
#ifdef CONFIG_OMAP_DSP_TASK_MULTIOPEN
	int sem_st;
#endif
	int ret = 0;

	if (count == 0) {
		return 0;
	} else if (count == 1) {
		printk(KERN_ERR
		       "omapdsp: count == 1 is illegal "
		       "for omap_dsp_task_read().\n");
		return -EINVAL;
	} else if (count & 0x1) {
		/* force even value */
		count--;
	}

#ifdef CONFIG_OMAP_DSP_TASK_MULTIOPEN
	if ((sem_st = down_tasksem_interruptible(task, &task->read_sem)) < 0)
		return sem_st;
#endif

	if (fifo_empty(fifo)) {
		long current_state;
		DECLARE_WAITQUEUE(wait, current);

		add_wait_queue(&task->read_wait_q, &wait);
		current_state = current->state;
		set_current_state(TASK_INTERRUPTIBLE);
		if (fifo_empty(fifo))	/* last check */
			schedule();
		set_current_state(current_state);
		remove_wait_queue(&task->read_wait_q, &wait);
		if (fifo_empty(fifo))
			goto up_out;
	}

	/* data protection while copying */
	disable_irq(INT_D2A_MB1);
	ret = copy_to_user_fm_fifo(buf, fifo, count);
	enable_irq(INT_D2A_MB1);

up_out:
#ifdef CONFIG_OMAP_DSP_TASK_MULTIOPEN
	up(&task->read_sem);
#endif
	return ret;
}

static ssize_t omap_dsp_task_read_bk_acv(struct file *file, char *buf,
					 size_t count, loff_t *ppos)
{
	unsigned int minor = MINOR(file->f_dentry->d_inode->i_rdev);
	struct dsptask *task = taskdev[minor]->task;
	struct rcvdt_bk_struct *rcvdt = &task->rcvdt.bk;
	int dspmem_access;
#ifdef CONFIG_OMAP_DSP_TASK_MULTIOPEN
	int sem_st;
#endif
	ssize_t ret = 0;

	if (count == 0) {
		return 0;
	} else if (count == 1) {
		printk(KERN_ERR
		       "omapdsp: count == 1 is illegal "
		       "for omap_dsp_task_read().\n");
		return -EINVAL;
	} else if ((int)buf & 0x1) {
		printk(KERN_ERR
		       "omapdsp: buf should be word aligned "
		       "for omap_dsp_task_read().\n");
		return -EINVAL;
	} else if (count & 0x1) {
		/* force even value */
		count--;
	}

#ifdef CONFIG_OMAP_DSP_TASK_MULTIOPEN
	if ((sem_st = down_tasksem_interruptible(task, &task->read_sem)) < 0)
		return sem_st;
#endif

	if (ipblink_empty(&rcvdt->link)) {
		long current_state;
		DECLARE_WAITQUEUE(wait, current);

		add_wait_queue(&task->read_wait_q, &wait);
		current_state = current->state;
		set_current_state(TASK_INTERRUPTIBLE);
		if (ipblink_empty(&rcvdt->link))	/* last check */
			schedule();
		set_current_state(current_state);
		remove_wait_queue(&task->read_wait_q, &wait);
		/* signal or 0-byte send from DSP */
		if (ipblink_empty(&rcvdt->link))
			goto up_out;
	}

	/* data protection while copying */
	disable_irq(INT_D2A_MB1);

	/* copy from delayed IPBUF */
	if (sndtyp_pvt(task->ttyp)) {
		/* private */
		if (!ipblink_empty(&rcvdt->link)) {
			struct ipbuf_p *ipbp = rcvdt->ipbuf_pvt_r;
			unsigned long dspadr = MKLONG(ipbp->ah, ipbp->al);
			unsigned char *src;
			size_t bkcnt;

			if ((dspmem_access = is_dspword_internal_mem(dspadr)) != 0)
				omap_dsp_enable_dspmem();
			src = dspword_to_virt(dspadr) + rcvdt->rp;
			bkcnt = ((unsigned long)ipbp->c) * 2 - rcvdt->rp;
			if (bkcnt > count) {
				if (copy_to_user_dsp(buf, src, count)) {
					ret = -EFAULT;
					goto copy_done_pvt;
				}
				ret = count;
				rcvdt->rp += count;
			} else {
				if (copy_to_user_dsp(buf, src, bkcnt)) {
					ret = -EFAULT;
					goto copy_done_pvt;
				}
				ret = bkcnt;
				ipblink_del_pvt(&rcvdt->link);
				release_ipbuf_pvt(ipbp);
				rcvdt->rp = 0;
			}
copy_done_pvt:
			if (dspmem_access)
				omap_dsp_disable_dspmem();
		}
	} else {
		/* global */
		if ((dspmem_access = is_ipbuf_internal_mem()) != 0)
			omap_dsp_enable_dspmem();
		while (!ipblink_empty(&rcvdt->link)) {
			unsigned char *src;
			size_t bkcnt;
			unsigned short bid = rcvdt->link.top;
			struct ipbuf *ipbp = ipbuf[bid];

			src = ipbp->d + rcvdt->rp;
			bkcnt = ((unsigned long)ipbp->c) * 2 - rcvdt->rp;
			if (bkcnt > count) {
				if (copy_to_user_dsp(buf, src, count)) {
					ret = -EFAULT;
					goto copy_done_gbl;
				}
				ret += count;
				rcvdt->rp += count;
				break;
			} else {
				if (copy_to_user_dsp(buf, src, bkcnt)) {
					ret = -EFAULT;
					goto copy_done_gbl;
				}
				ret += bkcnt;
				buf += bkcnt;
				count -= bkcnt;
				ipblink_del_top(&rcvdt->link, ipbuf);
				unuse_ipbuf(bid, MBSENDTYPE_NORMAL);
				rcvdt->rp = 0;
			}
		}
copy_done_gbl:
		if (dspmem_access)
			omap_dsp_disable_dspmem();
	}

	enable_irq(INT_D2A_MB1);
up_out:
#ifdef CONFIG_OMAP_DSP_TASK_MULTIOPEN
	up(&task->read_sem);
#endif
	return ret;
}

static ssize_t omap_dsp_task_read_wd_psv(struct file *file, char *buf,
					 size_t count, loff_t *ppos)
{
	unsigned int minor = MINOR(file->f_dentry->d_inode->i_rdev);
	struct dsptask *task = taskdev[minor]->task;
	struct fifo_struct *fifo = &task->rcvdt.fifo;
#ifdef CONFIG_OMAP_DSP_TASK_MULTIOPEN
	int sem_st;
#endif
	int ret = 0;

	if (count == 0) {
		return 0;
	} else if (count == 1) {
		printk(KERN_ERR
		       "omapdsp: count == 1 is illegal "
		       "for omap_dsp_task_read().\n");
		return -EINVAL;
	} else {
		/* force! */
		count = 2;
	}

#ifdef CONFIG_OMAP_DSP_TASK_MULTIOPEN
	if ((sem_st = down_tasksem_interruptible(task, &task->read_sem)) < 0)
		return sem_st;
#endif

	omap_dsp_mbsend_and_wait(MBCMD(WDREQ), task->tid, 0,
				 &task->read_wait_q);

	if (fifo_empty(fifo))
		goto up_out;
	ret = copy_to_user_fm_fifo(buf, fifo, count);

up_out:
#ifdef CONFIG_OMAP_DSP_TASK_MULTIOPEN
	up(&task->read_sem);
#endif
	return ret;
}

static ssize_t omap_dsp_task_read_bk_psv(struct file *file, char *buf,
					 size_t count, loff_t *ppos)
{
	unsigned int minor = MINOR(file->f_dentry->d_inode->i_rdev);
	struct dsptask *task = taskdev[minor]->task;
	struct rcvdt_bk_struct *rcvdt = &task->rcvdt.bk;
	int dspmem_access;
#ifdef CONFIG_OMAP_DSP_TASK_MULTIOPEN
	int sem_st;
#endif
	int ret = 0;

	if (count == 0) {
		return 0;
	} else if (count == 1) {
		printk(KERN_ERR
		       "omapdsp: count == 1 is illegal "
		       "for omap_dsp_task_read().\n");
		return -EINVAL;
	} else if ((int)buf & 0x1) {
		printk(KERN_ERR
		       "omapdsp: buf should be word aligned "
		       "for omap_dsp_task_read().\n");
		return -EINVAL;
	} else if (count & 0x1) {
		/* force even value */
		count--;
	}

#ifdef CONFIG_OMAP_DSP_TASK_MULTIOPEN
	if ((sem_st = down_tasksem_interruptible(task, &task->read_sem)) < 0)
		return sem_st;
#endif

	omap_dsp_mbsend_and_wait(MBCMD(BKREQ), task->tid, count/2,
				 &task->read_wait_q);

	/* signal or 0-byte send from DSP */
	if (ipblink_empty(&rcvdt->link))
		goto up_out;

	/*
	 * We will not receive more than requested count.
	 */
	if (sndtyp_pvt(task->ttyp)) {
		/* private */
		struct ipbuf_p *ipbp = rcvdt->ipbuf_pvt_r;
		size_t rcvcnt = ((unsigned long)ipbp->c) * 2;
		unsigned long dspadr = MKLONG(ipbp->ah, ipbp->al);
		void *src;

		if (count > rcvcnt)
			count = rcvcnt;
		if ((dspmem_access = is_dspword_internal_mem(dspadr)) != 0)
			omap_dsp_enable_dspmem();
		src = dspword_to_virt(dspadr);
		if (copy_to_user_dsp(buf, src, count)) {
			ret = -EFAULT;
			goto copy_done_pvt;
		}
		ipblink_del_pvt(&rcvdt->link);
		release_ipbuf_pvt(ipbp);
		ret = count;
copy_done_pvt:
		if (dspmem_access)
			omap_dsp_disable_dspmem();
	} else {
		/* global */
		unsigned short bid = rcvdt->link.top;
		struct ipbuf *ipbp = ipbuf[bid];
		size_t rcvcnt = ((unsigned long)ipbp->c) * 2;

		if ((dspmem_access = is_ipbuf_internal_mem()) != 0)
			omap_dsp_enable_dspmem();
		if (count > rcvcnt)
			count = rcvcnt;
		if (copy_to_user_dsp(buf, ipbp->d, count)) {
			ret = -EFAULT;
			goto copy_done_gbl;
		}
		ipblink_del_top(&rcvdt->link, ipbuf);
		unuse_ipbuf(bid, MBSENDTYPE_NORMAL);
		ret = count;
copy_done_gbl:
		if (dspmem_access)
			omap_dsp_disable_dspmem();
	}

up_out:
#ifdef CONFIG_OMAP_DSP_TASK_MULTIOPEN
	up(&task->read_sem);
#endif
	return count;
}

static ssize_t omap_dsp_task_write_wd(struct file *file, const char *buf,
				      size_t count, loff_t *ppos)
{
	unsigned int minor = MINOR(file->f_dentry->d_inode->i_rdev);
	struct dsptask *task = taskdev[minor]->task;
	unsigned short wd;
#ifdef CONFIG_OMAP_DSP_TASK_MULTIOPEN
	int sem_st;
#endif
	int ret = 0;

	if (count == 0) {
		return 0;
	} else if (count & 0x1) {
		printk(KERN_ERR
		       "omapdsp: odd count is illegal "
		       "for omap_dsp_task_write().\n");
		return -EINVAL;
	} else {
		/* force! */
		count = 2;
	}

#ifdef CONFIG_OMAP_DSP_TASK_MULTIOPEN
	if ((sem_st = down_tasksem_interruptible(task, &task->write_sem)) < 0)
		return sem_st;
#endif

	if (task->wsz == 0) {
		long current_state;
		DECLARE_WAITQUEUE(wait, current);

		add_wait_queue(&task->write_wait_q, &wait);
		current_state = current->state;
		set_current_state(TASK_INTERRUPTIBLE);
		if (task->wsz == 0)	/* last check */
			schedule();
		set_current_state(current_state);
		remove_wait_queue(&task->write_wait_q, &wait);
		/* signal or 0-byte send from DSP */
		if (task->wsz == 0)
			goto up_out;
	}

	if (copy_from_user(&wd, buf, count)) {
		ret = -EFAULT;
		goto up_out;
	}

	/* wsz protection */
	disable_irq(INT_D2A_MB1);
	if (omap_dsp_mbsend(MBCMD(WDSND), task->tid, wd) < 0)
		goto enirq_up_out;
	ret = count;
	if (rcvtyp_acv(task->ttyp))
		task->wsz = 0;

enirq_up_out:
	enable_irq(INT_D2A_MB1);
up_out:
#ifdef CONFIG_OMAP_DSP_TASK_MULTIOPEN
	up(&task->write_sem);
#endif
	return ret;
}

static ssize_t omap_dsp_task_write_bk(struct file *file, const char *buf,
				      size_t count, loff_t *ppos)
{
	unsigned int minor = MINOR(file->f_dentry->d_inode->i_rdev);
	struct dsptask *task = taskdev[minor]->task;
	unsigned char tid = task->tid;
	int dspmem_access;
#ifdef CONFIG_OMAP_DSP_TASK_MULTIOPEN
	int sem_st;
#endif
	int ret = 0;

	if (count == 0) {
		return 0;
	} else if ((int)buf & 0x1) {
		printk(KERN_ERR
		       "omapdsp: buf should be word aligned "
		       "for omap_dsp_task_write().\n");
		return -EINVAL;
	} else if (count & 0x1) {
		printk(KERN_ERR
		       "omapdsp: odd count is illegal "
		       "for omap_dsp_task_write().\n");
		return -EINVAL;
	}

#ifdef CONFIG_OMAP_DSP_TASK_MULTIOPEN
	if ((sem_st = down_tasksem_interruptible(task, &task->write_sem)) < 0)
		return sem_st;
#endif

	if (task->wsz == 0) {
		long current_state;
		DECLARE_WAITQUEUE(wait, current);

		add_wait_queue(&task->write_wait_q, &wait);
		current_state = current->state;
		set_current_state(TASK_INTERRUPTIBLE);
		if (task->wsz == 0)	/* last check */
			schedule();
		set_current_state(current_state);
		remove_wait_queue(&task->write_wait_q, &wait);
		/* interrupted by signal */
		if (task->wsz == 0)
			goto up_out;
	}

	/* wsz protection */
	disable_irq(INT_D2A_MB1);

	if (count > task->wsz)
		count = task->wsz;

	if (rcvtyp_pvt(task->ttyp)) {
		/* private */
		struct ipbuf_p *ipbp = task->ipbuf_pvt_w;
		unsigned long dspadr = MKLONG(ipbp->ah, ipbp->al);
		unsigned char *dst;

		if (ipbp->s != OMAP_DSP_TID_FREE)
			goto enirq_up_out;
		if ((dspmem_access = is_dspword_internal_mem(dspadr)) != 0)
			omap_dsp_enable_dspmem();
		dst = dspword_to_virt(dspadr);
		if (copy_from_user_dsp(dst, buf, count)) {
			ret = -EFAULT;
			if (dspmem_access)
				omap_dsp_disable_dspmem();
			goto enirq_up_out;
		}
		ipbp->c = count/2;
		ipbp->s = tid;
		if (dspmem_access)
			omap_dsp_disable_dspmem();
		if (omap_dsp_mbsend(MBCMD(BKSNDP), tid, 0) < 0)
			goto enirq_up_out;
	} else {
		/* global */
		unsigned short bid;

		if ((dspmem_access = is_ipbuf_internal_mem()) != 0)
			omap_dsp_enable_dspmem();
		bid = get_free_ipbuf(tid);
		if (bid == OMAP_DSP_BID_NULL)
			goto enirq_up_out;
		if (copy_from_user_dsp(ipbuf[bid]->d, buf, count)) {
			release_ipbuf(bid);
			ret = -EFAULT;
			if (dspmem_access)
				omap_dsp_disable_dspmem();
			goto enirq_up_out;
		}
		ipbuf[bid]->c  = count/2;
		ipbuf[bid]->sa = tid;
		if (dspmem_access)
			omap_dsp_disable_dspmem();
		if (omap_dsp_mbsend(MBCMD(BKSND), tid, bid) < 0) {
			release_ipbuf(bid);
			goto enirq_up_out;
		}
		ipb_bsycnt_inc(&ipbcfg);
	}
	ret = count;

	if (rcvtyp_acv(task->ttyp))
		task->wsz = 0;
enirq_up_out:
	enable_irq(INT_D2A_MB1);
up_out:
#ifdef CONFIG_OMAP_DSP_TASK_MULTIOPEN
	up(&task->write_sem);
#endif
	return ret;
}

static unsigned int omap_dsp_task_poll(struct file * file, poll_table * wait)
{
	unsigned int minor = MINOR(file->f_dentry->d_inode->i_rdev);
	struct dsptask *task = taskdev[minor]->task;
	unsigned int mask = 0;

	poll_wait(file, &task->read_wait_q, wait);
	poll_wait(file, &task->write_wait_q, wait);
	if (sndtyp_psv(task->ttyp) ||
	    (sndtyp_wd(task->ttyp) && !fifo_empty(&task->rcvdt.fifo)) ||
	    (sndtyp_bk(task->ttyp) && !ipblink_empty(&task->rcvdt.bk.link)))
		mask |= POLLIN | POLLRDNORM;
	if (task->wsz)
		mask |= POLLOUT | POLLWRNORM;

	return mask;
}

static int omap_dsp_task_ioctl(struct inode *inode, struct file *file,
			       unsigned int cmd, unsigned long arg)
{
	unsigned int minor = MINOR(inode->i_rdev);
	struct taskdev *dev = taskdev[minor];
	struct dsptask *task = dev->task;
	unsigned char tid = task->tid;
	unsigned short mbarg[1];
	int argc, interactive;
#ifdef CONFIG_OMAP_DSP_TASK_MULTIOPEN
	int sem_st;
#endif
	int ret;

	/* LOCK / UNLOCK operations */
	switch (cmd) {
	case OMAP_DSP_TASK_IOCTL_LOCK:
		return omap_dsp_task_lock(task);
	case OMAP_DSP_TASK_IOCTL_UNLOCK:
		return omap_dsp_task_unlock(task);
	}

#ifdef CONFIG_OMAP_DSP_TASK_MULTIOPEN
	/*
	 * actually only interractive commands need to lock
	 * the semaphore, but here all commands do it for simplicity.
	 */
	if ((sem_st = down_tasksem_interruptible(task, &task->ioctl_sem)) < 0)
		return sem_st;
#endif

	if ((cmd >= 0x0080) && (cmd < 0x0100)) {
		/*
		 * 0x0080 - 0x00ff
		 * reserved for backward compatibility
		 * user-defined TCTL commands: no arg, non-interactive
		 */
		argc = 0;
		interactive = 0;
	} else if (cmd < 0x8000) {
		/*
		 * 0x0000 - 0x7fff (except 0x0080 - 0x00ff)
		 * system reserved TCTL commands
		 */
		switch (cmd) {
		case OMAP_DSP_MBCMD_TCTL_TEN:
		case OMAP_DSP_MBCMD_TCTL_TDIS:
			argc = 0;
			interactive = 0;
			break;
		default:
			ret = -ENOIOCTLCMD;
			goto up_out;
		}
	}
	/*
	 * 0x8000 - 0xffff
	 * user-defined TCTL commands
	 */
	else if (cmd < 0x8100) {
		/* 0x8000-0x80ff: no arg, non-interactive */
		argc = 0;
		interactive = 0;
	} else if (cmd < 0x8200) {
		/* 0x8100-0x81ff: 1 arg, non-interactive */
		argc = 1;
		mbarg[0] = arg & 0xffff;
		interactive = 0;
	} else if (cmd < 0x9000) {
		/* 0x8200-0x8fff: reserved */
		ret = -ENOIOCTLCMD;
		goto up_out;
	} else if (cmd < 0x9100) {
		/* 0x9000-0x90ff: no arg, interactive */
		argc = 0;
		interactive = 1;
	} else if (cmd < 0x9200) {
		/* 0x9100-0x91ff: 1 arg, interactive */
		argc = 1;
		mbarg[0] = arg & 0xffff;
		interactive = 1;
	} else if (cmd < 0x10000) {
		/* 0x9200-0xffff: reserved */
		ret =  -ENOIOCTLCMD;
		goto up_out;
	} else {
		/*
		 * 0x10000 -
		 * non TCTL ioctls
		 */
		switch (cmd) {
		case OMAP_DSP_TASK_IOCTL_BFLSH:
			ret = omap_dsp_task_flush_buf(task);
			break;
		case OMAP_DSP_TASK_IOCTL_SETBSZ:
			ret = omap_dsp_task_set_fifosz(task, arg);
			break;
		case OMAP_DSP_TASK_IOCTL_GETNAME:
			ret = 0;
			if (copy_to_user((void *)arg, dev->name,
					 strlen(dev->name) + 1))
				ret = -EFAULT;
			break;
		default:
			ret = -ENOIOCTLCMD;
		}
		goto up_out;
	}

	/*
	 * issue TCTL
	 */
	if (interactive) {
		task->tctl_stat = -EBUSY;
		omap_dsp_mbsend_and_wait_exarg(MBCMD(TCTL), tid, cmd,
					       tid, argc, mbarg,
					       &task->ioctl_wait_q);
		ret = task->tctl_stat;
		if (ret < 0) {
			printk(KERN_ERR "omapdsp: TCTL not responding.\n");
			goto up_out;
		}
	} else {
		omap_dsp_mbsend_exarg(MBCMD(TCTL), tid, cmd, tid, argc, mbarg);
		ret = 0;
	}

up_out:
#ifdef CONFIG_OMAP_DSP_TASK_MULTIOPEN
	up(&task->ioctl_sem);
#endif
	return ret;
}

static void omap_dsp_task_vma_open(struct vm_area_struct *vma)
{
}

static void omap_dsp_task_vma_close(struct vm_area_struct *vma)
{
}

/**
 * On demand page allocation is not allowed. The mapping area is defined by 
 * corresponding DSP tasks.
 */
static struct page *omap_dsp_task_nopage_mmap(struct vm_area_struct *vma,
					      unsigned long address, int *type)
{
	return NOPAGE_SIGBUS;
}

static struct vm_operations_struct omap_dsp_task_vm_ops = {
	.open   = omap_dsp_task_vma_open,
	.close  = omap_dsp_task_vma_close,
	.nopage = omap_dsp_task_nopage_mmap,
};

static int omap_dsp_task_mmap(struct file *filp, struct vm_area_struct *vma)
{
	void *tmp_vadr;
	unsigned long tmp_padr, tmp_vmadr, off;
	size_t req_len, tmp_len;
	unsigned int minor = MINOR(filp->f_dentry->d_inode->i_rdev);
	struct dsptask *task = taskdev[minor]->task;

	if (task->map_length == 0) {
		printk(KERN_ERR
		       "omapdsp: task %s doesn't have mmap buffer.\n",
		       task->name);
		return -EINVAL;
	}
	if (is_dsp_internal_mem(task->map_base)) {
		printk(KERN_ERR
		       "omapdsp: task %s: map_base = %p\n"
		       "    DARAM/SARAM can't be used as mmap buffer.\n",
		       task->name, task->map_base);
		return -EINVAL;
	}

	/*
	 * Don't swap this area out
	 * Don't dump this area to a core file
	 */
	vma->vm_flags |= VM_RESERVED | VM_IO;

	/* Do not cache this area */
	vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);

	req_len = vma->vm_end - vma->vm_start;
	off = vma->vm_pgoff << PAGE_SHIFT;
	tmp_vmadr = vma->vm_start;
	tmp_vadr = task->map_base + off;
	do {
		tmp_padr = omap_dsp_virt_to_phys(tmp_vadr, &tmp_len);
		if (tmp_padr == 0) {
			printk(KERN_ERR
			       "omapdsp: task %s: illegal address "
			       "for mmap: %p", task->name, tmp_vadr);
			/* partial mapping will be cleared in upper layer */
			return -EINVAL;
		}
		if (tmp_len > req_len)
			tmp_len = req_len;

		printk(KERN_DEBUG
		       "omapdsp: mmap info: "
		       "vmadr = %08lx, padr = %08lx, len = %x\n",
		       tmp_vmadr, tmp_padr, tmp_len);
		if (remap_page_range(vma, tmp_vmadr, tmp_padr, tmp_len,
				     vma->vm_page_prot) != 0) {
			printk(KERN_ERR
			       "omapdsp: task %s: remap_page_range() failed.\n",
			       task->name);
			/* partial mapping will be cleared in upper layer */
			return -EAGAIN;
		}

		req_len   -= tmp_len;
		tmp_vmadr += tmp_len;
		tmp_vadr  += tmp_len;
	} while (req_len);

	vma->vm_ops = &omap_dsp_task_vm_ops;
	omap_dsp_task_vma_open(vma);
	return 0;
}

static int omap_dsp_task_open(struct inode *inode, struct file *file)
{
	unsigned int minor = MINOR(inode->i_rdev);
	struct taskdev *dev = taskdev[minor];
	int sem_st;

	if (minor >= n_task) {
		/* dynamic task */
		if ((sem_st = down_interruptible(&dev->dld_sem)) < 0)
			return sem_st;
#ifndef CONFIG_OMAP_DSP_TASK_MULTIOPEN
		if (dev->usecount > 0) {
			up(&dev->dld_sem);
			return -EBUSY;
		}
#endif

		if (dev->task == NULL) {
			/*
			 * ask for daemon to load the dsptask.
			 */
			long current_state;
			DECLARE_WAITQUEUE(wait, current);

			add_wait_queue(&dev->open_wait_q, &wait);
			current_state = current->state;
			set_current_state(TASK_INTERRUPTIBLE);
			dev->openstat = OMAP_DSP_OPENSTAT_REQ;
			/* wake up twch daemon for tadd */
			omap_dsp_twch_touch();
			if (dev->task == NULL)	/* last check */
				schedule();
			set_current_state(current_state);
			remove_wait_queue(&dev->open_wait_q, &wait);
			if (dev->task == NULL) {
				/*
				 * FIXME: error handling should be in
				 * more safe way (lock control, etc)
				 */
				printk(KERN_ERR
				       "omapdsp: task attach failed for %s!\n",
				       dev->name);
				dev->openstat = OMAP_DSP_OPENSTAT_CLOSE;
				/* wake up twch daemon */
				omap_dsp_twch_touch();
				up(&dev->dld_sem);
				return -EBUSY;
			}
		}
	} else {
		/* static task */
		preempt_disable();
#ifndef CONFIG_OMAP_DSP_TASK_MULTIOPEN
		if (dev->usecount > 0)
			return -EBUSY;
#endif
	}

	/*
	 * disable preemption while handling dev struct members
	 */
	dev->openstat = OMAP_DSP_OPENSTAT_OPEN;
	dev->usecount++;
	proc_list_add(&dev->task->proc_list, current);

	if (minor >= n_task)	/* dynamic task */
		up(&dev->dld_sem);
	else			/* static task */
		preempt_enable();

	file->f_op = &dev->fops;
	omap_dsp_map_update(current);
	omap_dsp_cur_users_add(current);
	return 0;
}

static int omap_dsp_task_release(struct inode *inode, struct file *file)
{
	unsigned int minor = MINOR(inode->i_rdev);
	struct taskdev *dev = taskdev[minor];
	struct dsptask *task = dev->task;

#ifdef CONFIG_OMAP_DSP_TASK_MULTIOPEN
	if (task->lock_pid == current->pid)
		omap_dsp_task_unlock(task);
#endif

	omap_dsp_cur_users_del(current);

	if (minor >= n_task)	/* dynamic task */
		down(&dev->dld_sem);
	else			/* static task */
		preempt_disable();

	/*
	 * disable preemption while handling dev struct members
	 */
	proc_list_del(&dev->task->proc_list, current);
	if (--dev->usecount == 0) {
		dev->openstat = OMAP_DSP_OPENSTAT_CLOSE;
		omap_dsp_task_flush_buf(task);
		if (minor >= n_task) {	/* dynamic task */
			/* wake up twch daemon for tdel */
			omap_dsp_twch_touch();
			/* do not release dld_sem */
			return 0;
		}
	}

	if (minor >= n_task)	/* dynamic task */
		up(&dev->dld_sem);
	else			/* static task */
		preempt_enable();
	return 0;
}

/*
 * mkdev / rmdev
 */
int omap_dsp_mkdev(char *name)
{
	struct taskdev *dev;
	int status;
	unsigned char minor;

	if (taskmod_cfgstat != CFGSTAT_DONE) {
		printk(KERN_ERR "omapdsp: dsp has not been configured.\n");
		return -EINVAL;
	}
	for (minor = n_task; minor < TASKDEV_MAX; minor++) {
		if (taskdev[minor] == NULL)
			goto do_make;
	}
	printk(KERN_ERR "omapdsp: Too many task devices.\n");
	return -EBUSY;

do_make:
	if ((dev = kmalloc(sizeof(struct taskdev), GFP_KERNEL)) == NULL)
		return -ENOMEM;
	memset(dev, 0, sizeof(struct taskdev));
	if ((status = omap_dsp_taskdev_init(dev, name, minor)) < 0) {
		kfree(dev);
		return status;
	}
	taskdev[minor] = dev;
	return minor;
}

int omap_dsp_rmdev(char *name)
{
	unsigned char minor;
	int ret;

	if (taskmod_cfgstat != CFGSTAT_DONE) {
		printk(KERN_ERR "omapdsp: dsp has not been configured.\n");
		return -EINVAL;
	}
	for (minor = n_task; minor < TASKDEV_MAX; minor++) {
		if (taskdev[minor] && !strcmp(taskdev[minor]->name, name)) {
			if ((ret = omap_dsp_rmdev_minor(minor)) < 0)
				return ret;
			return minor;
		}
	}
	return -EINVAL;
}

static int omap_dsp_rmdev_minor(unsigned char minor)
{
	struct taskdev *dev = taskdev[minor];
	struct dsptask *task = dev->task;

	if ((dev->usecount > 0) || waitqueue_active(&dev->open_wait_q))
		return -EBUSY;
	if (task) {
		/* FIXME: tdel? */
		omap_dsp_taskdev_detach_task(dev);
		omap_dsp_task_unconfig(task);
		kfree(task);
	}

	omap_dsp_taskdev_delete(minor);
	kfree(dev);

	return 0;
}

struct file_operations omap_dsp_task_fops = {
	.owner   = THIS_MODULE,
	.poll    = omap_dsp_task_poll,
	.ioctl   = omap_dsp_task_ioctl,
	.open    = omap_dsp_task_open,
	.release = omap_dsp_task_release,
	.mmap    = omap_dsp_task_mmap,
};

static int omap_dsp_taskdev_init(struct taskdev *dev, char *name,
				 unsigned char minor)
{
	taskdev[minor] = dev;

	strncpy(dev->name, name, OMAP_DSP_TNM_LEN);
	dev->enable = 1;
	dev->name[OMAP_DSP_TNM_LEN-1] = '\0';
	dev->openstat = OMAP_DSP_OPENSTAT_CLOSE;
	dev->usecount = 0;
	memcpy(&dev->fops, &omap_dsp_task_fops, sizeof(struct file_operations));
	devfs_mk_cdev(MKDEV(OMAP_DSP_TASK_MAJOR, minor),
		      S_IFCHR | S_IRUGO | S_IWUGO, "dsptask/%s", dev->name);
#ifdef CONFIG_PROC_FS
	omap_dsp_taskdev_create_proc(dev);
#endif

	init_waitqueue_head(&dev->open_wait_q);
	init_MUTEX(&dev->dld_sem);

	return 0;
}

static void omap_dsp_taskdev_delete(unsigned char minor)
{
	struct taskdev *dev = taskdev[minor];

	if (!dev->enable)
		return;
#ifdef CONFIG_PROC_FS
	omap_dsp_taskdev_remove_proc(dev);
#endif
	devfs_remove("dsptask/%s", dev->name);
	taskdev[minor] = NULL;
}

static void omap_dsp_taskdev_attach_task(struct taskdev *dev,
					struct dsptask *task)
{
	unsigned short ttyp = task->ttyp;

	dev->task = task;
	dev->fops.read =
		sndtyp_acv(ttyp) ?
			sndtyp_wd(ttyp) ? omap_dsp_task_read_wd_acv:
			/* sndtyp_bk */   omap_dsp_task_read_bk_acv:
		/* sndtyp_psv */
			sndtyp_wd(ttyp) ? omap_dsp_task_read_wd_psv:
			/* sndtyp_bk */   omap_dsp_task_read_bk_psv;
	dev->fops.write =
		rcvtyp_wd(ttyp) ? omap_dsp_task_write_wd:
		/* rcvbyp_bk */	  omap_dsp_task_write_bk;
	printk(KERN_INFO "omapdsp: taskdev %s enabled.\n", dev->name);
}

static void omap_dsp_taskdev_detach_task(struct taskdev *dev)
{
	if (dev->task) {
		dev->task = NULL;
		dev->fops.read = NULL;
		dev->fops.write = NULL;
		printk(KERN_INFO "omapdsp: taskdev %s disabled.\n", dev->name);
	}
}

/*
 * tadd / tdel
 */
int omap_dsp_tadd(unsigned char minor, unsigned long adr)
{
	struct taskdev *dev;
	struct dsptask *task;
	unsigned short argv[2];
	int ret = minor;

	if (minor >= TASKDEV_MAX) {
		printk(KERN_ERR
		       "omapdsp: illegal minor number (%d) for tadd\n", minor);
		return -EINVAL;
	}
	dev = taskdev[minor];
	if (!dev) {
		printk(KERN_ERR
		       "omapdsp: no task device with minor %d\n", minor);
		return -EINVAL;
	}
	if (dev->task) {
		printk(KERN_ERR
		       "omapdsp: task has already been attached "
		       "for taskdev %s\n", dev->name);
		return -EINVAL;
	}

	if (adr == OMAP_DSP_TADD_ABORTADR) {
		ret = 0;	/* FIXME: Is it ok? */
		goto done;
	}
	if (adr >= DSPSPACE_SIZE) {
		printk(KERN_ERR
		       "omapdsp: illegal address 0x%08lx for tadd\n", adr);
		ret = -EINVAL;
		goto done;
	}

	adr >>= 1;	/* word address */
	argv[0] = adr >> 16;		/* addrh */
	argv[1] = adr & 0xffff;	/* addrl */

	tadd_tid = OMAP_DSP_TID_ANON;
	tcfg_wait_cmd = MBCMD(TADD);
	omap_dsp_mbsend_and_wait_exarg(MBCMD(TADD), 0, 0,
				      OMAP_DSP_TID_ANON, 2, argv, &tcfg_wait_q);

	if (tadd_tid == OMAP_DSP_TID_ANON) {
		printk(KERN_ERR "omapdsp: tadd failed!\n");
		ret = -EINVAL;
		goto done;
	}
	if ((tadd_tid < n_task) || dsptask[tadd_tid]) {
		printk(KERN_ERR "omapdsp: illegal tid (%d)!\n", tadd_tid);
		ret = -EINVAL;
		goto done;
	}
	if ((task = kmalloc(sizeof(struct dsptask), GFP_KERNEL)) == NULL)
		return -ENOMEM;
	memset(task, 0, sizeof(struct dsptask));

	if ((ret = omap_dsp_task_config(task, tadd_tid)) < 0)
		goto done;

	if (strcmp(dev->name, task->name)) {
		printk(KERN_ERR
		       "omapdsp: task name (%s) doesn't match with "
		       "device name (%s).\n", task->name, dev->name);
		ret = -EINVAL;
		goto done;
	}

	omap_dsp_taskdev_attach_task(dev, task);

done:
	wake_up_interruptible(&dev->open_wait_q);
	return ret;
}

int omap_dsp_tdel(unsigned int minor)
{
	struct taskdev *dev;
	struct dsptask *task;
	unsigned char tid;

	if (minor >= TASKDEV_MAX) {
		printk(KERN_ERR
		       "omapdsp: illegal minor number (%d) for tdel\n", minor);
		return -EINVAL;
	}
	dev = taskdev[minor];
	if (!dev) {
		printk(KERN_ERR
		       "omapdsp: no task device with minor %d\n", minor);
		return -EINVAL;
	}
	task = dev->task;
	if (!task) {
		printk(KERN_ERR
		       "omapdsp: task has not been attached "
		       "for taskdev %s\n", dev->name);
		return -EINVAL;
	}
	if (dev->usecount > 0) {
		printk(KERN_ERR "omapdsp: taskdev %s is busy.\n", dev->name);
		return -EBUSY;
	}

	tid = task->tid;
	tdel_tid = OMAP_DSP_TID_ANON;
	tcfg_wait_cmd = MBCMD(TDEL);
	omap_dsp_mbsend_and_wait(MBCMD(TDEL), tid, 0, &tcfg_wait_q);

	omap_dsp_taskdev_detach_task(dev);
	omap_dsp_task_unconfig(task);
	kfree(task);
	up(&dev->dld_sem);

	if (tdel_tid != tid) {
		printk(KERN_ERR "omapdsp: tdel failed!\n");
		return -EINVAL;
	}

	return minor;
}

/*
 * openstat inquiry
 */
int omap_dsp_taskdev_openstat(unsigned char minor)
{
	return taskdev[minor] ? taskdev[minor]->openstat :
				OMAP_DSP_OPENSTAT_CLOSE;
}

/*
 * functions called from mailbox1 interrupt routine
 */

void mailbox1_wdsnd(unsigned char tid, unsigned short data)
{
	struct dsptask *task = dsptask[tid];

	if ((tid >= TASKDEV_MAX) || (task == NULL)) {
		printk(KERN_ERR "mailbox: WDSND with illegal tid! %d\n", tid);
		return;
	}
	if (sndtyp_bk(task->ttyp)) {
		printk(KERN_ERR
		       "mailbox: WDSND from block sending task! "
		       "(task%d)\n", tid);
		return;
	}
	if (sndtyp_psv(task->ttyp) && !waitqueue_active(&task->read_wait_q)) {
		printk(KERN_WARNING
		       "mailbox: WDSND from passive sending task "
		       "(task%d) without request!\n", tid);
		return;
	}

	write_word_to_fifo(&task->rcvdt.fifo, data);
	wake_up_interruptible(&task->read_wait_q);
}

void mailbox1_wdreq(unsigned char tid)
{
	struct dsptask *task = dsptask[tid];

	if ((tid >= TASKDEV_MAX) || (task == NULL)) {
		printk(KERN_ERR "mailbox: WDREQ with illegal tid! %d\n", tid);
		return;
	}
	if (rcvtyp_psv(task->ttyp)) {
		printk(KERN_ERR
		       "mailbox: WDREQ from passive receiving task! "
		       "(task%d)\n", tid);
		return;
	}

	task->wsz = 2;
	wake_up_interruptible(&task->write_wait_q);
}

void mailbox1_bksnd(unsigned char tid, unsigned short bid)
{
	struct dsptask *task = dsptask[tid];
	unsigned short cnt;

	if (bid >= ipbcfg.ln) {
		printk(KERN_ERR "mailbox: BKSND with illegal bid! %d\n", bid);
		return;
	}
	ipb_bsycnt_dec(&ipbcfg);
	if ((tid >= TASKDEV_MAX) || (task == NULL)) {
		printk(KERN_ERR "mailbox: BKSND with illegal tid! %d\n", tid);
		unuse_ipbuf(bid, MBSENDTYPE_NOWAIT);
		return;
	}
	if (sndtyp_wd(task->ttyp)) {
		printk(KERN_ERR
		       "mailbox: BKSND from word sending task! "
		       "(task%d)\n", tid);
		unuse_ipbuf(bid, MBSENDTYPE_NOWAIT);
		return;
	}
	if (sndtyp_pvt(task->ttyp)) {
		printk(KERN_ERR
		       "mailbox: BKSND from private sending task! "
		       "(task%d)\n", tid);
		unuse_ipbuf(bid, MBSENDTYPE_NOWAIT);
		return;
	}
	if (sync_with_dsp(&ipbuf[bid]->sd, tid, 10) < 0) {
		printk(KERN_ERR "mailbox: BKSND - IPBUF sync failed!\n");
		return;
	}

	/* should be done in DSP, but just in case. */
	ipbuf[bid]->next = OMAP_DSP_BID_NULL;

	cnt = ipbuf[bid]->c;
	if (cnt > ipbcfg.lsz) {
		printk(KERN_ERR
		       "mailbox: BKSND cnt(%d) > ipbuf line size(%d)!\n",
		       cnt, ipbcfg.lsz);
		unuse_ipbuf(bid, MBSENDTYPE_NOWAIT);
		return;
	}

	if (cnt == 0) {
		/* 0-byte send from DSP */
		unuse_ipbuf(bid, MBSENDTYPE_NOWAIT);
		goto done;
	}
	ipblink_add_tail(&task->rcvdt.bk.link, bid, ipbuf);
	/* we keep coming bid and return alternative line to DSP. */
	balance_ipbuf(MBSENDTYPE_NOWAIT);

done:
	wake_up_interruptible(&task->read_wait_q);
}

void mailbox1_bkreq(unsigned char tid, unsigned short cnt)
{
	struct dsptask *task = dsptask[tid];

	if ((tid >= TASKDEV_MAX) || (task == NULL)) {
		printk(KERN_ERR "mailbox: BKREQ with illegal tid! %d\n", tid);
		return;
	}
	if (rcvtyp_wd(task->ttyp)) {
		printk(KERN_ERR
		       "mailbox: BKREQ from word receiving task! "
		       "(task%d)\n", tid);
		return;
	}
	if (rcvtyp_pvt(task->ttyp)) {
		printk(KERN_ERR
		       "mailbox: BKREQ from private receiving task! "
		       "(task%d)\n", tid);
		return;
	}
	if (rcvtyp_psv(task->ttyp)) {
		printk(KERN_ERR
		       "mailbox: BKREQ from passive receiving task! "
		       "(task%d)\n", tid);
		return;
	}

	task->wsz = cnt*2;
	wake_up_interruptible(&task->write_wait_q);
}

void mailbox1_bkyld(unsigned short bid)
{
	if (bid >= ipbcfg.ln) {
		printk(KERN_ERR "mailbox: BKYLD with illegal bid! %d\n", bid);
		return;
	}

	/* should be done in DSP, but just in case. */
	ipbuf[bid]->next = OMAP_DSP_BID_NULL;

	/* we don't need to sync with DSP */
	ipb_bsycnt_dec(&ipbcfg);
	release_ipbuf(bid);
}

void mailbox1_bksndp(unsigned char tid)
{
	struct dsptask *task = dsptask[tid];
	struct rcvdt_bk_struct *rcvdt = &task->rcvdt.bk;
	struct ipbuf_p *ipbp = rcvdt->ipbuf_pvt_r;

	if ((tid >= TASKDEV_MAX) || (task == NULL)) {
		printk(KERN_ERR "mailbox: BKSNDP with illegal tid! %d\n", tid);
		return;
	}
	if (sndtyp_wd(task->ttyp)) {
		printk(KERN_ERR
		       "mailbox: BKSNDP from word sending task! "
		       "(task%d)\n", tid);
		return;
	}
	if (sndtyp_gbl(task->ttyp)) {
		printk(KERN_ERR
		       "mailbox: BKSNDP from non-private sending task! "
		       "(task%d)\n", tid);
		return;
	}

	/*
	 * we should not have delayed block at this point
	 * because read() routine releases the lock of the buffer and
	 * until then DSP can't send next data.
	 */

	if (sync_with_dsp(&ipbp->s, tid, 10) < 0) {
		printk(KERN_ERR "mailbox: BKSNDP - IPBUF sync failed!\n");
		return;
	}
	printk(KERN_DEBUG
	       "mailbox: ipbuf_pvt_r->a = 0x%08lx\n",
	       MKLONG(ipbp->ah, ipbp->al));
	ipblink_add_pvt(&rcvdt->link);
	wake_up_interruptible(&task->read_wait_q);
}

void mailbox1_bkreqp(unsigned char tid)
{
	struct dsptask *task = dsptask[tid];
	struct ipbuf_p *ipbp = task->ipbuf_pvt_w;

	if ((tid >= TASKDEV_MAX) || (task == NULL)) {
		printk(KERN_ERR "mailbox: BKREQP with illegal tid! %d\n", tid);
		return;
	}
	if (rcvtyp_wd(task->ttyp)) {
		printk(KERN_ERR
		       "mailbox: BKREQP from word receiving task! "
		       "(task%d)\n", tid);
		return;
	}
	if (rcvtyp_gbl(task->ttyp)) {
		printk(KERN_ERR
		       "mailbox: BKREQP from non-private receiving task! "
		       "(task%d)\n", tid);
		return;
	}
	if (rcvtyp_psv(task->ttyp)) {
		printk(KERN_ERR
		       "mailbox: BKREQP from passive receiving task! "
		       "(task%d)\n", tid);
		return;
	}

	if (sync_with_dsp(&ipbp->s, OMAP_DSP_TID_FREE, 10) < 0) {
		printk(KERN_ERR "mailbox: BKREQP - IPBUF sync failed!\n");
		return;
	}
	printk(KERN_DEBUG
	       "mailbox: ipbuf_pvt_w->a = 0x%08lx\n",
	       MKLONG(ipbp->ah, ipbp->al));
	task->wsz = ipbp->c*2;
	wake_up_interruptible(&task->write_wait_q);
}

void mailbox1_tctl(unsigned char tid, unsigned short data)
{
	struct dsptask *task = dsptask[tid];

	if ((tid >= TASKDEV_MAX) || (task == NULL)) {
		printk(KERN_ERR "mailbox: TCTL with illegal tid! %d\n", tid);
		return;
	}

	if (!waitqueue_active(&task->ioctl_wait_q)) {
		printk(KERN_WARNING "mailbox: unexpected TCTL from DSP!\n");
		return;
	}

	task->tctl_stat = data;
	wake_up_interruptible(&task->ioctl_wait_q);
}

void mailbox1_tcfg(unsigned char tid, unsigned short data)
{
	struct dsptask *task = dsptask[tid];
	unsigned long tmp_ipbp_r, tmp_ipbp_w;
	unsigned long tmp_mapstart, tmp_maplen;
	unsigned long tmp_tnm;
	unsigned short *tnm;
	volatile unsigned short *buf;
	int i;

	if ((tid >= TASKDEV_MAX) || (task == NULL)) {
		printk(KERN_ERR "mailbox: TCFG with illegal tid! %d\n", tid);
		return;
	}
	if ((task->cfgstat != CFGSTAT_GOING) ||
	    tcfg_wait_cmd != MBCMD(TCFG)) {
		printk(KERN_WARNING "mailbox: unexpected TCFG from DSP!\n");
		return;
	}

	if (sync_with_dsp(&ipbuf_sys_da->s, tid, 10) < 0) {
		printk(KERN_ERR "mailbox: TCFG - IPBUF sync failed!\n");
		return;
	}

	/*
	 * read configuration data on system IPBUF
	 */
	buf = ipbuf_sys_da->d;
	task->ttyp   = buf[0];
	tmp_ipbp_r   = MKLONG(buf[1], buf[2]);
	tmp_ipbp_w   = MKLONG(buf[3], buf[4]);
	tmp_mapstart = MKLONG(buf[5], buf[6]);
	tmp_maplen   = MKLONG(buf[7], buf[8]);
	tmp_tnm      = MKLONG(buf[9], buf[10]);

	task->rcvdt.bk.ipbuf_pvt_r = dspword_to_virt(tmp_ipbp_r);
	task->ipbuf_pvt_w          = dspword_to_virt(tmp_ipbp_w);
	task->map_base   = dspword_to_virt(tmp_mapstart);
	task->map_length = tmp_maplen << 1;	/* word -> byte */
	tnm = dspword_to_virt(tmp_tnm);
	for (i = 0; i < OMAP_DSP_TNM_LEN-1; i++) {
		/* avoiding byte access */
		unsigned short tmp = tnm[i];
		task->name[i] = tmp & 0x00ff;
		if (!tmp)
			break;
	}
	task->name[OMAP_DSP_TNM_LEN-1] = '\0';

	release_ipbuf_pvt(ipbuf_sys_da);
	task->cfgstat = CFGSTAT_DONE;
	wake_up_interruptible(&tcfg_wait_q);
}

void mailbox1_tadd(unsigned char tid)
{
	if (!waitqueue_active(&tcfg_wait_q) ||
	    tcfg_wait_cmd != MBCMD(TADD)) {
		printk(KERN_WARNING "mailbox: unexpected TADD from DSP!\n");
		return;
	}
	tadd_tid = tid;
	wake_up_interruptible(&tcfg_wait_q);
}

void mailbox1_tdel(unsigned char tid)
{
	if (!waitqueue_active(&tcfg_wait_q) ||
	    tcfg_wait_cmd != MBCMD(TDEL)) {
		printk(KERN_WARNING "mailbox: unexpected TDEL from DSP!\n");
		return;
	}
	tdel_tid = tid;
	wake_up_interruptible(&tcfg_wait_q);
}

void mailbox1_err_fatal(unsigned char tid)
{
	struct dsptask *task = dsptask[tid];
	struct list_head *ptr;
	siginfo_t info;

	if ((tid >= TASKDEV_MAX) || (task == NULL)) {
		printk(KERN_ERR
		       "mailbox: FATAL ERR with illegal tid! %d\n", tid);
		return;
	}

	info.si_signo = SIGBUS;
	info.si_errno = 0;
	info.si_code = SI_KERNEL;
	info._sifields._sigfault._addr = NULL;
	list_for_each(ptr, &task->proc_list) {
		send_sig_info(SIGBUS, &info, ((struct proc_list *)ptr)->tsk);
	}
}

void mailbox1_dbg(unsigned char tid)
{
	char s[64];
	unsigned short *src;
	volatile unsigned short *buf;
	int cnt;
	int i, j;

	if (((tid >= TASKDEV_MAX) || (dsptask[tid] == NULL)) &&
	    (tid != OMAP_DSP_TID_ANON)) {
		printk(KERN_ERR "mailbox: DBG with illegal tid! %d\n", tid);
		return;
	}
	if (sync_with_dsp(&ipbuf_sys_da->s, tid, 10) < 0) {
		printk(KERN_ERR "mailbox: DBG - IPBUF sync failed!\n");
		return;
	}
	buf = ipbuf_sys_da->d;
	cnt = buf[0];
	src = dspword_to_virt(MKLONG(buf[1], buf[2]));
	j = 0;
	for (i = 0; i < cnt; i++) {
		unsigned short tmp;
		/*
		 * Be carefull that ipbuf should not be read with
		 * 1-byte access since it might be placed in DARAM/SARAM
		 * and it can cause unexpected byteswap.
		 * For example,
		 *   s[j++] = *(src++) & 0xff;
		 * causes 1-byte access!
		 */
		tmp = *src++;
		s[j++] = tmp & 0xff;
		if (s[j-1] == '\n') {
			s[j] = '\0';
			printk(KERN_INFO "%s", s);
			j = 0;
			continue;
		}
		if (j == 63) {
			s[j] = '\0';
			printk(KERN_INFO "%s\n", s);
			j = 0;
			continue;
		}
	}
	if (j > 0) {
		s[j] = '\0';
		printk(KERN_INFO "%s\n", s);
	}

	release_ipbuf_pvt(ipbuf_sys_da);
}


#ifdef CONFIG_PROC_FS
/*
 * proc file entry
 */

#define openstat_name(stat) (\
	((stat) == OMAP_DSP_OPENSTAT_CLOSE) ? "CLOSE" :\
	((stat) == OMAP_DSP_OPENSTAT_OPEN)  ? "OPEN" :\
	((stat) == OMAP_DSP_OPENSTAT_REQ)   ? "REQ" :\
					      "unknown")

#define cfgstat_name(stat) (\
	((stat) == CFGSTAT_ERR)   ? "ERR" :\
	((stat) == CFGSTAT_GOING) ? "GOING" :\
	((stat) == CFGSTAT_DONE)  ? "DONE" :\
	((stat) == CFGSTAT_ABORT) ? "ABORT" :\
				    "unknown")

static inline char *bid_name(unsigned short bid)
{
	static char s[6];

	switch (bid) {
	case OMAP_DSP_BID_NULL:
		return "NULL";
	case OMAP_DSP_BID_PVT:
		return "PRIVATE";
	default:
		sprintf(s, "%d", bid);
		return s;
	}
}

static inline char *sndtyp_name(unsigned short ttyp)
{
	static char s[32];

	sprintf(s, "%s %s send",
		(sndtyp_acv(ttyp)) ? "active" :
				     "passive",
		(sndtyp_wd(ttyp))  ? "word" :
		(sndtyp_pvt(ttyp)) ? "private block" :
				     "global block");
	return s;
}

static inline char *rcvtyp_name(unsigned short ttyp)
{
	static char s[32];

	sprintf(s, "%s %s receive",
		(rcvtyp_acv(ttyp)) ? "active" :
				     "passive",
		(rcvtyp_wd(ttyp))  ? "word" :
		(rcvtyp_pvt(ttyp)) ? "private block" :
				     "global block");
	return s;
}

static int omap_dsp_proc_taskdev_stat_read(char *page, char **start,
					   off_t off, int count,
					   int *eof, void *data)
{
	struct taskdev *dev;
	struct dsptask *task;
	struct list_head *ptr;
	char *out;
	int len;

	preempt_disable();

	dev = (struct taskdev *)data;
	task = dev->task;

	out = page;
	out += sprintf(out,
		       "device name: %s\n"
		       "openstat     %s\n",
		       dev->name, openstat_name(dev->openstat));

	if (task) {
		unsigned short ttyp = task->ttyp;

		out += sprintf(out,
			       "  attached task: %s\n"
			       "    cfgstat    %s\n"
			       "    tid        %d\n"
			       "    ttyp       0x%04x (%s, %s)\n",
			       task->name,
			       cfgstat_name(task->cfgstat),
			       task->tid,
			       ttyp, sndtyp_name(ttyp), rcvtyp_name(ttyp));

		if (sndtyp_wd(ttyp)) {
			struct fifo_struct *fifo = &task->rcvdt.fifo;
			out += sprintf(out,
				       "    fifo->buf     0x%p\n"
				       "    fifo->sz      %d\n"
				       "    fifo->cnt     %d\n"
				       "    fifo->wp      %d\n",
				       fifo->buf, fifo->sz,
				       fifo->cnt, fifo->wp);
		} else {	/* sndtyp_bk */
			struct rcvdt_bk_struct *rcvdt = &task->rcvdt.bk;
			out += sprintf(out,
				       "    delayed list top   %s\n"
				       "                 tail  %s\n",
				       bid_name(rcvdt->link.top),
				       bid_name(rcvdt->link.tail));
			if (sndtyp_pvt(ttyp)) {
				out += sprintf(out,
					       "    private IPBUF for read   0x%p\n",
					       rcvdt->ipbuf_pvt_r);
			}
		}

		out += sprintf(out, "    wsz        %d\n", task->wsz);
		if (rcvtyp_pvt(ttyp)) {
			out += sprintf(out,
				       "    private IPBUF for write  0x%p\n",
				       task->ipbuf_pvt_w);
		}

		if (task->map_length) {
			out += sprintf(out,
				       "    map_base   %p\n"
				       "    map_length %d\n",
				       task->map_base, task->map_length);
		}

		out += sprintf(out, "    user pid ");
		list_for_each(ptr, &task->proc_list) {
			out += sprintf(out, "  %d",
				       ((struct proc_list *)ptr)->tsk->pid);
		}
		out += sprintf(out, "\n");
	} else {
		out += sprintf(out, "task not attached.\n");
	}

	preempt_enable();

	len = out - page;
	len -= off;
	if (len < count) {
		*eof = 1;
		if (len <= 0)
			return 0;
	} else {
		len = count;
	}
	*start = page + off;

	return len;
}

static int omap_dsp_proc_taskdev_fifosz_read(char *page, char **start,
					     off_t off, int count,
					     int *eof, void *data)
{
	struct taskdev *dev;
	struct dsptask *task;
	size_t fifosz;
	int len;

	dev = (struct taskdev *)data;
	task = dev->task;
	fifosz = (task && sndtyp_wd(task->ttyp)) ? task->rcvdt.fifo.sz : 0;
	len = sprintf(page, "%d\n", fifosz);

	return len;
}

static int omap_dsp_proc_taskdev_fifosz_write(struct file *file,
					      const char *buffer,
					      unsigned long count, void *data)
{
	struct taskdev *dev;
	struct dsptask *task;
	int len;
	char tmp[16];
	unsigned long fifosz;
	int ret;

	dev = (struct taskdev *)data;
	task = dev->task;

	if (task == NULL) {
		printk(KERN_ERR "task not attached.\n");
		return -EINVAL;
	}

	len = (count > 15) ? 15 : count;
	if (copy_from_user(tmp, buffer, len))
		return -EFAULT;
	tmp[len] = '\0';

	fifosz = simple_strtol(tmp, NULL, 10);
	if ((ret = omap_dsp_task_set_fifosz(task, fifosz)) < 0)
		return ret;

	return len;
}

/*
 * called from omap_dsp_proc_ipbuf_read()
 */
int omap_dsp_ipbuf_held(unsigned char tid, unsigned short bid)
{
	struct dsptask *task = dsptask[tid];
	unsigned short b;
	int ret = 0;

	if (task == NULL)
		return 0;

	disable_irq(INT_D2A_MB1);
	ipblink_for_each(b, &task->rcvdt.bk.link, ipbuf) {
		if (b == bid) {	/* found */
			ret = 1;
			break;
		}
	}
	enable_irq(INT_D2A_MB1);

	return ret;
}

static void omap_dsp_taskdev_create_proc(struct taskdev *dev)
{
	struct proc_dir_entry *ent;

	if (procdir_dsptask == NULL)
		return;

	dev->procent_dir = proc_mkdir(dev->name, procdir_dsptask);
	if (dev->procent_dir == NULL) {
		printk(KERN_ERR
		       "omapdsp: failed to register proc directory: "
		       "%s\n", dev->name);
		return;
	}

	ent = create_proc_read_entry("status", 0, dev->procent_dir,
				     omap_dsp_proc_taskdev_stat_read, dev);
	if (ent == NULL) {
		printk(KERN_ERR
		       "omapdsp: failed to register "
		       "status proc device for: %s\n", dev->name);
		return;
	}

	ent = create_proc_entry("fifosz", S_IFREG | S_IWUGO | S_IRUGO,
				dev->procent_dir);
	if (ent == NULL) {
		printk(KERN_ERR
		       "omapdsp: failed to register fifosz "
		       "proc device for: %s\n", dev->name);
		return;
	}
	ent->read_proc  = omap_dsp_proc_taskdev_fifosz_read;
	ent->write_proc = omap_dsp_proc_taskdev_fifosz_write;
	ent->data = dev;
}

static void omap_dsp_taskdev_remove_proc(struct taskdev *dev)
{
	if (dev->procent_dir == NULL)
		return;

	remove_proc_entry("status", dev->procent_dir);
	remove_proc_entry("fifosz", dev->procent_dir);
	remove_proc_entry(dev->name, procdir_dsptask);
	dev->procent_dir = NULL;
}

void __init omap_dsp_taskdev_create_procdir(void)
{
	procdir_dsptask = proc_mkdir("task", procdir_dsp);
	if (procdir_dsptask == NULL) {
		printk(KERN_ERR
		       "omapdsp: failed to register proc device: "
		       "dsp/task\n");
	}
}

void omap_dsp_taskdev_remove_procdir(void)
{
	if (procdir_dsptask == NULL)
		return;

	remove_proc_entry("task", procdir_dsp);
	procdir_dsptask = NULL;
}

#endif /* CONFIG_PROC_FS */

int __init omap_dsp_taskmod_init(void)
{
	int retval;

	taskmod_cfgstat = CFGSTAT_ERR;
	init_waitqueue_head(&tcfg_wait_q);
	memset(taskdev, 0, sizeof(void *) * TASKDEV_MAX);
	memset(dsptask, 0, sizeof(void *) * TASKDEV_MAX);

	retval = register_chrdev(OMAP_DSP_TASK_MAJOR, "dsptask",
				 &omap_dsp_task_fops);
	if (retval < 0) {
		printk(KERN_ERR
		       "omapdsp: failed to register task device: %d\n", retval);
		return retval;
	}
	devfs_mk_dir("dsptask");
#ifdef CONFIG_PROC_FS
	omap_dsp_taskdev_create_procdir();
#endif

	return 0;
}

void omap_dsp_taskmod_exit(void)
{
	devfs_remove("dsptask");
	unregister_chrdev(OMAP_DSP_TASK_MAJOR, "dsptask");
#ifdef CONFIG_PROC_FS
	omap_dsp_taskdev_remove_procdir();
#endif
}
