/*
 * linux/arch/arm/mach-omap/pm.c
 *
 * OMAP Power Management Routines
 *
 * Original code for the SA11x0:
 * Copyright (c) 2001 Cliff Brake <cbrake@accelent.com>
 *
 * Modified for the PXA250 by Nicolas Pitre:
 * Copyright (c) 2002 Monta Vista Software, Inc.
 *
 * Modified for the OMAP1510 by David Singleton:
 * Copyright (c) 2002 Monta Vista Software, Inc.
 *
 * Cleanup 2004 for OMAP1510/1610 by Dirk Behme <dirk.behme@de.bosch.com>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN
 * NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/config.h>
#include <linux/init.h>
#include <linux/pm.h>
#include <linux/slab.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/proc_fs.h>
#include <linux/sysctl.h>
#include <linux/errno.h>
#include <linux/serial_reg.h>
#include <linux/delay.h>
#include <linux/pm.h>

#include <asm/memory.h>
#include <asm/system.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/serial.h>
#include <asm/mach-types.h>
#include <asm/arch/hardware.h>
#include <asm/arch/omap16xx.h>
#include <asm/arch/pm.h>
#include <asm/arch/mux.h>
#include <asm/arch/tps65010.h>

#include "clock.h"
#include <linux/rtc.h>
#include <linux/time.h>

#include <linux/kmsg.h>
#include "../../../drivers/char/omap-rtc.h"

#define WIFI_PHONE_POWER_KEY
#define WIFI_PHONE_PM		//ron_lee 2005-11-08
#define WIFI_PHONE_LCD
#define WIFI_PHONE_RTC




#define DEBUG_PM
#ifdef DEBUG_PM
#define info(format, arg...)	printk(KERN_INFO "pm" ": " format "\n",  ## arg ) 
#define dbg(format, arg...) 	printk(KERN_DEBUG "pm" ": " format "\n",  ## arg ) 
#else
#define info(format, arg...)
#define dbg(format, arg...)
#endif	//DEBUG_PM

#ifdef WIFI_PHONE_POWER_KEY
//added by ron lee 2005-10-31
#include <asm/arch/irqs.h>
#include <asm/arch/gpio.h>
#endif

#ifdef WIFI_PHONE_PM
#include <asm/arch/usb.h>
#include <linux/suspend.h>
#include "clock.h"
#include "../../../drivers/char/watchdog/omap_wdt_hardware.h"


#endif

#ifdef WIFI_PHONE_LCD
#if defined(VM1185T_R01) || defined(VD1185T_R01)
void set_lcd_brightness(int);
#endif
void set_lcd_backlight(int);
#endif

static unsigned int arm_sleep_save[ARM_SLEEP_SAVE_SIZE];
static unsigned short ulpd_sleep_save[ULPD_SLEEP_SAVE_SIZE];
static unsigned int mpui1510_sleep_save[MPUI1510_SLEEP_SAVE_SIZE];
static unsigned int mpui1610_sleep_save[MPUI1610_SLEEP_SAVE_SIZE];

#ifdef WIFI_PHONE_PMxx
static unsigned short dsp_sleep_save[DSP_SLEEP_SAVE_SIZE];
#endif

#ifdef WIFI_PHONE_RTC
static struct timespec rtc_delta; 
#endif

/*
 * Let's power down on idle, but only if we are really
 * idle, because once we start down the path of
 * going idle we continue to do idle even if we get
 * a clock tick interrupt . .
 */
void omap_pm_idle(void)
{
	int (*func_ptr)(void) = 0;
	unsigned int mask32 = 0;

	/*
	 * If the DSP is being used let's just idle the CPU, the overhead
	 * to wake up from Big Sleep is big, milliseconds versus micro
	 * seconds for wait for interrupt.
	 */

	cli();
	clf();
	if (need_resched()) {
		stf();
		sti();
		return;
	}
	mask32 = omap_readl(ARM_SYSST);
	stf();
	sti();
	if ((mask32 & DSP_IDLE) == 0) {
		__asm__ volatile ("mcr	p15, 0, r0, c7, c0, 4");
	} else {

		if (cpu_is_omap1510()) {
			func_ptr = (void *)(OMAP1510_SRAM_IDLE_SUSPEND);
		} else if (cpu_is_omap1610() || cpu_is_omap1710()) {
			func_ptr = (void *)(OMAP1610_SRAM_IDLE_SUSPEND);
		} else if (cpu_is_omap5912()) {
			func_ptr = (void *)(OMAP5912_SRAM_IDLE_SUSPEND);
		}

		func_ptr();
	}
}
/*
 * Configuration of the wakeup event is board specific. For the
 * moment we put it into this helper function. Later it may move
 * to board specific files.
 */
/*
 * Configuration of the wakeup event is board specific. For the
 * moment we put it into this helper function. Later it may move
 * to board specific files.
 */
static void omap_pm_wakeup_setup(void)
{
	/*
	 * Enable ARM XOR clock and release peripheral from reset by
	 * writing 1 to PER_EN bit in ARM_RSTCT2, this is required
	 * for UART configuration to use UART2 to wake up.
	 */
	//added by ron lee 2005-10-31 
	u32 level1_wake = OMAP_IRQ_BIT(INT_IH2_IRQ) | OMAP_IRQ_BIT(INT_GPIO_BANK1);	//GPIO0
	u32 level2_wake = OMAP_IRQ_BIT(INT_UART2) | OMAP_IRQ_BIT(INT_KEYBOARD) | OMAP_IRQ_BIT(INT_RTC_ALARM);

	omap_writel(omap_readl(ARM_IDLECT2) | ENABLE_XORCLK, ARM_IDLECT2);
	omap_writel(omap_readl(ARM_RSTCT2) | PER_EN, ARM_RSTCT2);
	omap_writew(MODEM_32K_EN, ULPD_CLOCK_CTRL);

	/*
	 * Turn off all interrupts except L1-2nd level cascade,
	 * and the L2 wakeup interrupts: keypad and UART2.
	 */
	
	 /* Turn off all interrupts except GPIO bank 1, L1-2nd level cascade,
	 * and the L2 wakeup interrupts: keypad and UART2. Note that the
	 * drivers must still separately call omap_set_gpio_wakeup() to
	 * wake up to a GPIO interrupt.*/
	 //added by ron lee 2005-10-31
	if (cpu_is_omap1510() || cpu_is_omap16xx())
		level1_wake |= OMAP_IRQ_BIT(INT_GPIO_BANK1);
	//omap_set_gpio_wakeup(9, 1);		//Enable GPIO 9 wakeup
	
	//omap_writel(~IRQ_LEVEL2, OMAP_IH1_MIR);	//ron lee 2005-10-31
	//ron lee 2005-10-31
	omap_writel(~level1_wake, OMAP_IH1_MIR);
	
	if (cpu_is_omap1510()) {
		omap_writel(~(IRQ_UART2 | IRQ_KEYBOARD),  OMAP_IH2_MIR);
	}

	if (cpu_is_omap16xx()) {
		//omap_writel(~(IRQ_UART2 | IRQ_KEYBOARD), OMAP_IH2_0_MIR);		//ron lee 2005-10-31
		//ron lee: GPIO 40 Wakeup: (2005-10-31)
		//omap_cfg_reg(W15_1710_GPIO40);
		omap_set_gpio_wakeup(40, 1);	//Enable GPIO 40 wakeup (Power Key)
		/*don't use GPIO31 as wifi host wakeup */
		//omap_set_gpio_wakeup(31, 1);	//Enable GPIO31 wakeup (WIFI wakeup host)
		omap_set_gpio_wakeup(11, 1);	//Enable SPI Interrupt wakeup
		omap_set_gpio_wakeup(0, 1);		//Enable USB VBus wakeup (USB adapter wakeup)
		omap_writel(~level2_wake, OMAP_IH2_0_MIR);	//enable UART2 & Keypad Wakeup
		
		level2_wake = OMAP_IRQ_BIT(INT_1610_GPIO_WAKEUP) | OMAP_IRQ_BIT(INT_1610_GPIO_BANK2)  | OMAP_IRQ_BIT(INT_1610_GPIO_BANK3);	//enable IRQ 46: Peripheral Wakeup
		omap_writel(~level2_wake, OMAP_IH2_1_MIR);
		
		//omap_writel(~0x0, OMAP_IH2_1_MIR);
		omap_writel(~0x0, OMAP_IH2_2_MIR);
		omap_writel(~0x0, OMAP_IH2_3_MIR);
	}

	/*  New IRQ agreement */
 	omap_writel(1, OMAP_IH1_CONTROL);
	omap_writel(1, OMAP_IH2_CONTROL);

	/* external PULL to down, bit 22 = 0 */
	omap_writel(omap_readl(PULL_DWN_CTRL_2) & ~(1<<22), PULL_DWN_CTRL_2);
}

#define SYSDMA_LCD_CTRL      0xfffee3c4
#define OTG_SYSCON_2		0xfffb0408
#define SAVE 0
#define RESTORE 1

//static unsigned short save_dsp_idlect1;
//static unsigned short save_dsp_idlect2;
//static unsigned short save_dsp_ckctl;
static void save_chip_context(int command)
{
	if(command == SAVE) {
		info("save_chip_context(SAVE)");
		MPUI1610_SAVE(OMAP_IH1_MIR);
		MPUI1610_SAVE(OMAP_IH2_0_MIR);
		MPUI1610_SAVE(OMAP_IH2_1_MIR);
		MPUI1610_SAVE(OMAP_IH2_2_MIR);
		MPUI1610_SAVE(OMAP_IH2_3_MIR);
		MPUI1610_SAVE(MPUI_CTRL);
		MPUI1610_SAVE(MPUI_DSP_BOOT_CONFIG);
		MPUI1610_SAVE(MPUI_DSP_API_CONFIG);
		MPUI1610_SAVE(EMIFS_CONFIG);
		MPUI1610_SAVE(EMIFF_SDRAM_CONFIG);
		MPUI1610_SAVE(SYSDMA_LCD_CTRL);
		MPUI1610_SAVE(OTG_SYSCON_2);

		info("save arm registers");
		ARM_SAVE(ARM_CKCTL);
		ARM_SAVE(ARM_IDLECT1);
		ARM_SAVE(ARM_IDLECT2);
		ARM_SAVE(ARM_IDLECT3);
		ARM_SAVE(ARM_EWUPCT);
		ARM_SAVE(ARM_RSTCT1);
		ARM_SAVE(ARM_RSTCT2);
		ARM_SAVE(ARM_SYSST);
		ARM_SAVE(DPLL_CTL);

		info("save dsp registers");
		//save_dsp_idlect1 = __raw_readw(DSP_IDLECT1);
		//save_dsp_idlect2 = __raw_readw(DSP_IDLECT2);
		//save_dsp_ckctl = __raw_readw(DSP_CKCTL);
		//DSP_SAVE(DSP_IDLECT1);
		//DSP_SAVE(DSP_IDLECT2);
		//DSP_SAVE(DSP_CKCTL);

		info("save ulpd registers");
		ULPD_SAVE(ULPD_SOFT_DISABLE_REQ);
		ULPD_SAVE(ULPD_SOFT_REQ);
		ULPD_SAVE(ULPD_CLOCK_CTRL);
		ULPD_SAVE(ULPD_STATUS_REQ);
		
	} else {
		info("save_chip_context(RESTORE)");
		MPUI1610_RESTORE(OMAP_IH1_MIR);
		MPUI1610_RESTORE(OMAP_IH2_0_MIR);
		MPUI1610_RESTORE(OMAP_IH2_1_MIR);
		MPUI1610_RESTORE(OMAP_IH2_2_MIR);
		MPUI1610_RESTORE(OMAP_IH2_3_MIR);
		MPUI1610_RESTORE(MPUI_CTRL);
		MPUI1610_RESTORE(MPUI_DSP_BOOT_CONFIG);
		MPUI1610_RESTORE(MPUI_DSP_API_CONFIG);
		MPUI1610_RESTORE(EMIFS_CONFIG);
		MPUI1610_RESTORE(EMIFF_SDRAM_CONFIG);
		MPUI1610_RESTORE(SYSDMA_LCD_CTRL);
		MPUI1610_RESTORE(OTG_SYSCON_2);

		info("restore arm registers");
		ARM_RESTORE(ARM_CKCTL);
		ARM_RESTORE(ARM_IDLECT1);
		ARM_RESTORE(ARM_IDLECT2);
		ARM_RESTORE(ARM_IDLECT3);
		ARM_RESTORE(ARM_EWUPCT);
		ARM_RESTORE(ARM_RSTCT1);
		ARM_RESTORE(ARM_RSTCT2);
		ARM_RESTORE(ARM_SYSST);
		ARM_RESTORE(DPLL_CTL);

		info("restore dsp registers");
		//__raw_writew(save_dsp_idlect1, DSP_IDLECT1);
		//__raw_writew(save_dsp_idlect2, DSP_IDLECT2);
		//__raw_writew(save_dsp_ckctl, DSP_CKCTL);
		//DSP_RESTORE(DSP_IDLECT1);
		//DSP_RESTORE(DSP_IDLECT2);
		//DSP_RESTORE(DSP_CKCTL);

		info("restore ulpd registers");
		ULPD_RESTORE(ULPD_SOFT_DISABLE_REQ);
		ULPD_RESTORE(ULPD_SOFT_REQ);
		ULPD_RESTORE(ULPD_CLOCK_CTRL);
		ULPD_RESTORE(ULPD_STATUS_REQ);

	}

}

static void shutdown_DSP()
{
	//unsigned short save_dsp_idlect2;
	
	info("shutdown_DSP");
	//__raw_writew(__raw_readw(DSP_IDLECT1) | 0x0107, DSP_IDLECT1);	//IDLTIM_DSP | IDLPER_DSP | IDLXORP_DSP | IDLWDT_DSP
	//save_dsp_idlect2 = __raw_readw(DSP_IDLECT2);
	//__raw_writew(0, DSP_IDLECT2);

	//Fix me: this may be useless
	omap_clearbitw(ARM_RSTCT1, 1);	//DSP_EN = 0;
	omap_writel(0x0003ff1f, MPUI_CTRL);	//set MPUI to default configuration
	omap_writel(0x00000002, MPUI_DSP_BOOT_CONFIG);	//mode2: boot to idle
	omap_setbitw(ARM_RSTCT1, 1);	//DSP_EN = 1, Run DSP boot ROM code into DSP IDLE
	//while(omap_getbitw(ARM_SYSST, 6) == 0);	//wait until IDLE_DSP = 0;

	
	omap_clearbitw(ARM_CKCTL, 13);		//EN_DSPCK = 0, disable DSP clock
	//ULPD_POWER_CTRL_REG[ISOLATION_CONTROL(12)] =1

	//Power down using tps65013[10][GPIO3] to shutdown DSP power
	
}

static void shutdown_MPU()
{
	info("shutdown_MPU");
	omap_writel(0x00F5, OMAP_WDT_TIMER_MODE);
	omap_writel(0x00A0, OMAP_WDT_TIMER_MODE);
	while(omap_getbitl(OMAP_WDT_TIMER_MODE, 15) == 1);	//wait until switch from WD to timer mode

	omap_writel(0xAAAA, OMAP_WATCHDOG_SPR);	//disable 32k watchdog sequence
	while(omap_getbitl(OMAP_WATCHDOG_WPS, 4) == 1);

	omap_writel(0x5555, OMAP_WATCHDOG_SPR);	//disable 32k watchdog sequence
	while(omap_getbitl(OMAP_WATCHDOG_WPS, 4) == 1);

	//Fix me: this may be not enough
	omap_setbitw(ARM_IDLECT2, 2);		//EN_PERCK = 1

	omap_setregw(ARM_IDLECT1, 0x13c7);	//IDL_CLKOUT_ARM | IDLTIM_ARM | IDLAPI_ARM | IDLDPLL_ARM | IDLIF_ARM | IDLPER_ARM | IDLXORP_ARM | IDLWDT_ARM
	omap_clearregw(ARM_IDLECT2, 0x0348)	;	//EN_GPIOCK, DMACK_REQ, EN_APICK, EN_LCDCK = 0
	//where is the ARM_IDLECT3?
	
	omap_clearbitw(SYSDMA_LCD_CTRL, 8);		//LDP = 0, Disable SOSSI external LCD clock request
	omap_clearbitl(OMAP16XX_RESET_CONTROL, 6);	//CONF_RNG_IDLE_MODE
	omap_setregl(ARM_RETCTL, 0);		//EN_SLPZVSS = 1, configure the diode drop of the VSS of memory cell

	//UART2_CLK_MODE_R, MMC_SD_CLK_REQ_R, MMC_SD2_CLK_REQ_R, COM_MCLK_12_48_SEL, USB_HOST_HHC_UHOST_EN_R
	omap_clearregl(MOD_CONF_CTRL_0, 0x40901200);

	//omap_clearbitl(0xFFFB0408/*OTG_SYSCON_2_REG*/, 8);	//UHOST_EN = 0, reset USB host controller & kill 48MHz clock request
	omap_setbitw(ULPD_CLOCK_CTRL, 5);		//DIS_USB_PVCI_CLK = 1, Disable USB Device PVCI clock
	omap_clearbitw(ULPD_CLOCK_CTRL, 4);		//USB_MCLK_EN = 0, disable USB hub clock output
	omap_clearbitw(ULPD_CLOCK_CTRL, 3);		//TI_RESERVED_EN =0, disable clock on SYS_CLK_OUT

	omap_setregw(ULPD_SOFT_DISABLE_REQ, 0x0043);	//DIS_USB_HOST_DPLL_REQ | DIS_SDW_MCLK_REQ | DIS_COM_MCLK_REQ
	omap_clearbitw(ULPD_SDW_CLK_DIV_CTL_SEL, 1);	//SDW_ULPD_PLL_CLK_REQ = 0, disable SW request for BCLK
	omap_clearbitw(ULPD_COM_CLK_DIV_CTL_SEL, 1);	//COM_ULPD_PLL_CLK_REQ = 0, disable SW request for MCLK
	omap_writew(0x0000, ULPD_SOFT_REQ);		//ignore USB Device DPLL request

	omap_setbitw(ULPD_ENA_MCPM, 0);		//EN_SLPZVSS = 1, Enable MCPM on VSS side
		
	omap_clearbitw(ARM_IDLECT1, 10);		//WKUP_MODE = 0, wakeup fully controlled by the ULPD instead of simple IRQ

	omap_clearbitw(ULPD_POWER_CTRL, 1);		//LOW_PWR_REQ = 0
	omap_setregw(ULPD_POWER_CTRL, 0x0211);	//OSC_STOP_EN | DEEP_SLEEP_TRANSITION_EN | LOW_PWR_EN

	//omap_writew(150, ULPD_SETUP_ANALOG_CELL3);	//delay when exiting deep sleep
	
	
}

static void shutdown_TC()
{
	info("shutdown_TC");

	//oh! yes, ARM_IDLECT3 is here
	omap_setregw(ARM_IDLECT3, 0x002A);	//IDLTC2 | IDLTC1 | IDLOCPI
	omap_setregl(EMIFS_CONFIG, 0x0000000c);	//PDE | PWD_EN, system power down ack enable, dynamic auto idle
	
	
}

static int wakeup_reason = 0;
#define WAKEUP_BY_KEYPAD		(1 << 0)
#define WAKEUP_BY_WIFI			(1 << 1)
#define WAKEUP_BY_USB			(1 << 2)
#define WAKEUP_BY_RTC			(1 << 3)
static void determine_wakeup_reason(void)
{
	unsigned int ih2_0_sir, ih2_1_sir, ih2_2_sir;
	unsigned int wifi_irqstatus, pwr_irqstatus, usb_irqstatus;
	unsigned char alarm_status;
	
	ih2_0_sir = omap_readl(OMAP_IH2_0_SIR_IRQ);
	ih2_1_sir = omap_readl(OMAP_IH2_1_SIR_IRQ);
	ih2_2_sir = omap_readl(OMAP_IH2_2_SIR_IRQ);

	//wifi_irqstatus = omap_get_gpio_irqstatus(31);
	wifi_irqstatus = omap_get_gpio_irqstatus(11);
	pwr_irqstatus = omap_get_gpio_irqstatus(40);
	usb_irqstatus = omap_get_gpio_irqstatus(0);
	
	alarm_status = omap_readb(OMAP_RTC_STATUS_REG);

	wakeup_reason = 0;
	if(ih2_0_sir & 0x01)
		wakeup_reason |= WAKEUP_BY_KEYPAD;
	if(pwr_irqstatus & 0x0100)	//power key
		wakeup_reason |= WAKEUP_BY_KEYPAD;
	if(wifi_irqstatus & 0x8000)
		wakeup_reason |= WAKEUP_BY_WIFI;
	if(usb_irqstatus & 0x01)
		wakeup_reason |=WAKEUP_BY_USB;
	if(alarm_status & OMAP_RTC_STATUS_ALARM)
		wakeup_reason |= WAKEUP_BY_RTC;
		
	/*printk(KERN_INFO"PM: 0x%08x\n", ih2_0_sir);
	printk(KERN_INFO"PM: 0x%08x\n", ih2_1_sir);
	printk(KERN_INFO"PM: 0x%08x\n", ih2_2_sir);
	printk(KERN_INFO"PM: WiFi IRQ=%d, Power Key IRQ=%d, USB IRQ=%d\n", wifi_irqstatus, pwr_irqstatus, usb_irqstatus);
	printk(KERN_INFO"PM: RTC Status=0x%02x\n", alarm_status);*/
	
}

int get_wakeup_reason()
{
	return wakeup_reason;
}
EXPORT_SYMBOL(get_wakeup_reason);
static int alarm_wakeup_second = 0;
int set_alarm_wakeup(int second)
{
	alarm_wakeup_second = second;
}
EXPORT_SYMBOL(set_alarm_wakeup);
void set_alarm()
{
	struct rtc_time rtc_tm;
	
	if(alarm_wakeup_second == 0)
		return;
	get_rtc_time(&rtc_tm);
	printk("RTC:%d-%d-%d, %02d:%02d:%02d\n", 	//for testing, should be remove
			rtc_tm.tm_mday, rtc_tm.tm_mon + 1, rtc_tm.tm_year + 1900,
			rtc_tm.tm_hour, rtc_tm.tm_min, rtc_tm.tm_sec);
	
	rtc_tm.tm_sec += alarm_wakeup_second;
	if(rtc_tm.tm_sec >= 60) {
		rtc_tm.tm_min += (rtc_tm.tm_sec / 60);
		rtc_tm.tm_sec %= 60;
	}
	if(rtc_tm.tm_min >= 60) {
		rtc_tm.tm_min %=60;
		rtc_tm.tm_hour ++;
	}
	if(rtc_tm.tm_hour == 24) {
		rtc_tm.tm_hour = 0;
		rtc_tm.tm_mday ++;	//Fix Me: dangerous, normalize in set_rtc_alm_time()
	}

	set_rtc_alm_time(&rtc_tm);
	enable_alarm_irq();

}

void omap_pm_suspend(void)
{
	//unsigned int mask32 = 0;
	unsigned long arg0 = 0, arg1 = 0;
	int (*func_ptr)(unsigned short, unsigned short) = 0;
	//unsigned short save_dsp_idlect2;
	//int count = 0;
	


#ifdef VM1185T_R01
	if(pm_send_by_id(87, PM_SUSPEND, NULL) < 0) {
		printk("suspend wifi fail\n");
		return;
	}
#endif
#ifdef WIFI_PHONE_RTC
	struct rtc_time rtc_tm; 
	struct timespec time;

	info("save system time");
	time.tv_nsec = 0;
	get_rtc_time(&rtc_tm);
	rtc_tm_to_time(&rtc_tm, &time.tv_sec); 
	save_time_delta(&rtc_delta, &time);
	set_alarm();
#endif

	printk("PM: OMAP%x is entering deep sleep now ...\n", system_rev);

	if (machine_is_omap_osk()) {
		/* Stop LED1 (D9) blink */
		tps65010_set_led(LED1, OFF);
	}
#ifdef WIFI_PHONE_PM
	//turn off LCD backlight
	info("Turn off LCD backlight");
	set_lcd_backlight(0);

	//turn off Keypad backlight
	info("Turn off keypad backlight");
	set_keypad_backlight(0);

#endif
	/*
	 * Step 1: turn off interrupts
	 */

	cli();
	clf();

	/*
	 * Step 2: save registers
	 *
	 * The omap is a strange/beautiful device. The caches, memory
	 * and register state are preserved across power saves.
	 * We have to save and restore very little register state to
	 * idle the omap.
         *
 	 * Save interrupt, MPUI, ARM and UPLD control registers.
	 */
#ifdef WIFI_PHONE_PM
	save_chip_context(SAVE);
#endif

	


#if 0
	if (cpu_is_omap1510()) {
		MPUI1510_SAVE(OMAP_IH1_MIR);
		MPUI1510_SAVE(OMAP_IH2_MIR);
		MPUI1510_SAVE(MPUI_CTRL);
		MPUI1510_SAVE(MPUI_DSP_BOOT_CONFIG);
		MPUI1510_SAVE(MPUI_DSP_API_CONFIG);
		MPUI1510_SAVE(EMIFS_CONFIG);
		MPUI1510_SAVE(EMIFF_SDRAM_CONFIG);
	} else if (cpu_is_omap16xx()) {
		MPUI1610_SAVE(OMAP_IH1_MIR);
		MPUI1610_SAVE(OMAP_IH2_0_MIR);
		MPUI1610_SAVE(OMAP_IH2_1_MIR);
		MPUI1610_SAVE(OMAP_IH2_2_MIR);
		MPUI1610_SAVE(OMAP_IH2_3_MIR);
		MPUI1610_SAVE(MPUI_CTRL);
		MPUI1610_SAVE(MPUI_DSP_BOOT_CONFIG);
		MPUI1610_SAVE(MPUI_DSP_API_CONFIG);
		MPUI1610_SAVE(EMIFS_CONFIG);
		MPUI1610_SAVE(EMIFF_SDRAM_CONFIG);
	}

	ARM_SAVE(ARM_CKCTL);
	ARM_SAVE(ARM_IDLECT1);
	ARM_SAVE(ARM_IDLECT2);
	ARM_SAVE(ARM_EWUPCT);
	ARM_SAVE(ARM_RSTCT1);
	ARM_SAVE(ARM_RSTCT2);
	ARM_SAVE(ARM_SYSST);
	ULPD_SAVE(ULPD_CLOCK_CTRL);
	ULPD_SAVE(ULPD_STATUS_REQ);
#endif
	/*
	 * Step 3: LOW_PWR signal enabling
	 *
	 * Allow the LOW_PWR signal to be visible on MPUIO5 ball.
	 */
#if 0
	if (cpu_is_omap1510()) {
		/* POWER_CTRL_REG = 0x1 (LOW_POWER is available) */
		omap_writew(omap_readw(ULPD_POWER_CTRL) | 
			    OMAP1510_ULPD_LOW_POWER_REQ, ULPD_POWER_CTRL);
	} else if (cpu_is_omap16xx()) {
		/* POWER_CTRL_REG = 0x1 (LOW_POWER is available) */
		omap_writew(omap_readw(ULPD_POWER_CTRL) |
			    OMAP1610_ULPD_LOW_POWER_REQ, ULPD_POWER_CTRL);
	}
#endif
	
	/*
	 * Step 4: OMAP DSP Shutdown
	 */
#ifdef WIFI_PHONE_PM
	shutdown_DSP();
#endif
#if 0
	/* Set DSP_RST = 1 and DSP_EN = 0, put DSP block into reset */
	omap_writel((omap_readl(ARM_RSTCT1) | DSP_RST) & ~DSP_ENABLE,
		    ARM_RSTCT1);

	/* Set DSP boot mode to DSP-IDLE, DSP_BOOT_MODE = 0x2 */
        omap_writel(DSP_IDLE_MODE, MPUI_DSP_BOOT_CONFIG);

	/* Set EN_DSPCK = 0, stop DSP block clock */
	omap_writel(omap_readl(ARM_CKCTL) & ~DSP_CLOCK_ENABLE, ARM_CKCTL);

	/* Stop any DSP domain clocks */
	omap_writel(omap_readl(ARM_IDLECT2) | (1<<EN_APICK), ARM_IDLECT2);
	save_dsp_idlect2 = __raw_readw(DSP_IDLECT2);
	__raw_writew(0, DSP_IDLECT2);
#endif
	/*
	 * Step 5: Wakeup Event Setup
	 */
	info("omap_pm_wakeup_setup");
	omap_pm_wakeup_setup();
	/*
	 * Step 6a: ARM and Traffic controller shutdown
	 *
	 * Step 6 starts here with clock and watchdog disable
	 */
#ifdef WIFI_PHONE_PM
	shutdown_MPU();
	shutdown_TC();
	
	// Disable 48 MHz clocks for UART1/2/3
	info("Disable 48 MHz for UART1/2/3");
	omap_clearregl(MOD_CONF_CTRL_0, 0xE0000000);

	
	//Set DPLL bypass to 12 MHz for a well defined post-WFI delay of NOPs
	info("Set DPLL to bypass mode");
	omap_clearbitw(DPLL_CTL, 4);
#endif
#if 0
	/* stop clocks */
	mask32 = omap_readl(ARM_IDLECT2);
	mask32 &= ~(1<<EN_WDTCK);  /* bit 0 -> 0 (WDT clock) */
	mask32 |=  (1<<EN_XORPCK); /* bit 1 -> 1 (XORPCK clock) */
	mask32 &= ~(1<<EN_PERCK);  /* bit 2 -> 0 (MPUPER_CK clock) */
	mask32 &= ~(1<<EN_LCDCK);  /* bit 3 -> 0 (LCDC clock) */
	mask32 &= ~(1<<EN_LBCK);   /* bit 4 -> 0 (local bus clock) */
	mask32 |=  (1<<EN_APICK);  /* bit 6 -> 1 (MPUI clock) */
	mask32 &= ~(1<<EN_TIMCK);  /* bit 7 -> 0 (MPU timer clock) */
	mask32 &= ~(1<<DMACK_REQ); /* bit 8 -> 0 (DMAC clock) */
	mask32 &= ~(1<<EN_GPIOCK); /* bit 9 -> 0 (GPIO clock) */
	omap_writel(mask32, ARM_IDLECT2);

	/* disable ARM watchdog */
	omap_writel(0x00F5, OMAP_WDT_TIMER_MODE);
	omap_writel(0x00A0, OMAP_WDT_TIMER_MODE);
#endif
	/*
	 * Step 6b: ARM and Traffic controller shutdown
	 *
	 * Step 6 continues here. Prepare jump to power management
	 * assembly code in internal SRAM.
	 *
	 * Since the omap_cpu_suspend routine has been copied to
	 * SRAM, we'll do an indirect procedure call to it and pass the
	 * contents of arm_idlect1 and arm_idlect2 so it can restore
	 * them when it wakes up and it will return.
	 */

	arg0 = arm_sleep_save[ARM_SLEEP_SAVE_ARM_IDLECT1];
	arg1 = arm_sleep_save[ARM_SLEEP_SAVE_ARM_IDLECT2];

	if (cpu_is_omap1510()) {
		func_ptr = (void *)(OMAP1510_SRAM_API_SUSPEND);
	} else if (cpu_is_omap1610() || cpu_is_omap1710()) {
		func_ptr = (void *)(OMAP1610_SRAM_API_SUSPEND);
	} else if (cpu_is_omap5912()) {
		func_ptr = (void *)(OMAP5912_SRAM_API_SUSPEND);
	}

	/*
	 * Step 6c: ARM and Traffic controller shutdown
	 *
	 * Jump to assembly code. The processor will stay there
 	 * until wake up.
	 */

        func_ptr(arg0, arg1);

	//Set DPLL to lock mode
	omap_setbitw(DPLL_CTL, 4);	
	/*
	 * If we are here, processor is woken up!
	 */
	//info("wakeup from sleep.S");
	if (cpu_is_omap1510()) {
		/* POWER_CTRL_REG = 0x0 (LOW_POWER is disabled) */
		omap_writew(omap_readw(ULPD_POWER_CTRL) & 
			    ~OMAP1510_ULPD_LOW_POWER_REQ, ULPD_POWER_CTRL);
	} else if (cpu_is_omap16xx()) { 
		/* POWER_CTRL_REG = 0x0 (LOW_POWER is disabled) */
		omap_writew(omap_readw(ULPD_POWER_CTRL) &
			    ~OMAP1610_ULPD_LOW_POWER_REQ, ULPD_POWER_CTRL);
	}

	determine_wakeup_reason();
	
#ifdef WIFI_PHONE_PM
	save_chip_context(RESTORE);
#endif

	/* Restore DSP clocks */
#if 0
	omap_writel(omap_readl(ARM_IDLECT2) | (1<<EN_APICK), ARM_IDLECT2);
	__raw_writew(save_dsp_idlect2, DSP_IDLECT2);
	//ARM_RESTORE(ARM_IDLECT2);

	/*
	 * Restore ARM state, except ARM_IDLECT1/2 which omap_cpu_suspend did
	 */

	ARM_RESTORE(ARM_CKCTL);
	ARM_RESTORE(ARM_EWUPCT);
	ARM_RESTORE(ARM_RSTCT1);
	ARM_RESTORE(ARM_RSTCT2);
	ARM_RESTORE(ARM_SYSST);
	ULPD_RESTORE(ULPD_CLOCK_CTRL);
	ULPD_RESTORE(ULPD_STATUS_REQ);

	if (cpu_is_omap1510()) {
		MPUI1510_RESTORE(MPUI_CTRL);
		MPUI1510_RESTORE(MPUI_DSP_BOOT_CONFIG);
		MPUI1510_RESTORE(MPUI_DSP_API_CONFIG);
		MPUI1510_RESTORE(EMIFS_CONFIG);
		MPUI1510_RESTORE(EMIFF_SDRAM_CONFIG);
		MPUI1510_RESTORE(OMAP_IH1_MIR);
		MPUI1510_RESTORE(OMAP_IH2_MIR);
	} else if (cpu_is_omap16xx()) {
		MPUI1610_RESTORE(MPUI_CTRL);
		MPUI1610_RESTORE(MPUI_DSP_BOOT_CONFIG);
		MPUI1610_RESTORE(MPUI_DSP_API_CONFIG);
		MPUI1610_RESTORE(EMIFS_CONFIG);
		MPUI1610_RESTORE(EMIFF_SDRAM_CONFIG);

		MPUI1610_RESTORE(OMAP_IH1_MIR);
		MPUI1610_RESTORE(OMAP_IH2_0_MIR);
		MPUI1610_RESTORE(OMAP_IH2_1_MIR);
		MPUI1610_RESTORE(OMAP_IH2_2_MIR);
		MPUI1610_RESTORE(OMAP_IH2_3_MIR);
	}
#endif
	/*
	 * Reenable interrupts
	 */

	sti();
	stf();

	printk("PM: OMAP%x is re-starting from deep sleep...\n", system_rev);

	printk(KERN_INFO"PM: Wakeup Reason = 0x%04x\n", wakeup_reason);
	
	if (machine_is_omap_osk()) {
		/* Let LED1 (D9) blink again */
		tps65010_set_led(LED1, BLINK);
	}
#ifdef VM1185T_R01
	//attach netdev of wifi module
	info("Resume the Wi-Fi module");
	pm_send_by_id(87, PM_RESUME, NULL);
#endif
#ifdef WIFI_PHONE_RTC
	info("restore system time");
	time.tv_nsec = 0;
	get_rtc_time(&rtc_tm);
	rtc_tm_to_time(&rtc_tm, &time.tv_sec); 
	restore_time_delta(&rtc_delta, &time);
	disable_alarm_irq();
#endif
}

#ifdef CONFIG_PROC_FS		
static int g_read_completed;
/*
 * Writing to /proc/pm puts the CPU in sleep mode
 */
static ssize_t
omap_pm_write_proc(struct file * file, const char * buf, size_t count,
    loff_t *ppos)
{
	int rc;
	rc = pm_send_all(PM_SUSPEND, (void *)3);

	if (rc == 0) {
		omap_pm_suspend();

		pm_send_all(PM_RESUME, (void *)0);
	}
	return count;
}


static int omap_pm_read_proc(
	char *page_buffer,
	char **my_first_byte,
	off_t virtual_start,
	int length,
	int *eof,
	void *data)
{
	int my_buffer_offset = 0;
	char * const my_base = page_buffer;

	ARM_SAVE(ARM_CKCTL);
	ARM_SAVE(ARM_IDLECT1);
	ARM_SAVE(ARM_IDLECT2);
	ARM_SAVE(ARM_EWUPCT);
	ARM_SAVE(ARM_RSTCT1);
	ARM_SAVE(ARM_RSTCT2);
	ARM_SAVE(ARM_SYSST);

	ULPD_SAVE(ULPD_IT_STATUS);
	ULPD_SAVE(ULPD_CLOCK_CTRL);
	ULPD_SAVE(ULPD_SOFT_REQ);
	ULPD_SAVE(ULPD_STATUS_REQ);
	ULPD_SAVE(ULPD_DPLL_CTRL);
	ULPD_SAVE(ULPD_POWER_CTRL);

	if (cpu_is_omap1510()) {
		MPUI1510_SAVE(MPUI_CTRL);
		MPUI1510_SAVE(MPUI_DSP_STATUS);
		MPUI1510_SAVE(MPUI_DSP_BOOT_CONFIG);
		MPUI1510_SAVE(MPUI_DSP_API_CONFIG);
		MPUI1510_SAVE(EMIFF_SDRAM_CONFIG);
		MPUI1510_SAVE(EMIFS_CONFIG);
	} else if (cpu_is_omap16xx()) {
		MPUI1610_SAVE(MPUI_CTRL);
		MPUI1610_SAVE(MPUI_DSP_STATUS);
		MPUI1610_SAVE(MPUI_DSP_BOOT_CONFIG);
		MPUI1610_SAVE(MPUI_DSP_API_CONFIG);
		MPUI1610_SAVE(EMIFF_SDRAM_CONFIG);
		MPUI1610_SAVE(EMIFS_CONFIG);
	}

	if (virtual_start == 0) {
		g_read_completed = 0;

		my_buffer_offset += sprintf(my_base + my_buffer_offset,
		   "ARM_CKCTL_REG:            0x%-8x     \n"
		   "ARM_IDLECT1_REG:          0x%-8x     \n"
		   "ARM_IDLECT2_REG:          0x%-8x     \n"
		   "ARM_EWUPCT_REG:           0x%-8x     \n"
		   "ARM_RSTCT1_REG:           0x%-8x     \n"
		   "ARM_RSTCT2_REG:           0x%-8x     \n"
		   "ARM_SYSST_REG:            0x%-8x     \n"
		   "ULPD_IT_STATUS_REG:       0x%-4x     \n"
		   "ULPD_CLOCK_CTRL_REG:      0x%-4x     \n"
		   "ULPD_SOFT_REQ_REG:        0x%-4x     \n"
		   "ULPD_DPLL_CTRL_REG:       0x%-4x     \n"
		   "ULPD_STATUS_REQ_REG:      0x%-4x     \n"
		   "ULPD_POWER_CTRL_REG:      0x%-4x     \n",
		   ARM_SHOW(ARM_CKCTL),
		   ARM_SHOW(ARM_IDLECT1),
		   ARM_SHOW(ARM_IDLECT2),
		   ARM_SHOW(ARM_EWUPCT),
		   ARM_SHOW(ARM_RSTCT1),
		   ARM_SHOW(ARM_RSTCT2),
		   ARM_SHOW(ARM_SYSST),
		   ULPD_SHOW(ULPD_IT_STATUS),
		   ULPD_SHOW(ULPD_CLOCK_CTRL),
		   ULPD_SHOW(ULPD_SOFT_REQ),
		   ULPD_SHOW(ULPD_DPLL_CTRL),
		   ULPD_SHOW(ULPD_STATUS_REQ),
		   ULPD_SHOW(ULPD_POWER_CTRL));

		if (cpu_is_omap1510()) {
			my_buffer_offset += sprintf(my_base + my_buffer_offset,
			   "MPUI1510_CTRL_REG             0x%-8x \n"
			   "MPUI1510_DSP_STATUS_REG:      0x%-8x \n"
			   "MPUI1510_DSP_BOOT_CONFIG_REG: 0x%-8x \n"
		   	   "MPUI1510_DSP_API_CONFIG_REG:  0x%-8x \n"
		   	   "MPUI1510_SDRAM_CONFIG_REG:    0x%-8x \n"
		   	   "MPUI1510_EMIFS_CONFIG_REG:    0x%-8x \n",
		   	   MPUI1510_SHOW(MPUI_CTRL),
		   	   MPUI1510_SHOW(MPUI_DSP_STATUS),
		   	   MPUI1510_SHOW(MPUI_DSP_BOOT_CONFIG),
		   	   MPUI1510_SHOW(MPUI_DSP_API_CONFIG),
		   	   MPUI1510_SHOW(EMIFF_SDRAM_CONFIG),
		   	   MPUI1510_SHOW(EMIFS_CONFIG));
		} else if (cpu_is_omap16xx()) {
			my_buffer_offset += sprintf(my_base + my_buffer_offset,
			   "MPUI1610_CTRL_REG             0x%-8x \n"
			   "MPUI1610_DSP_STATUS_REG:      0x%-8x \n"
			   "MPUI1610_DSP_BOOT_CONFIG_REG: 0x%-8x \n"
		   	   "MPUI1610_DSP_API_CONFIG_REG:  0x%-8x \n"
		   	   "MPUI1610_SDRAM_CONFIG_REG:    0x%-8x \n"
		   	   "MPUI1610_EMIFS_CONFIG_REG:    0x%-8x \n",
		   	   MPUI1610_SHOW(MPUI_CTRL),
		   	   MPUI1610_SHOW(MPUI_DSP_STATUS),
		   	   MPUI1610_SHOW(MPUI_DSP_BOOT_CONFIG),
		   	   MPUI1610_SHOW(MPUI_DSP_API_CONFIG),
		   	   MPUI1610_SHOW(EMIFF_SDRAM_CONFIG),
		   	   MPUI1610_SHOW(EMIFS_CONFIG));
		}

		g_read_completed++;
	} else if (g_read_completed >= 1) {
		 *eof = 1;
		 return 0;
	}
	g_read_completed++;

	*my_first_byte = page_buffer;
	return  my_buffer_offset;
}

/*
 *	omap_pm_prepare - Do preliminary suspend work.
 *	@state:		suspend state we're entering.
 *
 */

static int omap_pm_prepare(u32 state)
{
	int error = 0;

	switch (state)
	{
	case PM_SUSPEND_STANDBY:
	case PM_SUSPEND_MEM:
		break;

	case PM_SUSPEND_DISK:
		return -ENOTSUPP;

	default:
		return -EINVAL;
	}

	return error;
}


/*
 *	omap_pm_enter - Actually enter a sleep state.
 *	@state:		State we're entering.
 *
 */

static int omap_pm_enter(u32 state)
{
	switch (state)
	{
	case PM_SUSPEND_STANDBY:
	case PM_SUSPEND_MEM:
		omap_pm_suspend();
		break;

	case PM_SUSPEND_DISK:
		return -ENOTSUPP;

	default:
		return -EINVAL;
	}

	return 0;
}


/**
 *	omap_pm_finish - Finish up suspend sequence.
 *	@state:		State we're coming out of.
 *
 *	This is called after we wake back up (or if entering the sleep state
 *	failed).
 */

static int omap_pm_finish(u32 state)
{
	return 0;
}
#ifdef WIFI_PHONE_PM		// 2005-11-03
static int wifi_freq_scale(unsigned long rate)
{
	int err;
	
	if(freeze_processes()) {
		err = -EAGAIN;
		return err;
	}
	
	
	change_clk_rate(rate);
	

	thaw_processes();
	return 0;

}

//most of works are handled by user mode apmd
static int omap_pm_set_high_speed(u32 state)
{
	int err = 0;
	info("omap_pm_set_high_speed, state=%d", state);

#ifdef WIFI_PHONE_FREQ_SCALE	
	if(wifi_freq_scale(216000000) == 0) {
		info("set clock rate: 216MHz");
	}
#endif
	//set SDRAM to high speed

	//set WiFi full on

	//set LCD brightness 100%
	//info("set LCM brightness to 100%");
	//set_lcd_brightness(250);
	//info("Turn on LCD Backlight");
	//set_lcd_backlight(1);
	
	//resume Audio Codec
	//info("resume Audio Codec");
	//pm_send_by_id(0x41, PM_RESUME, state);

	//turn on keypad controller
	return err;
	
}

static int omap_pm_set_low_speed(u32 state)
{
	int err = 0;
	info("omap_pm_set_low_speed, state=%d", state);
#ifdef WIFI_PHONE_FREQ_SCALE
	if(wifi_freq_scale(120000000) == 0) {
		info("set clock rate: 120MHz");
	}
#endif
	//set SDRAM to low speed

	//set Wifi to Power Saving

	//set LCD brightness 70%
	//#ifdef VM1185T_R01
	//info("set LCM brightness to 70%");
	//set_lcd_brightness(178);
	//#endif

	//turn on LCD backlight
	//info("delete LCM backlight timer");
	//del_lcmbklight_timer();

	//info("turn on LCD backlight");
	//set_lcd_backlight(1);

	//suspend Audio Codec
	//info("suspend Audio Codec");
	//pm_send_by_id(0x41, PM_SUSPEND, state);

	//turn on keypad controller

	//keypad backlite depend on key pressed
	return err;
}

static int omap_pm_set_idle_mode(u32 state)
{
	int err = 0;
	info("omap_pm_set_idle_mode, state=%d", state);
#ifdef WIFI_PHONE_FREQ_SCALE	
	if(wifi_freq_scale(30000000) == 0) {
		info("set clock rate: 30MHz");
	}
#endif
	//set SDRAM to low speed

	//set Wifi to power saving mode
	//info("Suspend Wi-Fi");
	//pm_send_by_id(0x57, PM_SUSPEND, state);

	//set LCM brightness to 10%
	//#ifdef VM1185T_R01
	//info("set LCM brightness to 10%");
	//set_lcd_brightness(25);
	//#endif
	
	//turn off LCD backlight
	//info("Turn off LCD backlight");
	//set_lcd_backlight(0);
	
	//suspend Audio Codec
	//info("Suspend Audio Codec");
	//pm_send_by_id(0x41, PM_SUSPEND, state);
	

	//turn on keypad controller

	//turn off keypad backlite
	//info("Turn off keypad backlight");
	//set_keypad_backlight(0);

	return err;
}

static int omap_pm_set_power_saving(u32 state)
{
	int err = 0;
	info("omap_pm_set_power_saving, state=%d", state);
#ifdef WIFI_PHONE_FREQ_SCALE
	if(wifi_freq_scale(30000000) == 0) {
		info("set clock rate: 30MHz");
	}
#endif
	//set LCD brightness to 30%
	//info("set LCM brightness to 10%");
	//set_lcd_brightness(15);

	
	//info("Turn on LCD backlight");
	//set_lcd_backlight(1);
	
	//turn on keypad controller

	//turn off keypad backlite
	//info("Turn off keypad backlight");
	//set_keypad_backlight(0);
	
	return err;
	
	
}

static int omap_pm_set_charging(u32 state)
{
	int err =0;
	info("omap_pm_set_charging%d", state);
	switch(state) {
		case 1:	//charging mode 1: should never go here
			//CPU low speed
			
			//set SDRAM to low speed

			//set Wifi off

			//set LCD on
			//info("Turn on LCD backlight");
			//set_lcd_backlight(1);
			
			//suspend Audio Codec
			//info("suspend Audio Codec");
			//pm_send_by_id(0x41, PM_SUSPEND, state);

			//turn off keypad controller

			//turn off keypad backlite
			//info("Turn off keypad backlight");
			//set_keypad_backlight(0);
			break;
		case 2:	//charging mode 2
			//CPU high speed
			//SDRAM high speed 
			//Wifi on
			//LCD on
			//info("Turn on LCD backlight");
			//set_lcd_backlight(1);
			
			//Audio Codec on 
			//info("Resume Audio Codec");
			//pm_send_by_id(0x41, PM_RESUME, state);
			
			//keypad controller on
			//keypad backlite depend on key pressed
			break;
		case 3:	//charging mode 3
			//CPU low speed
			//SDRAM low speed
			//Wifi to Power Saving
			
			//LCD on
			//info("Turn on LCD backlight");
			//set_lcd_backlight(1);
			//info("Add LCM backlight idle timer");
			//add_lcmbklight_timer();
			
			//Audio Codec On
			//info("Resume Audio Codec");
			//pm_send_by_id(0x41, PM_RESUME, state);
			
			//Keypad Controller On
			
			//Keypad backlight off
			//info("Turn off keypad backlight");
			//set_keypad_backlight(0);
			break;
		default:
			break;
	}
	return err;
}

struct wifi_pm_ops wifi_phone_pm_ops = {
	.set_high_speed	= omap_pm_set_high_speed,
	.set_low_speed	= omap_pm_set_low_speed,
	.set_idle_mode	= omap_pm_set_idle_mode,
	.set_power_saving = omap_pm_set_power_saving,
	.set_charging		= omap_pm_set_charging,
};

#endif  //WIFI_PHONE_PM


#ifdef WIFI_PHONE_POWER_KEY
//ron lee: GPIO 40 wakeup: (2005-11-02)
static irqreturn_t omap_wakeup_interrupt(int irq, void * dev, struct pt_regs * regs)
{
	info("peripheral wakeup");
 	return IRQ_HANDLED;
}
 
 static struct irqaction omap_wakeup_irq = {
	.name = "peripheral wakeup",
	.flags = SA_INTERRUPT,
	.handler = omap_wakeup_interrupt
};
//ron lee
#endif	//WIFI_PHONE_POWER_KEY


struct pm_ops omap_pm_ops ={
	.pm_disk_mode = 0,
        .prepare        = omap_pm_prepare,
        .enter          = omap_pm_enter,
        .finish         = omap_pm_finish,
};

static int __init omap_pm_init(void)
{
	struct proc_dir_entry *entry;

	printk("Power Management for TI OMAP.\n");
	pm_idle = omap_pm_idle;
	/*
	 * We copy the assembler sleep/wakeup routines to SRAM.
	 * These routines need to be in SRAM as that's the only
	 * memory the MPU can see when it wakes up.
	 */

#ifdef	CONFIG_ARCH_OMAP1510
	if (cpu_is_omap1510()) {
		memcpy((void *)OMAP1510_SRAM_IDLE_SUSPEND, omap1510_idle_loop_suspend,
		       omap1510_idle_loop_suspend_sz);
		memcpy((void *)OMAP1510_SRAM_API_SUSPEND, omap1510_cpu_suspend,
		       omap1510_cpu_suspend_sz);
	} else
#endif
	if (cpu_is_omap1610() || cpu_is_omap1710()) {
		memcpy((void *)OMAP1610_SRAM_IDLE_SUSPEND, omap1610_idle_loop_suspend,
		       omap1610_idle_loop_suspend_sz);
		memcpy((void *)OMAP1610_SRAM_API_SUSPEND, omap1610_cpu_suspend,
		       omap1610_cpu_suspend_sz);
	} else if (cpu_is_omap5912()) {
		memcpy((void *)OMAP5912_SRAM_IDLE_SUSPEND, omap1610_idle_loop_suspend,
		       omap1610_idle_loop_suspend_sz);
		memcpy((void *)OMAP5912_SRAM_API_SUSPEND, omap1610_cpu_suspend,
		       omap1610_cpu_suspend_sz);
	}
	
#ifdef WIFI_PHONE_POWER_KEY				
	omap_cfg_reg(W15_1710_GPIO40);

	/* don't use GPIO31 as wifi host wakeup GPIO*/
#if 0
	omap_cfg_reg(C17_1710_GPIO31);		//ron_lee 2005-12-09 WIFI wakeup GPIO
	omap_request_gpio(31);
	omap_set_gpio_edge_ctrl(31, OMAP_GPIO_BOTH_EDGES);		//Wifi gpio edge trigger
#endif
//	omap_set_gpio_edge_ctrl(13, OMAP_GPIO_RISING_EDGE);	//USB VBus edge trigger
	//omap_set_gpio_direction(40, 1);	//Set GPIO 40 as input
	//omap_set_gpio_wakeup(40, 1);	//Enable GPIO 40 wakeup
	//omap_set_gpio_edge_ctrl(40, OMAP_GPIO_BOTH_EDGES); //detect rising edge
	//omap_set_gpio_irqenable(40, 1);	//enable GPIO 40 Interrupt
	//ron lee: GPIO 40 wakeup: (2005-11-02)
	if (cpu_is_omap16xx())
		setup_irq(INT_1610_GPIO_WAKEUP, &omap_wakeup_irq);
#endif	//WIFI_PHONE_POWER_KEY

	pm_set_ops(&omap_pm_ops);
	
#ifdef WIFI_PHONE_PM
	pm_register_wifi_pmops(&wifi_phone_pm_ops);
#endif	//WIFI_PHONE_PM

	entry = create_proc_read_entry("pm", S_IWUSR | S_IRUGO, NULL,
	    omap_pm_read_proc, 0);
	if (entry) {
		entry->write_proc = (write_proc_t *)omap_pm_write_proc;
	}
	
	//ron lee: GPIO 40 wakeup (2005-11-02)
	/* configure LOW_PWR pin */
	omap_cfg_reg(T20_1610_LOW_PWR);


	return 0;
}
__initcall(omap_pm_init);

#endif /* CONFIG_PROC_FS */

#ifdef CONFIG_SYSCTL

/*
 * ARGH!  ACPI people defined CTL_ACPI in linux/acpi.h rather than
 * linux/sysctl.h.
 */

#define CTL_ACPI 9999
#define APM_S1_SLP_TYP 19

/*
 * Send us to sleep.
 */

static int sysctl_omap_pm_suspend(void)
{
	int retval;

	retval = pm_send_all(PM_SUSPEND, (void *)3);

	if (retval == 0) {
		omap_pm_suspend();

		pm_send_all(PM_RESUME, (void *)0);
	}

	return retval;
}

static struct ctl_table pm_table[] =
{
	{APM_S1_SLP_TYP, "suspend", NULL, 0, 0600, NULL,
	    (proc_handler *)&sysctl_omap_pm_suspend},

	{0}
};

static struct ctl_table pm_dir_table[] =
{
	{CTL_ACPI, "pm", NULL, 0, 0555, pm_table},
	{0}
};

/*
 * Initialize power interface
 */
static int __init pm_init(void)
{
	register_sysctl_table(pm_dir_table, 1);
	return 0;
}

__initcall(pm_init);

#endif /* CONFIG_SYSCTL */
