/*
 * drivers/media/video/omap/camera_core.c
 *
 * Video-for-Linux (Version 2) camera capture driver for
 * the OMAP H2 and H3 camera controller.
 *
 * auther:  
 * Adapted from omap24xx driver written by Andy Lowe (source@mvista.com)
 *
 * Copyright (C) 2004 Texas Instruments.
 *
 * This file is licensed under the terms of the GNU General Public License
 * version 2. This program is licensed "as is" without any warranty of any
 * kind, whether express or implied.
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/vmalloc.h>
#include <linux/slab.h>
#include <linux/proc_fs.h>
#include <linux/ctype.h>
#include <linux/pagemap.h>
#include <linux/mm.h>
#include <linux/device.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/videodev.h>
#include <linux/pci.h>
#include <asm/semaphore.h>
#include <asm/processor.h>
#include <linux/dma-mapping.h>
#include <linux/fb.h>

#include <asm/io.h>
#include <asm/byteorder.h>
#include <asm/irq.h>
#include <asm/hardware/clock.h>

#include "camera_core.h"
#include "camera_sensor_if.h"
#include "omap_camera_if.h"

struct camera_device *camera_dev;
extern struct camera_sensor camera_sensor_if;
extern struct camera_interface omap_camera_if;

static void camera_core_sg_dma_process(struct camera_device *);
static int capture_abort= 0;

/* module parameters */
static int video_nr = -1;	/* video device minor (-1 ==> auto assign) */

/* Maximum amount of memory to use for capture buffers.
 * Default is 4800KB, enough to double-buffer SXGA.
 */
static int capture_mem = 1280*960*2*2;
/*Size of video overlay framebuffer. This determines the maximum image size
 *that can be previewed. Default is 600KB, enough for sxga.
 */
static int overlay_mem = 640*480*2;

int omap_dma_start_linked (struct camera_device *, struct sgdma_state *, void *);

/* DMA completion routine for the scatter-gather DMA fragments. */
/* This function is called when a scatter DMA fragment is completed */
int
camera_core_callback_sg_dma(struct camera_device *cam, void * arg)
{
        struct sgdma_state *sgdma;

	int sgslot = (int) arg;
        spin_lock(&cam->sg_lock);
	sgdma = cam->sgdma + sgslot;
	if(! sgdma->queued_sglist)
	{
		spin_unlock(&cam->sg_lock);
		printk(" ERROR!!!   SGDMA COMPLETED WHEN NONE QUEUED \n");
		return -EINVAL;
	}
	if(!--sgdma->queued_sglist) {
		/* queue for this sglist is empty so check whether transfer
		** of the frame has been completed */
		if(sgdma->next_sglist == sgdma->sglen) {
			dma_callback_t callback = sgdma->callback;
			void *arg = sgdma->arg;
			/* all done with this sglist */
			cam->free_sgdma++;
			if (callback) {
				spin_unlock(&cam->sg_lock);
				(*callback)(cam, arg);
				camera_core_sg_dma_process(cam);
				return 0;
			}

		}
	}
	spin_unlock(&cam->sg_lock);
	camera_core_sg_dma_process(cam);
	return 0;
	

}

/* Initialize the camera DMA driver. */
static void
camera_core_dma_init(struct camera_device *cam)
{
        int sg, ch;

        spin_lock_init(&cam->sg_lock);
	spin_lock_init(&cam->dma_lock);
	cam->free_dmach = 2;
	/* we don't know the next channel number here */
	/* cam->next_dmach = 0;*/
        cam->free_sgdma = NUM_SG_DMA;
        cam->next_sgdma = 0;
        for (sg = 0; sg < NUM_SG_DMA; sg++) {
                cam->sgdma[sg].sglen = 0;
                cam->sgdma[sg].next_sglist = 0;
                cam->sgdma[sg].queued_sglist = 0;
                cam->sgdma[sg].csr = 0;
                cam->sgdma[sg].callback = NULL;
                cam->sgdma[sg].arg = NULL;
        }
	for ( ch = 0; ch < 2; ++ch) {
		cam->camdma[ch].callback = NULL;
		cam->camdma[ch].arg = NULL;
	}
	
}

/*
 * Process the scatter-gather DMA queue by starting queued transfers
 * This function is called to program the dma to start the transfer of an image.
 */
static void
camera_core_sg_dma_process(struct camera_device *cam)
{
        unsigned long irqflags;
        int queued_sgdma, sgslot;
        struct sgdma_state *sgdma;
        const struct scatterlist *sglist;
	
        spin_lock_irqsave(&cam->sg_lock, irqflags);

        queued_sgdma = NUM_SG_DMA - cam->free_sgdma;
        sgslot = (cam->next_sgdma + cam->free_sgdma)%34;
	while ( queued_sgdma >0 ) {
		sgdma = cam->sgdma + sgslot;
		while( sgdma->next_sglist < sgdma->sglen ) {
			sglist = sgdma->sglist + sgdma->next_sglist;
			if (omap_dma_start_linked(cam, sgdma, (void *) sgslot)){
				/* dma start failed */
				spin_unlock_irqrestore(&cam->sg_lock, irqflags);
				return;
			}
			else {
				/* dma start successful */
				sgdma->next_sglist ++;
				sgdma->queued_sglist ++;
			}
		}
		queued_sgdma-- ;
		sgslot=(sgslot+1)%34;	
	}

        spin_unlock_irqrestore(&cam->sg_lock, irqflags);
}

/* Queue a scatter-gather DMA transfer from the camera to memory.
 * Returns zero if the transfer was successfully queued, or
 * non-zero if all of the scatter-gather slots are already in use.
 */
static int
camera_core_sg_dma_queue(struct camera_device *cam,
        const struct scatterlist *sglist, int sglen, dma_callback_t callback,
        void *arg)
{
        unsigned long irqflags;
        struct sgdma_state *sgdma;
	int count1, count2;

	if ((sglen < 0)  || ((sglen > 0) & !sglist))
                return -EINVAL;

        spin_lock_irqsave(&cam->sg_lock, irqflags);

        if (!cam->free_sgdma) {
                spin_unlock_irqrestore(&cam->sg_lock, irqflags);
                return -EBUSY;
        }

        sgdma = cam->sgdma + cam->next_sgdma;
	sgdma->sglist = sglist;
        sgdma->sglen = sglen;
        sgdma->next_sglist = 0;
        sgdma->queued_sglist = 0;
        sgdma->csr = 0;
        sgdma->callback = callback;
        sgdma->arg = arg;

	count1 = (cam->next_sgdma + 1);
	count2 = count1%34;
        cam->next_sgdma = count2; 
        cam->free_sgdma--;

	spin_unlock_irqrestore(&cam->sg_lock, irqflags);

	camera_core_sg_dma_process(cam);

        return 0;
}


/* -------------------overlay routines ------------------------------*/
/* callback routine for overlay DMA completion. We just start another DMA
 * transfer unless overlay has been turned off
 */

static void
camera_core_overlay_callback(struct camera_device *cam, void *arg)
{
	int err;
	unsigned long irqflags;
        int i, j;
	int count, index;
        unsigned char *fb_buf = phys_to_virt ((unsigned long)camera_dev->fbuf.base);
	spin_lock_irqsave(&cam->overlay_lock, irqflags);
	if(!cam->previewing) {
		spin_unlock_irqrestore(&cam->overlay_lock, irqflags);
		return;
	}


	count = 0;
	j = ((cam->pix.width - 1) * cam->fbuf.fmt.bytesperline);

	for (i = 0 ; i < cam->pix.sizeimage; i+= cam->pix.bytesperline) {
		for (index = 0; index <  cam->pix.bytesperline; index ++) {
			fb_buf[j] = *(((unsigned char *) cam->overlay_base) +
								 i + index);
			index ++;
			fb_buf[j + 1] = *(((unsigned char *) cam->overlay_base) + i + index);
			j = j - cam->fbuf.fmt.bytesperline;
		}
		count += 2;
		j = ((cam->pix.width - 1) * cam->fbuf.fmt.bytesperline) + count;
	}

	if (cam->overlay_cnt > 0)
		--cam->overlay_cnt;
	sg_dma_address(&cam->overlay_sglist) = cam->overlay_base_phys;
	sg_dma_len(&cam->overlay_sglist) = cam->pix.sizeimage;
	while (cam->overlay_cnt < 2) {
		err = camera_core_sg_dma_queue(cam, &cam->overlay_sglist, 1,
			camera_core_overlay_callback, NULL);
		if (err)
			break;
		++cam->overlay_cnt;
	}

	spin_unlock_irqrestore(&cam->overlay_lock, irqflags);
}

 
static void
camera_core_start_overlay_dma(struct camera_device *cam)
{
	int err;
	unsigned long irqflags;

	if (!cam->previewing) 
		return;

	spin_lock_irqsave(&cam->overlay_lock, irqflags);
	sg_dma_address(&cam->overlay_sglist) = cam->overlay_base_phys;
	sg_dma_len(&cam->overlay_sglist)= cam->pix.sizeimage;
	while(cam->overlay_cnt < 2){
		err = camera_core_sg_dma_queue(cam, &cam->overlay_sglist, 1,
				camera_core_overlay_callback, NULL);
		if (err)
			break;
		++cam->overlay_cnt;
	}
	spin_unlock_irqrestore(&cam->overlay_lock, irqflags);
	
}

/* ------------------ videobuf_queue_ops ---------------------------------------- */

/* This routine is called from interrupt context when a scatter-gather DMA
 * transfer of a videobuf_buffer completes.
 */
static void
camera_core_vbq_complete(struct camera_device *cam, void *arg)
{
        struct videobuf_buffer *vb = (struct videobuf_buffer *) arg;

        spin_lock(&cam->vbq_lock);

        do_gettimeofday(&vb->ts);
        vb->field_count = cam->field_count;
        cam->field_count += 2;

        vb->state = STATE_DONE;

        wake_up(&vb->done);
        spin_unlock(&cam->vbq_lock);
}

static void
camera_core_vbq_release(struct file *file, struct videobuf_buffer *vb)
{
        videobuf_waiton(vb, 0, 0);
        videobuf_dma_pci_unmap(NULL, &vb->dma);
        videobuf_dma_free(&vb->dma);

        vb->state = STATE_NEEDS_INIT;
}

/* Limit the number of available kernel image capture buffers based on the
 * number requested, the currently selected image size, and the maximum
 * amount of memory permitted for kernel capture buffers.
 */
static int
camera_core_vbq_setup(struct file *file, unsigned int *cnt, unsigned int *size)
{
        struct camera_fh *fh = file->private_data;
        struct camera_device *cam = fh->cam;

        if (*cnt <= 0)
                *cnt = VIDEO_MAX_FRAME; /* supply a default number of buffers */

        if (*cnt > VIDEO_MAX_FRAME)
                *cnt = VIDEO_MAX_FRAME;

        spin_lock(&cam->img_lock);

        *size = cam->pix.sizeimage;

        spin_unlock(&cam->img_lock);

        while (*size * *cnt > capture_mem)
                (*cnt)--;

        return 0;
}

static int
camera_core_vbq_prepare(struct file *file, struct videobuf_buffer *vb,
        enum v4l2_field field)
{
        struct camera_fh *fh = file->private_data;
        struct camera_device *cam = fh->cam;
        int err = 0;

        spin_lock(&cam->img_lock);

        if (cam->pix.sizeimage > vb->bsize) {
                spin_unlock(&cam->img_lock);
                return -EINVAL;
        }

        vb->size = cam->pix.sizeimage; 
        vb->width = cam->pix.width;
        vb->height = cam->pix.height;
        vb->field = field;

        spin_unlock(&cam->img_lock);

        if (vb->state == STATE_NEEDS_INIT)
                err = videobuf_iolock(NULL, vb, NULL);

        if (!err)
                vb->state = STATE_PREPARED;
        else
                camera_core_vbq_release (file, vb);

        return err;
}

static void
camera_core_vbq_queue(struct file *file, struct videobuf_buffer *vb)
{
        struct camera_fh *fh = file->private_data;
        struct camera_device *cam = fh->cam;
        enum videobuf_state state = vb->state;
        int err;

        vb->state = STATE_QUEUED;
        err = camera_core_sg_dma_queue (cam, vb->dma.sglist, vb->dma.sglen,
                camera_core_vbq_complete, vb);
        if (err) {
                /* Oops.  We're not supposed to get any errors here.  The only
                 * way we could get an error is if we ran out of scatter-gather
                 * DMA slots, but we are supposed to have at least as many
                 * scatter-gather DMA slots as video buffers so that can't
                 * happen.
                 */
                 printk(KERN_DEBUG CAM_NAME
                        ": Failed to queue a video buffer for DMA!\n");
                 vb->state = state;
        }
}

/* ------------------ videobuf_queue_ops ---------------------------------------- */



/* Find the best match for a requested image capture size.  The best match
 * is chosen as the nearest match that has the same number or fewer pixels
 * as the requested size, or the smallest image size if the requested size
 * has fewer pixels than the smallest image.
 */
enum image_size
image_find_size(unsigned int width, unsigned int height)
{
        enum image_size isize;
        unsigned long pixels = width*height;

        for (isize = QQCIF; isize < SXGA; isize++) {
                if (image_sizes[isize + 1].height*
                        image_sizes[isize + 1].width > pixels) {
                        return isize;
                }
        }
        return SXGA;
}

/* Given the image capture format in cam->pix, the nominal frame period in 
 * cam->nominal_timeperframe, and the frequency of the camera interface 
 * functional clock (cam->ocp_clk), calculate the required xclk frequency 
 * (cam->xclk) and the actual frame period (cam->cparm.timeperframe).
 * xclk is obtained by dividing ocp_clk with a divisor in the range 1 to 30.
 * The nominal xclk input frequency of the OV9640 is 24MHz, maximum 
 * frequency is 48MHz, and minimum frequency is 10MHz.
 */
static void
camera_core_sensor_timeperframe(struct camera_device *cam)
{
	unsigned long tgt_xclk;			/* target xclk */
	unsigned long tgt_fps;			/* target frames per minute */
	enum image_size isize;

	/* We use arbitrary rules to select the xclk frequency.  If the 
	 * capture size is VGA and the frame rate is greater than 900 
	 * frames per minute, or if the capture size is SXGA and the 
	 * frame rate is greater than 450 frames per minutes, then the 
	 * xclk frequency will be set to 48MHz.  Otherwise, the xclk 
	 * frequency will be set to 24MHz.  If the ocp_clk frequency 
	 * is such that 
	 * the target xclk frequency is not achievable, then xclk will be set 
	 * as close as to the target as possible.
	 */
	if ((cam->nominal_timeperframe.numerator == 0) 
		|| (cam->nominal_timeperframe.denominator == 0))
	{
		/* supply a default nominal_timeperframe of 15 fps */
		cam->nominal_timeperframe.numerator = 1;
		cam->nominal_timeperframe.denominator = 15;
	}
	tgt_fps = (cam->nominal_timeperframe.denominator)
		/ cam->nominal_timeperframe.numerator;
	tgt_xclk = 24000000;
	isize = image_find_size(cam->pix.width, cam->pix.height);
	switch (isize) {
		case SXGA:
			if( tgt_fps >7)
				tgt_xclk = 48000000;
			break;
		case VGA:
			if (tgt_fps > 15)
				tgt_xclk = 48000000;
			break;

		case CIF:
			if (tgt_fps > 30)
                                tgt_xclk = 24000000;
			break;

		case QVGA:
			if (tgt_fps > 30)
                                tgt_xclk = 24000000;
			break;

		case QCIF:
			if (tgt_fps > 60)
                                tgt_xclk = 24000000;
			break;

		case QQVGA:
			if (tgt_fps > 30)
                                tgt_xclk = 24000000;
			break;

		case QQCIF:
			if (tgt_fps > 60)
                                tgt_xclk = 24000000;
			break;

		default:
			break;
	}

	cam->xclk = tgt_xclk;
	cam->cparm.timeperframe = cam->nominal_timeperframe;
}

static int
camera_core_do_ioctl(struct inode *inode, struct file *file, unsigned int cmd, 
		     void *arg)
{
	struct camera_fh *fh  = file->private_data;
	struct camera_device *cam = fh->cam;
	int err;

	switch (cmd) {
		case VIDIOC_ENUMINPUT:
		{
			/* default handler assumes 1 video input (the camera) */
			struct v4l2_input *input = (struct v4l2_input *) arg;
			int index = input->index;

			memset(input, 0, sizeof(*input));
			input->index = index;

			if (index > 0)
				return -EINVAL;

			strlcpy(input->name, "camera", sizeof(input->name));
			input->type = V4L2_INPUT_TYPE_CAMERA;

			return 0;
		}

		case VIDIOC_G_INPUT:
		{
			unsigned int *input = arg;
			*input = 0;

			return 0;
		}

		case VIDIOC_S_INPUT:
		{
			unsigned int *input = arg;

			if (*input > 0)
				return -EINVAL;

			return 0;
		}

		case VIDIOC_ENUM_FMT:
		{
			struct v4l2_fmtdesc *fmt = arg;
			return cam->cam_sensor->enum_pixformat(fmt);
		}	

		case VIDIOC_TRY_FMT:
		{
			struct v4l2_pix_format *fmt = arg;
			return cam->cam_sensor->try_format(fmt);

		}

		case VIDIOC_G_FMT:
		{
			struct v4l2_format *fmt = arg;

			/* get the current format */
			memset (&fmt->fmt.pix, 0, sizeof (fmt->fmt.pix));
			fmt->fmt.pix = cam->pix;
			
			return 0;
		}

		case VIDIOC_S_FMT:
		{
			struct v4l2_format *fmt = arg;
			unsigned int temp_sizeimage = 0;
			unsigned int clk = 0;

			temp_sizeimage = cam->pix.sizeimage;
			cam->cam_sensor->try_format(&fmt->fmt.pix);
			cam->pix = fmt->fmt.pix;

			camera_core_sensor_timeperframe (cam);
			clk = cam->cam_if->set_xclk(cam->xclk, cam->ocp_clk);

                        return cam->cam_sensor->set_format((struct v4l2_pix_format*)fmt, clk, &cam->cparm.timeperframe);
		}

		case VIDIOC_QUERYCTRL:
		{
			struct v4l2_queryctrl *qc = arg;
                        return cam->cam_sensor->query_control(qc);
		}

		case VIDIOC_G_CTRL:
		{
			struct v4l2_control *vc = arg;
                        return cam->cam_sensor->get_control(vc);
		}

		case VIDIOC_S_CTRL:
		{
			struct v4l2_control *vc = arg;
                        return cam->cam_sensor->set_control (vc);
		}
		
		case VIDIOC_QUERYCAP:
		{
			struct v4l2_capability *cap = 
				(struct v4l2_capability *) arg;

			memset(cap, 0, sizeof(*cap));
                	strlcpy(cap->driver, CAM_NAME, sizeof(cap->driver));
                	strlcpy(cap->card, cam->vfd->name, sizeof(cap->card));
			cap->bus_info[0] = '\0';
			cap->version = KERNEL_VERSION(0, 0, 0);
			cap->capabilities =
				V4L2_CAP_VIDEO_CAPTURE |
				V4L2_CAP_VIDEO_OVERLAY |
				V4L2_CAP_READWRITE | 
				V4L2_CAP_STREAMING;
			return 0;
		}

                case VIDIOC_G_FBUF: /* Get the frame buffer parameters */
                {
                        struct v4l2_framebuffer *fbuf =
                                (struct v4l2_framebuffer *) arg;

                        spin_lock(&cam->img_lock);
                        *fbuf = cam->fbuf;
                        spin_unlock(&cam->img_lock);
                        return 0;
                }

                case VIDIOC_S_FBUF: /* set the frame buffer parameters */
                {
                        struct v4l2_framebuffer *fbuf =
                                (struct v4l2_framebuffer *) arg;

                        spin_lock(&cam->img_lock);
                        if (cam->previewing) {
                                spin_unlock(&cam->img_lock);
                                return -EBUSY;
                        }
			cam->fbuf.base = fbuf->base;
			cam->fbuf.fmt = fbuf->fmt;	
			
                        spin_unlock(&cam->img_lock);
                        return 0;
                }

		case VIDIOC_OVERLAY:
		{
			int enable = *((int *) arg);

			/* 
			 * check whether the capture format and 
			 ** the display format matches 
			 * return failure if they are different
			 */
			if (cam->pix.pixelformat != cam->fbuf.fmt.pixelformat)
			{
			printk("VIDIOC_OVERLAY: Format is not matching\n");
				return -EINVAL;
			}

			/* If the camera image size is greater 
			** than LCD size return failure */
			if ((cam->pix.width > cam->fbuf.fmt.height) || 
				(cam->pix.height > cam->fbuf.fmt.width))
			{
			printk ("VIDIOC_OVERLAY: size of the LCD is less\n");
				return -EINVAL;
			}
			
			if ( !cam->previewing && enable)
			{
				cam->previewing = fh;
				cam->overlay_cnt = 0;
				camera_core_start_overlay_dma(cam);
			}
			else if (!enable)
			{
				cam->previewing = NULL;
			}
	
			return 0;
		}

                case VIDIOC_REQBUFS:
                        return videobuf_reqbufs(file, &fh->vbq, arg);

                case VIDIOC_QUERYBUF:
                        return videobuf_querybuf(&fh->vbq, arg);

                case VIDIOC_QBUF:
                        return videobuf_qbuf(file, &fh->vbq, arg);

                case VIDIOC_DQBUF:
                        return videobuf_dqbuf(file, &fh->vbq, arg);

                case VIDIOC_STREAMON:
                {
                        spin_lock(&cam->img_lock);

                        if (cam->streaming || cam->reading) {
                                spin_unlock(&cam->img_lock);
                                return -EBUSY;
                        }
                        else {
                                cam->streaming = fh;
                                /* FIXME: start camera interface */
                        }

                        spin_unlock(&cam->img_lock);

                        return videobuf_streamon(file, &fh->vbq);
                }
		case VIDIOC_STREAMOFF:
                {
                        err = videobuf_streamoff(file, &fh->vbq);
                        if (err < 0)
                                return err;

                        spin_lock(&cam->img_lock);
                        if (cam->streaming == fh) {
                                cam->streaming = NULL;
                                /* FIXME: stop camera interface */
                        }
                        spin_unlock(&cam->img_lock);
                        return 0;
                }
		case VIDIOC_ENUMSTD:
		case VIDIOC_G_STD:
		case VIDIOC_S_STD:
		case VIDIOC_QUERYSTD:
		{
			/* Digital cameras don't have an analog video standard, 
			 * so we don't need to implement these ioctls.
			 */
			 return -EINVAL;
		}
		case VIDIOC_G_AUDIO:
		case VIDIOC_S_AUDIO:
		case VIDIOC_G_AUDOUT:
		case VIDIOC_S_AUDOUT:
		{
			/* we don't have any audio inputs or outputs */
			return -EINVAL;
		}

		case VIDIOC_G_JPEGCOMP:
		case VIDIOC_S_JPEGCOMP:
		{
			/* JPEG compression is not supported */
			return -EINVAL;
		}

		case VIDIOC_G_TUNER:
		case VIDIOC_S_TUNER:
		case VIDIOC_G_MODULATOR:
		case VIDIOC_S_MODULATOR:
		case VIDIOC_G_FREQUENCY:
		case VIDIOC_S_FREQUENCY:
		{
			/* we don't have a tuner or modulator */
			return -EINVAL;
		}

		case VIDIOC_ENUMOUTPUT:
		case VIDIOC_G_OUTPUT:
		case VIDIOC_S_OUTPUT:
		{
			/* we don't have any video outputs */
			return -EINVAL;
		}

		default:
		{
			/* unrecognized ioctl */
			return -ENOIOCTLCMD;
		}
	}
	return 0;
}

/*
 *  file operations
 */

static unsigned
int camera_core_poll(struct file *file, struct poll_table_struct *wait)
{
	return -EINVAL;
}

/* ------------------------------------------------------------ */
/* callback routine for read DMA completion. We just start another DMA
 * transfer unless overlay has been turned off
 */
static void
camera_core_capture_callback(struct camera_device *cam, void *arg)
{
	int err;
	unsigned long irqflags;
	static int done = 0;

		spin_lock_irqsave(&cam->capture_lock, irqflags);
		if (capture_abort)
		{
			done = 0;
			cam->capture_started = 0;
			spin_unlock_irqrestore(&cam->capture_lock, irqflags);
			return;
		}
	if (done < 14) {
		++done;
		sg_dma_address(&cam->capture_sglist) = cam->capture_base_phys;
		sg_dma_len(&cam->capture_sglist) = cam->pix.sizeimage;
		err = camera_core_sg_dma_queue(cam, &cam->capture_sglist, 1,
			camera_core_capture_callback, NULL);
		
	} else {
		camera_dev->capture_completed = 1;
		if(!capture_abort)
		{

		/* Wake up any process which are waiting for the 
		** DMA to complete */
		wake_up_interruptible (&camera_dev->new_video_frame);
		sg_dma_address(&cam->capture_sglist) = cam->capture_base_phys;
		sg_dma_len(&cam->capture_sglist) = cam->pix.sizeimage;
		err = camera_core_sg_dma_queue(cam, &cam->capture_sglist, 1,
			camera_core_capture_callback, NULL);
		}
	}
		spin_unlock_irqrestore(&cam->capture_lock, irqflags);

}

 
static ssize_t
camera_core_read(struct file *file, char *data, size_t count, loff_t *ppos)
{
	struct camera_fh *fh = file->private_data;
	struct camera_device *cam = fh->cam;
	int err;
	unsigned long irqflags;
	long timeout;

	
	if (!cam->capture_base){
	cam->capture_base = (unsigned long) dma_alloc_coherent(NULL,
				cam->pix.sizeimage,
				(dma_addr_t *) &cam->capture_base_phys,
				GFP_KERNEL | GFP_DMA);
	}
	if (!cam->capture_base){
		printk(KERN_ERR CAM_NAME
			": cannot allocate overlay framebuffer \n");
		return 0;
	}
	spin_lock_irqsave(&cam->capture_lock, irqflags);
	capture_abort = 0;
	cam->capture_started = 1;
	sg_dma_address(&cam->capture_sglist) = cam->capture_base_phys;
	sg_dma_len(&cam->capture_sglist)= cam->pix.sizeimage;
	spin_unlock_irqrestore(&cam->overlay_lock, irqflags);
	err = camera_core_sg_dma_queue(cam, &cam->capture_sglist, 1,
			camera_core_capture_callback, NULL);
	/* Wait till DMA is completed */
	timeout = HZ * 10;
	cam->capture_completed = 0;
	while (cam->capture_completed == 0) {
		timeout = interruptible_sleep_on_timeout 
				(&cam->new_video_frame, timeout);
		if (timeout == 0){
			printk("TIMEOUT");	
			return 0; /* time out */
		}
	}
	/* copy the data to the user buffer */
	err = copy_to_user (data, (void *)cam->capture_base, cam->pix.sizeimage);
	return (cam->pix.sizeimage - err);
	
}

static int
camera_core_mmap(struct file *file, struct vm_area_struct *vma)
{
        struct camera_fh *fh = file->private_data;

        return videobuf_mmap_mapper(vma, &fh->vbq);
}

static int
camera_core_ioctl(struct inode *inode, struct file *file, unsigned int cmd, 
		  unsigned long arg)
{

	return video_usercopy(inode, file, cmd, arg, camera_core_do_ioctl);
}

static int
camera_core_release(struct inode *inode, struct file *file)
{
	struct camera_fh *fh = file->private_data;
	struct camera_device *cam = fh->cam;
	

	file->private_data = NULL;
	kfree(fh);

	spin_lock(&cam->img_lock);
	if (cam->previewing == fh) {
		cam->previewing = NULL;
	}
	spin_unlock(&cam->img_lock);

	capture_abort = 1;
	while(cam->free_dmach < 2)
		mdelay(5);

	if(cam->capture_base) {
		dma_free_coherent(NULL, cam->pix.sizeimage, 
					(void *)cam->capture_base, 
					cam->capture_base_phys);
		cam->capture_base = 0;
		cam->capture_base_phys = 0;
	}
	if (fh->vbq.read_buf) {
		camera_core_vbq_release(file, fh->vbq.read_buf);
		kfree(fh->vbq.read_buf);
	}
	camera_dev->cam_if->close ();
	return 0;
}

static int
camera_core_open(struct inode *inode, struct file *file)
{
	int minor = iminor(inode);
	struct camera_device *cam = camera_dev;
	struct camera_fh *fh;

	if (!cam || !cam->vfd || (cam->vfd->minor != minor))
		return -ENODEV;

	/* allocate per-filehandle data */
	fh = kmalloc(sizeof(*fh), GFP_KERNEL);
	if (NULL == fh)
		return -ENOMEM;
	file->private_data = fh;
	fh->cam = cam;
	fh->type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

	videobuf_queue_init(&fh->vbq, &cam->vbq_ops, NULL, &cam->vbq_lock,
                fh->type, V4L2_FIELD_NONE, sizeof(struct videobuf_buffer));

	cam->capture_completed = 0;
	cam->capture_started = 0;

	cam->cam_if->open (cam);
        /* program the sensor for the default capture format and rate */
        if (cam->cam_sensor->set_format (&cam->pix, cam->xclk, 
				&cam->cparm.timeperframe))
        {
                printk ("\nCamera sensor configuration failed");
                return -ENODEV;
        }


	return 0;
}

static struct file_operations camera_core_fops = 
{
	.owner			= THIS_MODULE,
	.llseek			= no_llseek,
	.read			= camera_core_read,
	.poll			= camera_core_poll,
	.ioctl			= camera_core_ioctl,
	.mmap			= camera_core_mmap,
	.open			= camera_core_open,
	.release		= camera_core_release,
};

void
camera_core_cleanup(void)
{
	struct camera_device *cam = camera_dev;
	struct video_device *vfd;
	if (!cam)
		return;
	vfd = cam->vfd;
	if(vfd){
		if(vfd->minor == -1) {
			/* The device never got registered, so release the 
			** video_device struct directly
			*/
			video_device_release(vfd);
		} else {
			/* The unregister function will release the video_device
			** struct as well as unregistering it.
			*/
			video_unregister_device(vfd);
		}
		cam->vfd = NULL;
	}
	if(cam->overlay_base) {
		dma_free_coherent(NULL, cam->overlay_size,
					(void *)cam->overlay_base, 
					cam->overlay_base_phys);
		cam->overlay_base = 0;
	}	
	cam->overlay_base_phys = 0;
	cam->cam_if->cleanup(cam);
	cam->iobase_phys = 0;
	kfree(cam);
	camera_dev=NULL;

	return;
}


int __init 
camera_core_init(void)
{
	struct camera_device *cam;
	struct video_device *vfd;

	cam = kmalloc(sizeof(struct camera_device), GFP_KERNEL);
	if (!cam) {
		printk(KERN_ERR CAM_NAME ": could not allocate memory\n");
		goto init_error;
	}
	memset(cam, 0, sizeof(struct camera_device));
	
	/* Save the pointer to camera device in a global variable */
	camera_dev = cam;

	/* initialize the video_device struct */
	vfd = cam->vfd = video_device_alloc();
	if (!vfd) {
		printk(KERN_ERR CAM_NAME 
			": could not allocate video device struct\n");
		goto init_error;
	}
	
 	vfd->release = video_device_release;
#if FIXME_FIXME   
/*  If the dev fields are not initlized (parent and such) sysfs will crash you
 *  when it tries to set you up....most drivers seem to chain in the info their 
 *  parent bus gave them.  Getting this associated right is likely necessary 
 * for power mangement.
 */
 	vfd->dev = &cam->dev;
#endif
 	strlcpy(vfd->name, CAM_NAME, sizeof(vfd->name));
 	vfd->type = VID_TYPE_CAPTURE | VID_TYPE_OVERLAY | VID_TYPE_CHROMAKEY;
 	
 	/* need to register for a VID_HARDWARE_* ID in videodev.h */
 	vfd->hardware = 0;
 	vfd->fops = &camera_core_fops;
 	video_set_drvdata(vfd, cam);
 	vfd->minor = -1;

        /* initialize the videobuf queue ops */
        cam->vbq_ops.buf_setup = camera_core_vbq_setup;
        cam->vbq_ops.buf_prepare = camera_core_vbq_prepare;
        cam->vbq_ops.buf_queue = camera_core_vbq_queue;
        cam->vbq_ops.buf_release = camera_core_vbq_release;

	cam->overlay_size = overlay_mem;
	if(cam->overlay_size >0)
	{
		cam->overlay_base = (unsigned long) dma_alloc_coherent(NULL,
					cam->overlay_size,
					(dma_addr_t *) &cam->overlay_base_phys,
					GFP_KERNEL | GFP_DMA);
		if (!cam->overlay_base){
			printk(KERN_ERR CAM_NAME
				": cannot allocate overlay framebuffer \n");
			goto init_error;
		}
	}
	memset ((void*)cam->overlay_base, 0, cam->overlay_size);
	/* initilize the overlay interface */
	spin_lock_init(&cam->overlay_lock);

 	/*Initialise the pointer to the sensor interface and camera interface */
 	cam->cam_sensor = &camera_sensor_if;
 	cam->cam_if = &omap_camera_if;

	/* initialize the camera interface */
	if( (cam->cam_if->init (NULL, cam)) < 0)
		goto init_error;
	
	/* initialise the camera sensor interface */
	cam->cam_sensor->init (&cam->pix);

	/* initialize the spinlock used to serialize access to the image 
	 * parameters
	 */
	spin_lock_init(&cam->img_lock);

	/* initialize the streaming capture parameters */
	cam->cparm.capability = V4L2_CAP_TIMEPERFRAME;

	/* get the input clock value to camera interface and store it */
	cam->ocp_clk= clk_get_rate(clk_get(0, "armper_ck"));

	/* Enable the xclk output.  The sensor may (and does, in the case of 
	 * the OV9640) require an xclk input in order for its initialization 
	 * routine to work.
	 */
	cam->xclk = 21000000;	/* choose an arbitrary xclk frequency */

	/* select an arbitrary default capture frame rate of 15fps */
	cam->nominal_timeperframe.numerator = 1;
	cam->nominal_timeperframe.denominator = 15;
	camera_core_sensor_timeperframe(cam);

	/* disable overlay */
	cam->overlay = 0;

	/* initialise the wait queue */
	init_waitqueue_head (&cam->new_video_frame);

	/* Initialise the DMA structures */
	camera_core_dma_init (cam);

	if (video_register_device(vfd, VFL_TYPE_GRABBER, video_nr) < 0) {
		printk(KERN_ERR CAM_NAME 
			": could not register Video for Linux device\n");
		vfd->minor = -1;
		goto init_error;
	}

	printk(KERN_INFO CAM_NAME 
		": registered device video%d [v4l2]\n", vfd->minor);

	return 0;

init_error:
	camera_core_cleanup();
	return -ENODEV;
}

MODULE_AUTHOR("Texas Instruments.");
MODULE_DESCRIPTION("OMAP Video for Linux camera driver");
MODULE_LICENSE("GPL");
module_param(video_nr, int, 0);
MODULE_PARM_DESC(video_nr, 
		"Minor number for video device (-1 ==> auto assign)");
module_param(capture_mem, int, 0);
MODULE_PARM_DESC(capture_mem,
        "Maximum amount of memory for capture buffers (default 4800KB)");

module_init(camera_core_init);
module_exit(camera_core_cleanup);


