/*
 *	Menelaus Real Time Clock interface for Linux	
 *
 *	Based on drivers/char/omap-rtc.c:
 *
 *	  Copyright (C) 2003 MontaVista Software, Inc.
 *        Author: George G. Davis <gdavis@mvista.com> or <source@mvista.com>
 *
 *	  Initially based on linux-2.4.20/drivers/char/rtc.c
 *	  Copyright (C) 1996 Paul Gortmaker
 *
 *	This driver allows use of the real time clock (built into
 *	nearly all computers) from user space. It exports the /dev/rtc
 *	interface supporting various ioctl() and also the
 *	/proc/driver/rtc pseudo-file for status information.
 *
 *	The ioctls can be used to set the interrupt behaviour from the
 *	RTC via IRQs. Then the /dev/rtc	interface can be used to make
 *	use of RTC interrupts, be they time update or alarm based.
 *
 *	The /dev/rtc interface will block on reads until an interrupt
 *	has been received. If a RTC interrupt has already happened,
 *	it will output an unsigned long and then block. The output value
 *	contains the interrupt status in the low byte and the number of
 *	interrupts since the last read in the remaining high bytes. The 
 *	/dev/rtc interface can also be used with the select(2) call.
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version
 *	2 of the License, or (at your option) any later version.
 *
 *	Based on other minimal char device drivers, like Alan's
 *	watchdog, Ted's random, etc. etc.
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/miscdevice.h>
#include <linux/ioport.h>
#include <linux/fcntl.h>
#include <linux/init.h>
#include <linux/poll.h>
#include <linux/proc_fs.h>
#include <linux/spinlock.h>
#include <linux/rtc.h>

#include <linux/interrupt.h>

#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <asm/hardware.h>
#include <asm/irq.h>
#include <asm/mach/irq.h>
#include <asm/arch/menelaus.h>
#include <asm/arch/sys_info.h>

extern spinlock_t rtc_lock;

/* Local BCD/BIN conversion macros: */
#ifdef BCD_TO_BIN
#undef BCD_TO_BIN
#endif
#define BCD_TO_BIN(val)	((val)=((val)&15) + ((val)>>4)*10)
 
#ifdef BIN_TO_BCD
#undef BIN_TO_BCD
#endif
#define BIN_TO_BCD(val)	((val)=(((val)/10)<<4) + (val)%10)

#define menelaus_read_err(r, reg, act_if_fail) \
	do { \
		if ((*r = menelaus_read(reg)) == -1) { \
			act_if_fail; \
		} \
	} while (0);

#define menelaus_write_err(w, reg, act_if_fail) \
	do { \
		if (menelaus_write(w, reg) == -1) { \
			act_if_fail; \
		} \
	} while (0);

static int rtc_alarm_irq = IH_MENELAUS_BASE + 9;
static int rtc_timer_irq = IH_MENELAUS_BASE + 8;

static struct fasync_struct *rtc_async_queue;

static DECLARE_WAIT_QUEUE_HEAD(rtc_wait);

static ssize_t rtc_read(struct file *file, char *buf,
			size_t count, loff_t *ppos);

static int rtc_ioctl(struct inode *inode, struct file *file,
		     unsigned int cmd, unsigned long arg);

static unsigned int rtc_poll(struct file *file, poll_table *wait);

static int get_rtc_time (struct rtc_time *rtc_tm);
static int get_rtc_alm_time (struct rtc_time *alm_tm);

static int rtc_read_proc(char *page, char **start, off_t off,
                         int count, int *eof, void *data);

/*
 *	Bits in rtc_status. (7 bits of room for future expansion)
 */

#define RTC_IS_OPEN		0x01	/* means /dev/rtc is in use */
#define RTC_AL_EN		0x02    /* means alarm interrupts are enabled */
#define RTC_TIMER_EN		0x03    /* means timer interrupts are enabled */

static unsigned long rtc_freq = 0;	/* Current periodic IRQ rate:
					   bit 0: every second
					   bit 1: every minute
					   bit 2: every hour
					   bit 3: every day */
static unsigned long rtc_status = 0;	/* bitmapped status byte.	*/
static unsigned long rtc_irq_data = 0;	/* our output to the world	*/

/*
 *	If this driver ever becomes modularised, it will be really nice
 *	to make the epoch retain its value across module reload...
 */

static unsigned long epoch = 1900;	/* year corresponding to 0x00	*/

static const unsigned char days_in_mo[] = 
{0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};

static irqreturn_t rtc_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	/*
	 * Alarm or timer interrupt
	 */

	spin_lock (&rtc_lock);

	rtc_irq_data += 0x100;
	rtc_irq_data &= ~0xff;
//	rtc_irq_data |= menelaus_read(OMAP_RTC_STATUS_REG);

	spin_unlock (&rtc_lock);

	/* Now do the rest of the actions */
	wake_up_interruptible(&rtc_wait);	

	kill_fasync (&rtc_async_queue, SIGIO, POLL_IN);
	return IRQ_HANDLED;
}

static ssize_t rtc_read(struct file *file, char *buf,
			size_t count, loff_t *ppos)
{
	DECLARE_WAITQUEUE(wait, current);
	unsigned long data;
	ssize_t retval;
	
	if (count < sizeof(unsigned long))
		return -EINVAL;

	add_wait_queue(&rtc_wait, &wait);
	set_current_state(TASK_INTERRUPTIBLE);

	for (;;) {
		spin_lock_irq (&rtc_lock);
		data = rtc_irq_data;
		if (data != 0) {
			rtc_irq_data = 0;
			break;
		}
		spin_unlock_irq (&rtc_lock);

		if (file->f_flags & O_NONBLOCK) {
			retval = -EAGAIN;
			goto out;
		}
		if (signal_pending(current)) {
			retval = -ERESTARTSYS;
			goto out;
		}
		schedule();
	}

	spin_unlock_irq (&rtc_lock);
	retval = put_user(data, (unsigned long *)buf); 
	if (!retval)
		retval = sizeof(unsigned long); 
 out:
	set_current_state(TASK_RUNNING);
	remove_wait_queue(&rtc_wait, &wait);

	return retval;
}

static int rtc_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
		     unsigned long arg)
{
	struct rtc_time wtime; 

	switch (cmd) {
	case RTC_AIE_OFF:	/* Disable alarm interrupts */
	{
		int en;

		if (rtc_alarm_irq == NO_IRQ)
			return -EINVAL;

		spin_lock_irq (&rtc_lock);
		en = rtc_status & RTC_AL_EN;
		spin_unlock_irq (&rtc_lock);

		if (en)	
			disable_irq(rtc_alarm_irq);
		return 0;
	}
	case RTC_AIE_ON:	/* Enable alarm interrupts */
	{
		int en;

		if (rtc_alarm_irq == NO_IRQ)
			return -EINVAL;

		spin_lock_irq (&rtc_lock);
		en = rtc_status & RTC_AL_EN;
		spin_unlock_irq (&rtc_lock);

		if (!en)	
			enable_irq(rtc_alarm_irq);
		return 0;
	}
	case RTC_ALM_READ:	/* Read the present alarm time */
	{
		/*
		 * This returns a struct rtc_time. Reading >= 0xc0
		 * means "don't care" or "match all". Only the tm_hour,
		 * tm_min, and tm_sec values are filled in.
		 */
		memset(&wtime, 0, sizeof(struct rtc_time));
		if (get_rtc_alm_time(&wtime) == -1)
			return -1;
		break; 
	}
	case RTC_ALM_SET:	/* Store a time into the alarm */
	{
		struct rtc_time alm_tm;
		unsigned char mon, day, hrs, min, sec, leap_yr;
		unsigned int yrs;

		if (copy_from_user(&alm_tm, (struct rtc_time*)arg,
				   sizeof(struct rtc_time)))
			return -EFAULT;

		yrs = alm_tm.tm_year + 1900;
		mon = alm_tm.tm_mon + 1;
		day = alm_tm.tm_mday;
		hrs = alm_tm.tm_hour;
		min = alm_tm.tm_min;
		sec = alm_tm.tm_sec;

		if (yrs < 1970)
			return -EINVAL;

		leap_yr = ((!(yrs % 4) && (yrs % 100)) || !(yrs % 400));

		if ((mon > 12) || (day == 0))
			return -EINVAL;

		if (day > (days_in_mo[mon] + ((mon == 2) && leap_yr)))
			return -EINVAL;
			
		if ((hrs >= 24) || (min >= 60) || (sec >= 60))
			return -EINVAL;

		if ((yrs -= epoch) > 255)    /* They are unsigned */
			return -EINVAL;

		if (yrs > 169) {
			return -EINVAL;
		}

		if (yrs >= 100)
			yrs -= 100;

		BIN_TO_BCD(sec);
		BIN_TO_BCD(min);
		BIN_TO_BCD(hrs);
		BIN_TO_BCD(day);
		BIN_TO_BCD(mon);
		BIN_TO_BCD(yrs);

		spin_lock_irq(&rtc_lock);
		{
			s32 r;
			menelaus_read_err(&r, MENELAUS_RTC_CTRL, return -EIO);
			menelaus_write_err(r | 1<<1, MENELAUS_RTC_CTRL, return -EIO);
			menelaus_write_err(yrs, MENELAUS_RTC_AL_YR, return -EIO);
			menelaus_write_err(mon, MENELAUS_RTC_AL_MON, return -EIO);
			menelaus_write_err(day, MENELAUS_RTC_AL_DAY, return -EIO);
			menelaus_write_err(hrs, MENELAUS_RTC_AL_HR, return -EIO);
			menelaus_write_err(min, MENELAUS_RTC_AL_MIN, return -EIO);
			menelaus_write_err(sec, MENELAUS_RTC_AL_SEC, return -EIO);
		}
		spin_unlock_irq(&rtc_lock);

		return 0;
	}
	case RTC_PIE_OFF:	/* Disable timer interrupts */
	{
		int en;

		if (rtc_timer_irq == NO_IRQ)
			return -EINVAL;

		spin_lock_irq (&rtc_lock);
		en = rtc_status & RTC_TIMER_EN;
		spin_unlock_irq (&rtc_lock);

		if (en)	
			disable_irq(rtc_timer_irq);
		return 0;
	}
	case RTC_PIE_ON:	/* Enable timer interrupts */
	{
		int en;

		if (rtc_timer_irq == NO_IRQ)
			return -EINVAL;

		spin_lock_irq (&rtc_lock);
		en = rtc_status & RTC_AL_EN;
		spin_unlock_irq (&rtc_lock);

		if (!en)	
			enable_irq(rtc_timer_irq);
		return 0;
	}
	case RTC_IRQP_READ:	/* Read the periodic IRQ rate.	*/
	{
		int r;

		if (rtc_timer_irq == NO_IRQ)
			return -EINVAL;

		spin_lock_irq(&rtc_lock);
		r = (rtc_freq == 0) ? 1 : 0;
		spin_unlock_irq(&rtc_lock);
		return put_user(r, (unsigned long __user *)arg);
	}
	case RTC_IRQP_SET:	/* Set periodic IRQ rate.	*/
	{
		if (arg != 1)
			return -EINVAL;

		if (rtc_timer_irq == NO_IRQ)
			return -EINVAL;


		spin_lock_irq(&rtc_lock);
		rtc_freq = 0;
		spin_unlock_irq(&rtc_lock);
		return 0;
	}
	case RTC_RD_TIME:	/* Read the time/date from RTC	*/
	{
		memset(&wtime, 0, sizeof(struct rtc_time));
		if (get_rtc_time(&wtime) == -1)
			return -1;
		break;
	}
	case RTC_SET_TIME:	/* Set the RTC */
	{
		struct rtc_time rtc_tm;
		unsigned char mon, day, hrs, min, sec, leap_yr;
		unsigned int yrs;
		int ret = 0;

		if (!capable(CAP_SYS_TIME))
			return -EACCES;

		if (copy_from_user(&rtc_tm, (struct rtc_time*)arg,
				   sizeof(struct rtc_time)))
			return -EFAULT;

		yrs = rtc_tm.tm_year + 1900;
		mon = rtc_tm.tm_mon + 1;   /* tm_mon starts at zero */
		day = rtc_tm.tm_mday;
		hrs = rtc_tm.tm_hour;
		min = rtc_tm.tm_min;
		sec = rtc_tm.tm_sec;

		if (yrs < 1970)
			return -EINVAL;

		leap_yr = ((!(yrs % 4) && (yrs % 100)) || !(yrs % 400));

		if ((mon > 12) || (day == 0))
			return -EINVAL;

		if (day > (days_in_mo[mon] + ((mon == 2) && leap_yr)))
			return -EINVAL;

		if ((hrs >= 24) || (min >= 60) || (sec >= 60))
			return -EINVAL;

		if ((yrs -= epoch) > 255)    /* They are unsigned */
			return -EINVAL;

		if (yrs > 169) {
			return -EINVAL;
		}

		if (yrs >= 100)
			yrs -= 100;

		BIN_TO_BCD(sec);
		BIN_TO_BCD(min);
		BIN_TO_BCD(hrs);
		BIN_TO_BCD(day);
		BIN_TO_BCD(mon);
		BIN_TO_BCD(yrs);

		spin_lock_irq(&rtc_lock);
		menelaus_write_err(yrs, MENELAUS_RTC_YR, ret = -EIO);
		menelaus_write_err(mon, MENELAUS_RTC_MON, ret = -EIO);
		menelaus_write_err(day, MENELAUS_RTC_DAY, ret = -EIO);
		menelaus_write_err(hrs, MENELAUS_RTC_HR, ret = -EIO);
		menelaus_write_err(min, MENELAUS_RTC_MIN, ret = -EIO);
		menelaus_write_err(sec, MENELAUS_RTC_SEC, ret = -EIO);
		menelaus_write_err(0x8, MENELAUS_RTC_UPDATE, ret = -EIO); /* update everything */
		spin_unlock_irq(&rtc_lock);

		return ret;
	}
	case RTC_EPOCH_READ:	/* Read the epoch.	*/
	{
		return put_user (epoch, (unsigned long *)arg);
	}
	case RTC_EPOCH_SET:	/* Set the epoch.	*/
	{
		/* 
		 * There were no RTC clocks before 1900.
		 */
		if (arg < 1900)
			return -EINVAL;

		if (!capable(CAP_SYS_TIME))
			return -EACCES;

		epoch = arg;
		return 0;
	}
	default:
#if	!defined(CONFIG_ARCH_OMAP)
		return -ENOTTY;
#else
		return -EINVAL;
#endif
	}
	return copy_to_user((void *)arg, &wtime, sizeof wtime) ? -EFAULT : 0;
}

/*
 *	We enforce only one user at a time here with the open/close.
 *	Also clear the previous interrupt data on an open, and clean
 *	up things on a close.
 */

static int rtc_open(struct inode *inode, struct file *file)
{
	spin_lock_irq (&rtc_lock); /* protect against concurrent opens */

	if(rtc_status & RTC_IS_OPEN)
		goto out_busy;

	rtc_status |= RTC_IS_OPEN;

	rtc_irq_data = 0;
	spin_unlock_irq (&rtc_lock);

	/* enable rtc */
	{
		s32 r;
		menelaus_read_err(&r, MENELAUS_RTC_CTRL, goto out_busy);
		menelaus_write_err(r | 1, MENELAUS_RTC_CTRL, goto out_busy);
	}
	return 0;

out_busy:
	spin_unlock_irq (&rtc_lock);
	return -EBUSY;
}

static int rtc_fasync (int fd, struct file *filp, int on)
{
	return fasync_helper (fd, filp, on, &rtc_async_queue);
}

static int rtc_release(struct inode *inode, struct file *file)
{
	if (file->f_flags & FASYNC) {
		rtc_fasync (-1, file, 0);
	}

	spin_lock_irq (&rtc_lock);
	rtc_irq_data = 0;
	spin_unlock_irq (&rtc_lock);

	/*
	 * No need for locking -- nobody else can do anything until this rmw
	 * is committed, and we don't implement timer support in menelaus-rtc.
	 */
	rtc_status &= ~RTC_IS_OPEN;
	return 0;
}

static unsigned int rtc_poll(struct file *file, poll_table *wait)
{
	unsigned long l;

	poll_wait(file, &rtc_wait, wait);

	spin_lock_irq (&rtc_lock);
	l = rtc_irq_data;
	spin_unlock_irq (&rtc_lock);

	if (l != 0)
		return POLLIN | POLLRDNORM;
	return 0;
}

static struct file_operations rtc_fops = {
	owner:		THIS_MODULE,
	llseek:		no_llseek,
	read:		rtc_read,
	poll:		rtc_poll,
	ioctl:		rtc_ioctl,
	open:		rtc_open,
	release:	rtc_release,
	fasync:		rtc_fasync,
};

static struct miscdevice rtc_dev = {
	minor:		RTC_MINOR,
	name:		"menelaus-rtc",
	fops:		&rtc_fops
};

static int __init rtc_init(void)
{
	u32 board = get_board_type();
#if 0
	if (request_irq(rtc_timer_irq, rtc_interrupt, SA_INTERRUPT,
			rtc_dev.name, NULL)) {
		printk(KERN_ERR "%s: RTC timer interrupt IRQ%d is not free.\n",
		       rtc_dev.name, INT_RTC_TIMER);
		return -EIO;
	}
#endif
	if(board != BOARD_H4_MENELAUS)
		return -ENODEV;

	if (request_irq(rtc_alarm_irq, rtc_interrupt, SA_INTERRUPT,
	                "Menelaus RTC alarm", NULL)) {
		rtc_alarm_irq = NO_IRQ;
	}
	disable_irq(rtc_alarm_irq);

	if (request_irq(rtc_timer_irq, rtc_interrupt, SA_INTERRUPT,
	                "Menelaus RTC timer", NULL)) {
		rtc_timer_irq = NO_IRQ;
	}
	disable_irq(rtc_timer_irq);

	if (misc_register(&rtc_dev)) {
		if (rtc_alarm_irq != NO_IRQ)
			free_irq (rtc_alarm_irq, NULL);
		if (rtc_timer_irq != NO_IRQ)
			free_irq (rtc_timer_irq, NULL);
		return -ENODEV;
	}
	if (!create_proc_read_entry ("driver/rtc", 0, NULL, rtc_read_proc, NULL)) {
		if (rtc_alarm_irq != NO_IRQ)
			free_irq (rtc_alarm_irq, NULL);
		if (rtc_timer_irq != NO_IRQ)
			free_irq (rtc_timer_irq, NULL);
		misc_deregister(&rtc_dev);
		return -ENOMEM;
	}
	return 0;
}

static void __exit rtc_exit (void)
{
	s32 r;
	u32 board = get_board_type();

	if (board != BOARD_H4_MENELAUS)
		return;

	menelaus_read_err(&r, MENELAUS_RTC_CTRL, );
	menelaus_write_err(r & ~(1<<1 | 1), MENELAUS_RTC_CTRL, ); /* disable rtc and rtc alarms */

	if (rtc_alarm_irq != NO_IRQ)
		free_irq (rtc_alarm_irq, NULL);
	if (rtc_timer_irq != NO_IRQ)
		free_irq (rtc_timer_irq, NULL);

	remove_proc_entry ("driver/rtc", NULL);
	misc_deregister(&rtc_dev);
}

/*
 *	Info exported via "/proc/driver/rtc".
 */

static int rtc_proc_output (char *buf)
{
#define YN(value) ((value) ? "yes" : "no")
#define NY(value) ((value) ? "no" : "yes")
	char *p;
	struct rtc_time tm;

	p = buf;

	if (get_rtc_time(&tm) == -1)
		return -1;

	/*
	 * There is no way to tell if the luser has the RTC set for local
	 * time or for Universal Standard Time (GMT). Probably local though.
	 */
	p += sprintf(p,
		     "rtc_time\t: %02d:%02d:%02d\n"
		     "rtc_date\t: %04d-%02d-%02d\n"
	 	     "rtc_epoch\t: %04lu\n",
		     tm.tm_hour, tm.tm_min, tm.tm_sec,
		     tm.tm_year + 1900, tm.tm_mon + 1, tm.tm_mday, epoch);

	if (get_rtc_alm_time(&tm) == -1)
		return -1;

	/*
	 * We implicitly assume 24hr mode here. Alarm values >= 0xc0 will
	 * match any value for that particular field. Values that are
	 * greater than a valid time, but less than 0xc0 shouldn't appear.
	 */
	p += sprintf(p,
		     "alarm_time\t: %02d:%02d:%02d\n"
		     "alarm_date\t: %04d-%02d-%02d\n",
		     tm.tm_hour, tm.tm_min, tm.tm_sec,
		     tm.tm_year + 1900, tm.tm_mon + 1, tm.tm_mday);

#if 0
	p += sprintf(p,
		     "BCD\t\t: %s\n"
		     "24hr\t\t: %s\n"
		     "alarm_IRQ\t: %s\n"
		     "update_IRQ\t: %s\n"
		     "update_rate\t: %ud\n",
		     YN(1),
		     YN(1),
		     YN(menelaus_read(OMAP_RTC_INTERRUPTS_REG) &
			OMAP_RTC_INTERRUPTS_IT_ALARM),
		     YN(menelaus_read(OMAP_RTC_INTERRUPTS_REG) &
			OMAP_RTC_INTERRUPTS_IT_TIMER),
		     menelaus_read(OMAP_RTC_INTERRUPTS_REG) & 3 /* REVISIT */);
#endif

	return  p - buf;
#undef YN
#undef NY
}

static int rtc_read_proc(char *page, char **start, off_t off,
                         int count, int *eof, void *data)
{
        int len = rtc_proc_output (page);
        if (len <= off+count) *eof = 1;
        *start = page + off;
        len -= off;
        if (len>count) len = count;
        if (len<0) len = 0;
        return len;
}

static int get_rtc_time(struct rtc_time *rtc_tm)
{
	/* REVISIT: Fix this comment!!!
	 * read RTC once any update in progress is done. The update
	 * can take just over 2ms. We wait 10 to 20ms. There is no need to
	 * to poll-wait (up to 1s - eeccch) for the falling edge of OMAP_RTC_STATUS_BUSY.
	 * If you need to know *exactly* when a second has started, enable
	 * periodic update complete interrupts, (via ioctl) and then 
	 * immediately read /dev/rtc which will block until you get the IRQ.
	 * Once the read clears, read the RTC time (again via ioctl). Easy.
	 */

#if	0 /* REVISIT: This need to do as the TRM says. */
	unsigned long uip_watchdog = jiffies;
	if (rtc_is_updating() != 0)
		while (jiffies - uip_watchdog < 2*HZ/100) {
			barrier();
			cpu_relax();
		}
#endif

	/*
	 * Only the values that we read from the RTC are set. We leave
	 * tm_wday, tm_yday and tm_isdst untouched. Even though the
	 * RTC has RTC_DAY_OF_WEEK, we ignore it, as it is only updated
	 * by the RTC when initially set to a non-zero value.
	 */
	spin_lock_irq(&rtc_lock);
	{
		s32 v;
		menelaus_read_err(&v, MENELAUS_RTC_SEC, goto out_err);
		rtc_tm->tm_sec = v;
		menelaus_read_err(&v, MENELAUS_RTC_MIN, goto out_err);
		rtc_tm->tm_min = v;
		menelaus_read_err(&v, MENELAUS_RTC_HR, goto out_err);
		rtc_tm->tm_hour = v;
		menelaus_read_err(&v, MENELAUS_RTC_DAY, goto out_err);
		rtc_tm->tm_mday = v;
		menelaus_read_err(&v, MENELAUS_RTC_MON, goto out_err);
		rtc_tm->tm_mon = v;
		menelaus_read_err(&v, MENELAUS_RTC_YR, goto out_err);
		rtc_tm->tm_year = v;
	}
	spin_unlock_irq(&rtc_lock);

	BCD_TO_BIN(rtc_tm->tm_sec);
	BCD_TO_BIN(rtc_tm->tm_min);
	BCD_TO_BIN(rtc_tm->tm_hour);
	BCD_TO_BIN(rtc_tm->tm_mday);
	BCD_TO_BIN(rtc_tm->tm_mon);
	BCD_TO_BIN(rtc_tm->tm_year);

	/*
	 * Account for differences between how the RTC uses the values
	 * and how they are defined in a struct rtc_time;
	 */
	if ((rtc_tm->tm_year += (epoch - 1900)) <= 69)
		rtc_tm->tm_year += 100;

	rtc_tm->tm_mon--;
	return 0;
out_err:
	spin_unlock_irq(&rtc_lock);
	return -1;
}

static int get_rtc_alm_time(struct rtc_time *alm_tm)
{
	s32 v;
	spin_lock_irq(&rtc_lock);
	menelaus_read_err(&v, MENELAUS_RTC_AL_SEC, goto out_err);
	alm_tm->tm_sec = v;
	menelaus_read_err(&v, MENELAUS_RTC_AL_MIN, goto out_err);
	alm_tm->tm_min = v;
	menelaus_read_err(&v, MENELAUS_RTC_AL_HR, goto out_err);
	alm_tm->tm_hour = v;
	menelaus_read_err(&v, MENELAUS_RTC_AL_DAY, goto out_err);
	alm_tm->tm_mday = v;
	menelaus_read_err(&v, MENELAUS_RTC_AL_MON, goto out_err);
	alm_tm->tm_mon = v;
	menelaus_read_err(&v, MENELAUS_RTC_AL_YR, goto out_err);
	alm_tm->tm_year = v;
	spin_unlock_irq(&rtc_lock);

	BCD_TO_BIN(alm_tm->tm_sec);
	BCD_TO_BIN(alm_tm->tm_min);
	BCD_TO_BIN(alm_tm->tm_hour);
	BCD_TO_BIN(alm_tm->tm_mday);
	BCD_TO_BIN(alm_tm->tm_mon);
	BCD_TO_BIN(alm_tm->tm_year);

	/*
	 * Account for differences between how the RTC uses the values
	 * and how they are defined in a struct rtc_time;
	 */
	if ((alm_tm->tm_year += (epoch - 1900)) <= 69)
		alm_tm->tm_year += 100;

	alm_tm->tm_mon--;
	return 0;
out_err:
	spin_unlock_irq(&rtc_lock);
	return -1;
}

MODULE_AUTHOR("Texas Instruments, Inc.");
MODULE_LICENSE("GPL");

module_init(rtc_init);
module_exit(rtc_exit);
