/*
 * BRIEF MODULE DESCRIPTION
 *
 *      Infra-red driver for the OMAP24XX Platform
 *          (SIR/MIR/FIR modes)
 *          (based on omap-sir.c)
 *
 * Copyright 2003 MontaVista Software Inc.
 * Author: MontaVista Software, Inc.
 *         source@mvista.com
 *
 * Copyright 2004 Texas Instruments.
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  either version 2 of the  License, or (at your
 *  option) any later version.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 *
 Modifications:
 Feb 2004, Texas Instruments
 - Ported to 2.6 kernel (Feb 2004).
 *
 Apr 2004, Texas Instruments
 - Added support for H3 (Apr 2004).
 *
 May 2004, Texas Instruments
 - Added support for H4 (May 2004).
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/init.h>
#include <linux/errno.h>
#include <linux/netdevice.h>
#include <linux/slab.h>
#include <linux/rtnetlink.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/ioport.h>
#include <linux/dma-mapping.h>

#include <net/irda/irda.h>
#include <net/irda/irmod.h>
#include <net/irda/wrapper.h>
#include <net/irda/irda_device.h>

#include <asm/irq.h>
#include <asm/io.h>
#include <asm/hardware.h>
#include <asm/serial.h>
#include <asm/mach-types.h>
#include <asm/dma.h>
#include <asm/arch/gpio.h>
#include <linux/i2c.h>
#include <asm/arch/mux.h>

#include "omap24xx-ir.h"

#define SIR_MODE 0
#define MIR_MODE 1
#define FIR_MODE 2

/*
 * Our netdevice.  There is only ever one of these.
 */

unsigned long uart3_base;
static struct net_device *netdev;	/* Our net device */
static int rx_state = 0;	/* RX state for IOCTL */

struct omap24xx_irda {
	unsigned char open;
	int speed;		/* Current IrDA speed */
	int newspeed;

	struct net_device_stats stats;
	struct irlap_cb *irlap;
	struct qos_info qos;

	int rx_dma_channel;
	int tx_dma_channel;

	dma_addr_t rx_buf_dma_phys;	/* Physical adress of RX DMA buffer */
	dma_addr_t tx_buf_dma_phys;	/* Physical adress of TX DMA buffer */

	void *rx_buf_dma_virt;	/* Virtual adress of RX DMA buffer */
	void *tx_buf_dma_virt;	/* Virtual adress of TX DMA buffer */

	struct device *dev;

};

/* Functions to read and write to the I/O expanders */
int read_gpio_expa(u8 * val, int addr);
int write_gpio_expa(u8 val, int addr);

/*
 * If you want to disable debug information
 * please uncomment line bellow
 */

//#define OMAP24XX_IR_DEBUG_ENABLE 
#undef OMAP24XX_IR_DEBUG_ENABLE

#ifdef  OMAP24XX_IR_DEBUG_ENABLE
#define __ECHO_IN printk(KERN_ERR "%s: enter\n",__FUNCTION__);
#define __ECHO_OUT printk(KERN_ERR "%s: exit\n",__FUNCTION__);
#define DBG(args...) printk(KERN_ERR __FUNCTION__"():"args);
#else
#define __ECHO_IN
#define __ECHO_OUT
#define DBG(args...)
#endif

#define DBG_IRQ(args...) printk(KERN_ERR __FUNCTION__"():"args);

/* forward declarations */

#define BASE_CLK  48000000

static int omap24xx_irda_set_speed(struct net_device *dev, int speed);

static void omap24xx_irda_start_rx_dma(struct omap24xx_irda *si)
{
	/* Configure DMA */
	omap_set_dma_src_params(si->rx_dma_channel, 0x00, OMAP2420_UART3_BASE,
				0, 0);

	omap_set_dma_dest_params(si->rx_dma_channel, 0x01, si->rx_buf_dma_phys,
				 0, 0);

	omap_set_dma_transfer_params(si->rx_dma_channel, OMAP_DMA_DATA_TYPE_S8,
				     4096, 1, 0x00, OMAP24XX_TRIGGER_RX, 0);

	omap_start_dma(si->rx_dma_channel);
}

static void omap24xx_start_tx_dma(struct omap24xx_irda *si, int size)
{
	omap_set_dma_dest_params(si->tx_dma_channel, 0x0, OMAP2420_UART3_BASE,
				 0, 0);

	omap_set_dma_src_params(si->tx_dma_channel, 0x01, si->tx_buf_dma_phys,
				0, 0);

	omap_set_dma_transfer_params(si->tx_dma_channel, OMAP_DMA_DATA_TYPE_S8,
				     size, 1, 0x00, OMAP24XX_TRIGGER_TX, 0);

	omap_start_dma(si->tx_dma_channel);
}

/* DMA RX callback - normally, we should not go here,
   it calls only if something is going wrong
 */

static
void omap24xx_irda_rx_dma_callback(int lch, u16 ch_status, void *data)
{
	struct net_device *dev = data;
	struct omap24xx_irda *si = dev->priv;

	printk(KERN_ERR "RX Transfer error or very big frame \n");

	/* Clear interrupts */
	readb(UART3_IIR);

	si->stats.rx_frame_errors++;

	readb(UART3_RESUME);

	/* Re-init RX DMA */
	omap24xx_irda_start_rx_dma(si);

}

/* DMA TX callback - calling when frame transfer has been finished */

static
void omap24xx_irda_tx_dma_callback(int lch, u16 ch_status, void *data)
{
	struct net_device *dev = data;
	struct omap24xx_irda *si = dev->priv;

	/*Stop DMA controller */
	omap_stop_dma(si->tx_dma_channel);
}

/*
 * Set the IrDA communications speed.
 * Interrupt have to be disabled here.
 */

static
int omap24xx_irda_startup(struct net_device *dev)
{
	__ECHO_IN;

	writeb(0x07, UART3_MDR1);	/* Put UART3 in reset mode */

	/* Clear DLH and DLL */
	writeb(1 << 7, UART3_LCR);

	writeb(0, UART3_DLL);
	writeb(0, UART3_DLH);

	writeb(0xbf, UART3_LCR);

	writeb(1 << 4, UART3_EFR);

	writeb(1 << 7, UART3_LCR);

	/* Enable access to UART3_TLR and UART3_TCR registers */
	writeb(1 << 6, UART3_MCR);

	writeb(0, UART3_SCR);

	/* Set Rx trigger to 1 and Tx trigger to 1 */
	writeb(0, UART3_TLR);

	/* Set LCR to 8 bits and 1 stop bit */
	writeb(0x03, UART3_LCR);

	/* Clear RX and TX FIFO and enable FIFO */
	/* Use DMA Req for transfers */

	writeb((1 << 2) | (1 << 1) | (1 << 3) | (1 << 4) | (1 << 6) | 1,
	       UART3_FCR);

	writeb(0, UART3_MCR);

	writeb((1 << 7) | (1 << 6), UART3_SCR);

	/* Enable UART3 SIR Mode,(Frame-length method to end frames) */
	writeb(1, UART3_MDR1);

	/* Set Status FIFO trig to 1 */
	writeb(0, UART3_MDR2);

	/* Enables RXIR input */
	/* and disable TX underrun */
	/* SEND_SIP pulse */

	//   writeb((1 << 7) | (1 << 6) | (1 << 4), UART3_ACREG);
	writeb((1 << 6) | (1 << 4), UART3_ACREG);

	/* Enable EOF Interrupt only */
	writeb((1 << 7) | (1 << 5), UART3_IER);

	/* Set Maximum Received Frame size to 2048 bytes */
	writeb(0x00, UART3_RXFLL);
	writeb(0x08, UART3_RXFLH);

	readb(UART3_RESUME);

	__ECHO_OUT;

	return 0;
}

static
int omap24xx_irda_shutdown(struct omap24xx_irda *si)
{

	/* Disable all UART3 Interrupts */
	writeb(0, UART3_IER);

	/* Disable UART3 and disable baud rate generator */
	writeb(0x07, UART3_MDR1);	/* Put UART3 in reset mode */

	writeb((1 << 5), UART3_ACREG);	/* set SD_MODE pin to high and Disable RX IR */

	/* Clear DLH and DLL */
	writeb(1 << 7, UART3_LCR);
	writeb(0, UART3_DLL);
	writeb(0, UART3_DLH);

	return 0;

}

static irqreturn_t
omap24xx_irda_irq(int irq, void *dev_id, struct pt_regs *hw_regs)
{
	struct net_device *dev = dev_id;
	struct omap24xx_irda *si = dev->priv;
	struct sk_buff *skb;

	u8 status;
	int w = 0;

	__ECHO_IN;

	/* Clear EOF interrupt */
	status = readb(UART3_IIR);

	if (status & (1 << 5)) {
		u8 mdr2 = readb(UART3_MDR2);
		if (mdr2 & 1)
			printk(KERN_ERR "IRDA Buffer underrun error\n");

		si->stats.tx_packets++;

		if (si->newspeed) {
			omap24xx_irda_set_speed(dev, si->newspeed);
			si->newspeed = 0;
		}

		netif_wake_queue(dev);

		if (!(status & 0x80))
			return IRQ_HANDLED;
	}

	/* Stop DMA and if there are no errors, send frame to upper layer */

	omap_stop_dma(si->rx_dma_channel);

	status = readb(UART3_SFLSR);	/* Take a frame status */

	if (status != 0) {	/* Bad frame? */
		si->stats.rx_frame_errors++;
		readb(UART3_RESUME);
	} else {
		/* We got a frame! */
		skb = alloc_skb(4096, GFP_ATOMIC);

		if (!skb) {
			printk(KERN_ERR "omap_sir: out of memory for RX SKB\n");
			return IRQ_HANDLED;
		}
		/*
		 * Align any IP headers that may be contained
		 * within the frame.
		 */

		skb_reserve(skb, 1);

		w = readw(OMAP_DMA4_CDAC_REG(si->rx_dma_channel));
		w -= readw(OMAP_DMA4_CDSA_REG(si->rx_dma_channel));
		if (si->speed != 4000000) {
			memcpy(skb_put(skb, w - 2), si->rx_buf_dma_virt, w - 2);	/* Copy DMA buffer to skb */
		} else {
			memcpy(skb_put(skb, w - 4), si->rx_buf_dma_virt, w - 4);	/* Copy DMA buffer to skb */
		}

		skb->dev = dev;
		skb->mac.raw = skb->data;
		skb->protocol = htons(ETH_P_IRDA);
		si->stats.rx_packets++;
		si->stats.rx_bytes += skb->len;
		netif_receive_skb(skb);	/* Send data to upper level */
	}

	/* Re-init RX DMA */
	omap24xx_irda_start_rx_dma(si);

	dev->last_rx = jiffies;

	__ECHO_OUT;

	return IRQ_HANDLED;
}

static
int omap24xx_irda_hard_xmit(struct sk_buff *skb, struct net_device *dev)
{
	struct omap24xx_irda *si = dev->priv;
	int speed = irda_get_next_speed(skb);
	int mtt = irda_get_mtt(skb);
	int xbofs = irda_get_next_xbofs(skb);

	__ECHO_IN;

	/*
	 * Does this packet contain a request to change the interface
	 * speed?  If so, remember it until we complete the transmission
	 * of this frame.
	 */
	if (speed != si->speed && speed != -1)
		si->newspeed = speed;

	if (xbofs) {
		/* Set number of addtional BOFS */
		writeb(xbofs + 1, UART3_EBLR);
	}

	/*
	 * If this is an empty frame, we can bypass a lot.
	 */
	if (skb->len == 0) {
		if (si->newspeed) {
			si->newspeed = 0;
			omap24xx_irda_set_speed(dev, speed);
		}
		dev_kfree_skb(skb);
		return 0;
	}
	netif_stop_queue(dev);

	/* Copy skb data to DMA buffer */
	memcpy(si->tx_buf_dma_virt, skb->data, skb->len);

	si->stats.tx_bytes += skb->len;

	/* Set frame length */
	writeb((skb->len & 0xff), UART3_TXFLL);
	writeb((skb->len >> 8), UART3_TXFLH);

	if (mtt > 1000)
		mdelay(mtt / 1000);
	else
		udelay(mtt);

	/* Start TX DMA transfer */
	omap24xx_start_tx_dma(si, skb->len);

	/* We can free skb now because it's already in DMA buffer */
	dev_kfree_skb(skb);

	dev->trans_start = jiffies;

	__ECHO_OUT;

	return 0;
}

static int
omap24xx_irda_ioctl(struct net_device *dev, struct ifreq *ifreq, int cmd)
{
	struct if_irda_req *rq = (struct if_irda_req *)ifreq;
	struct omap24xx_irda *si = dev->priv;
	int ret = -EOPNOTSUPP;

	__ECHO_IN;

	switch (cmd) {
	case SIOCSBANDWIDTH:
		if (capable(CAP_NET_ADMIN)) {
			/*
			 * We are unable to set the speed if the
			 * device is not running.
			 */
			if (si->open) {
				ret =
				    omap24xx_irda_set_speed(dev,
							    rq->ifr_baudrate);
			} else {
				printk
				    (KERN_ERR
				     "omap_irda_ioctl: SIOCSBANDWIDTH: !netif_running\n");
				ret = 0;
			}
		}
		break;

	case SIOCSMEDIABUSY:
		ret = -EPERM;
		if (capable(CAP_NET_ADMIN)) {
			irda_device_set_media_busy(dev, TRUE);
			ret = 0;
		}
		break;

	case SIOCGRECEIVING:
		rq->ifr_receiving = rx_state;
		break;

	default:
		break;
	}
	__ECHO_OUT;

	return ret;
}

static struct net_device_stats *omap24xx_irda_stats(struct net_device *dev)
{
	struct omap24xx_irda *si = dev->priv;
	return &si->stats;
}

#if 1
static
int omap24xx_irda_setup(void)
{
	omap2_cfg_reg(K15_2420_UART3_TX);
	omap2_cfg_reg(K14_2420_UART3_RX);
#if 0
	reg_val = omap_prcmreg_read(PRCM_FCLK_EN2_CORE);
	omap_prcmreg_write(reg_val | (1 << 2), PRCM_FCLK_EN2_CORE);

	reg_val = omap_prcmreg_read(PRCM_ICLK_EN2_CORE);
	omap_prcmreg_write(reg_val | (1 << 2), PRCM_ICLK_EN2_CORE);
#endif
}
#endif

static
int omap24xx_irda_start(struct net_device *dev)
{
	struct omap24xx_irda *si = dev->priv;
	int err;
	unsigned long flags = 0;
	u8 ioExpanderVal = 0;

	__ECHO_IN;
	si->speed = 9600;

	omap24xx_irda_setup();

	err = request_irq(dev->irq, omap24xx_irda_irq, 0, dev->name, dev);
	if (err)
		goto err_irq;

	/*
	 * The interrupt must remain disabled for now.
	 */

	disable_irq(dev->irq);

	/*  Request DMA channels for IrDA hardware */

	if (omap_request_dma(OMAP24XX_TRIGGER_RX, "IrDA Rx DMA",
			     (void *)omap24xx_irda_rx_dma_callback,
			     dev, &(si->rx_dma_channel))) {
		printk(KERN_ERR "Failed to request IrDA Rx DMA \n");
		goto err_irq;
	}

	if (omap_request_dma(OMAP24XX_TRIGGER_TX, "IrDA Tx DMA",
			     (void *)omap24xx_irda_tx_dma_callback,
			     dev, &(si->tx_dma_channel))) {
		printk(KERN_ERR "Failed to request IrDA Tx DMA \n");
		goto err_irq;
	}

	/* Allocate TX and RX buffers for DMA channels */

	si->rx_buf_dma_virt =
	    dma_alloc_coherent(NULL, 4096, &(si->rx_buf_dma_phys), flags);

	si->tx_buf_dma_virt =
	    dma_alloc_coherent(NULL, 4096, &(si->tx_buf_dma_phys), flags);

	/*
	 * Setup the serial port for the specified config.
	 */

	if ((err = read_gpio_expa(&ioExpanderVal, 0x21))) {
		printk(KERN_ERR ": Error reading from I/O EXPANDER \n");
		return err;
	}

	ioExpanderVal |= 0x01;	/* 'P6' Enable IRDA_TX and IRDA_RX */

	if ((err = write_gpio_expa(ioExpanderVal, 0x21))) {
		printk(KERN_ERR ": Error writing to I/O EXPANDER \n");
		return err;
	}

	if ((err = read_gpio_expa(&ioExpanderVal, 0x21))) {
		printk(KERN_ERR " Error reading from I/O EXPANDER \n");
		return err;
	}

	err = omap24xx_irda_startup(dev);

	if (err)
		goto err_startup;

	omap24xx_irda_set_speed(dev, si->speed = 9600);

	/*
	 * Open a new IrLAP layer instance.
	 */

	si->irlap = irlap_open(dev, &si->qos, "omap_sir");

	err = -ENOMEM;
	if (!si->irlap)
		goto err_irlap;

	/* Now enable the interrupt and start the queue  */
	si->open = 1;
	/* Start RX DMA */

	omap24xx_irda_start_rx_dma(si);

	enable_irq(dev->irq);
	netif_start_queue(dev);

	__ECHO_OUT;

	return 0;

      err_irlap:
	si->open = 0;
	omap24xx_irda_shutdown(si);
      err_startup:
      err_irq:
	free_irq(dev->irq, dev);
	MOD_DEC_USE_COUNT;
	return err;
}

static
int omap24xx_irda_stop(struct net_device *dev)
{
	struct omap24xx_irda *si = dev->priv;

	__ECHO_IN;

	disable_irq(dev->irq);

	netif_stop_queue(dev);

	omap_free_dma(si->rx_dma_channel);
	omap_free_dma(si->tx_dma_channel);

	dma_free_coherent(NULL, 4096, si->rx_buf_dma_virt, si->rx_buf_dma_phys);
	dma_free_coherent(NULL, 4096, si->tx_buf_dma_virt, si->tx_buf_dma_phys);

	omap24xx_irda_shutdown(si);

	/* Stop IrLAP */
	if (si->irlap) {
		irlap_close(si->irlap);
		si->irlap = NULL;
	}

	si->open = 0;

	/*
	 * Free resources
	 */

	free_irq(dev->irq, dev);

	__ECHO_OUT;

	return 0;
}

static void set_h4_gpio_expa(u8 FIR_SEL)
{
	u8 ioExpanderVal = 0;

	if (read_gpio_expa(&ioExpanderVal, 0x20) != 0) {
		printk("Error reading from I/O EXPANDER \n");
		return;
	}

	ioExpanderVal &= ~0x01;
	ioExpanderVal |= FIR_SEL;

	if (write_gpio_expa(ioExpanderVal, 0x20) != 0) {
		printk("Error writing to I/O EXPANDER \n");
		return;
	}
	if (read_gpio_expa(&ioExpanderVal, 0x20) != 0) {
		printk("Error reading from I/O EXPANDER \n");
		return;
	}
}

static int which_speed;

static void set_h4_gpio_expa_handler(void *data)
{
	int *mode = data;

	if (*mode == SIR_MODE)
		set_h4_gpio_expa(0);
	else if (*mode == MIR_MODE)
		set_h4_gpio_expa(1);
	else if (*mode == FIR_MODE)
		set_h4_gpio_expa(1);
}

DECLARE_WORK(set_h4_gpio_expa_work, &set_h4_gpio_expa_handler, &which_speed);

static
int omap24xx_irda_set_speed(struct net_device *dev, int speed)
{
	struct omap24xx_irda *si = dev->priv;
	int divisor;

	__ECHO_IN;

	/* Set IrDA speed */
	if (speed <= 115200) {

		if (in_interrupt()) {
			cancel_delayed_work(&set_h4_gpio_expa_work);
			which_speed = SIR_MODE;
			schedule_work(&set_h4_gpio_expa_work);
		} else
			set_h4_gpio_expa(0);

		/* SIR mode */
		printk("Set SIR Mode! Speed: %d\n", speed);

		writeb(1, UART3_MDR1);	/* Set SIR mode */

		writeb(1, UART3_EBLR);

		divisor = BASE_CLK / (16 * speed);	/* Base clock 48 MHz */

		writeb(1 << 7, UART3_LCR);

		writeb((divisor & 0xFF), UART3_DLL);

		writeb((divisor >> 8), UART3_DLH);

		writeb(0x03, UART3_LCR);

		writeb(0, UART3_MCR);

	} else if (speed <= 1152000) {

		printk("Set MIR Mode! Speed: %d\n", speed);

		writeb((1 << 2) | (1 << 6), UART3_MDR1);	/* Set MIR mode with
								   SIP after each frame */
		writeb(2, UART3_EBLR);

		divisor = BASE_CLK / (41 * speed);

		writeb(1 << 7, UART3_LCR);

		writeb((divisor & 0xFF), UART3_DLL);

		writeb((divisor >> 8), UART3_DLH);

		writeb(0x03, UART3_LCR);

		if (in_interrupt()) {
			cancel_delayed_work(&set_h4_gpio_expa_work);
			which_speed = MIR_MODE;
			schedule_work(&set_h4_gpio_expa_work);
		} else
			set_h4_gpio_expa(1);

	} else {
		/* FIR mode */

		if (in_interrupt()) {
			cancel_delayed_work(&set_h4_gpio_expa_work);
			which_speed = MIR_MODE;
			schedule_work(&set_h4_gpio_expa_work);
		} else
			set_h4_gpio_expa(1);

		printk("Set FIR Mode! Speed: %d\n", speed);
		writeb((1 << 2) | (1 << 6) | 1, UART3_MDR1);	/* Set FIR mode
								   with SIP after each frame */
	}

	si->speed = speed;

	__ECHO_OUT;

	return 0;
}

static
int omap24xx_irda_net_init(struct net_device *dev)
{
	struct omap24xx_irda *si = dev->priv;
	unsigned int baudrate_mask;
	int err = -ENOMEM;

	si = kmalloc(sizeof(struct omap24xx_irda), GFP_KERNEL);
	if (!si)
		goto out;

	memset(si, 0, sizeof(*si));

	/*
	 * Initialize structures
	 */
	dev->priv = si;
	dev->hard_start_xmit = omap24xx_irda_hard_xmit;
	dev->open = omap24xx_irda_start;
	dev->stop = omap24xx_irda_stop;
	dev->do_ioctl = omap24xx_irda_ioctl;
	dev->get_stats = omap24xx_irda_stats;

	irda_device_setup(dev);
	irda_init_max_qos_capabilies(&si->qos);

	baudrate_mask = IR_9600 | IR_19200 | IR_38400 | IR_57600 | IR_115200;

	baudrate_mask |= (IR_4000000 << 8);

	si->qos.baud_rate.bits &= baudrate_mask;
	si->qos.min_turn_time.bits = 7;

	uart3_base =
	    (unsigned long)ioremap(OMAP2420_UART3_BASE, OMAP24XX_REG_SIZE);

	irda_qos_bits_to_value(&si->qos);
	return 0;
      out:
	kfree(si);

	return err;
}

/*
 * Remove all traces of this driver module from the kernel, so we can't be
 * called.  Note that the device has already been stopped, so we don't have
 * to worry about interrupts or dma.
 */

static void omap24xx_irda_net_uninit(struct net_device *dev)
{
	struct omap24xx_irda *si = dev->priv;

	dev->hard_start_xmit = NULL;
	dev->open = NULL;
	dev->stop = NULL;
	dev->do_ioctl = NULL;
	dev->get_stats = NULL;
	dev->priv = NULL;

	kfree(si);
}

#ifdef MODULE
static
#endif
int __init omap24xx_irda_init(void)
{
	struct net_device *dev;
	int err = 0;

	/* OMAP24xx IR  net device register stage */
	dev = alloc_irdadev(sizeof(struct omap24xx_irda));

	if (dev) {
		dev->irq = INT_IRDA;
		dev->init = omap24xx_irda_net_init;
		dev->uninit = omap24xx_irda_net_uninit;

		err = register_netdev(dev);
		if (err) {
			kfree(dev);
		} else {
			netdev = dev;
		}
	}
	return err;
}

#ifdef MODULE
static
#endif
void __exit omap24xx_irda_exit(void)
{
	struct net_device *dev = netdev;

	if (dev) {
		unregister_netdev(dev);
		free_netdev(dev);
	}

	/*
	 * We now know that the netdevice is no longer in use, and all
	 * references to our driver have been removed.  The only structure
	 * which may still be present is the netdevice, which will get
	 * cleaned up by net/core/dev.c
	 */
}

int write_gpio_expa(u8 val, int addr)
{
	struct i2c_adapter *adap;
	int err;
	struct i2c_msg msg[1];
	unsigned char data[1];

	adap = i2c_get_adapter(0);
	if (!adap)
		return -ENODEV;
	msg->addr = addr;	/* I2C address of GPIO EXPA */
	msg->flags = 0;
	msg->len = 1;
	msg->buf = data;
	data[0] = val;
	err = i2c_transfer(adap, msg, 1);
	if (err >= 0)
		return 0;
	return err;
}

int read_gpio_expa(u8 * val, int addr)
{
	struct i2c_adapter *adap;
	int err;
	struct i2c_msg msg[1];
	unsigned char data[1];

	adap = i2c_get_adapter(0);
	if (!adap)
		return -ENODEV;
	msg->addr = addr;	/* I2C address of GPIO EXPA */
	msg->flags = I2C_M_RD;
	msg->len = 2;
	msg->buf = data;
	err = i2c_transfer(adap, msg, 1);
	*val = data[0];
	printk("IRDA: I2C: Read data is %x", (u8) * data);
	if (err >= 0)
		return 0;
	return err;
}

module_init(omap24xx_irda_init);
module_exit(omap24xx_irda_exit);

MODULE_AUTHOR("MontaVista");
MODULE_DESCRIPTION("OMAP24XX IR");
MODULE_LICENSE("GPL");
