/*
 *  linux/drivers/serial/8250_omap.c
 *  Partially copied from 8250_acorn.c
 *
 *  Copyright (C) 1996-2003 Russell King.
 *  Copyright (C) 2004 Nokia Corporation.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */
#include <linux/module.h>
#include <linux/types.h>
#include <linux/tty.h>
#include <linux/serial_core.h>
#include <linux/serial.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/slab.h>
#include <linux/device.h>
#include <linux/init.h>

#include <asm/io.h>
#include <asm/arch/serial.h>
#include <asm/string.h>
#include <asm/arch/cpu.h>

#define UART_SYSC       0x15

#define MAX_PORTS       3

struct omap_serial_port {
	unsigned long baddr;
	unsigned int irq;
};

#ifdef CONFIG_ARCH_OMAP24XX
static const struct omap_serial_port omap2420_serial_ports[] = {
	{ .baddr = (unsigned long)IO_ADDRESS(OMAP24XX_UART1_BASE), .irq = INT_UART1_IRQ },
	{ .baddr = (unsigned long)IO_ADDRESS(OMAP24XX_UART2_BASE), .irq = INT_UART2_IRQ },
	{ .baddr = (unsigned long)IO_ADDRESS(OMAP24XX_UART3_BASE), .irq = INT_UART3_IRQ },
};
#else
static const struct omap_serial_port omap1510_serial_ports[] = {
	{ .baddr = (unsigned long)IO_ADDRESS(OMAP_UART1_BASE), .irq = INT_UART1 },
	{ .baddr = (unsigned long)IO_ADDRESS(OMAP_UART2_BASE), .irq = INT_UART2 },
	{ .baddr = (unsigned long)IO_ADDRESS(OMAP_UART3_BASE), .irq = INT_UART3 },
};

static const struct omap_serial_port omap730_serial_ports[] = {
	{ .baddr = (unsigned long)IO_ADDRESS(OMAP_UART1_BASE), .irq = INT_730_UART_MODEM_1 },
	{ .baddr = (unsigned long)IO_ADDRESS(OMAP_UART2_BASE), .irq = INT_730_UART_MODEM_IRDA_2 },
};
#endif

static void
reset_port(struct serial_struct *port)
{
	if (cpu_is_omap1510() || cpu_is_omap2420())
		return;
	writeb(0x01, port->iomap_base + (UART_SYSC << port->iomem_reg_shift));
	while (!(readb(port->iomap_base + (UART_SYSC << port->iomem_reg_shift)) & 0x01));
}

static inline int
serial_register_onedev(const struct omap_serial_port *port)
{
	struct serial_struct req;

	memset(&req, 0, sizeof(req));
	req.type                = PORT_OMAP;
	req.irq			= port->irq;
	req.flags		= UPF_SKIP_TEST;
	req.io_type		= UPIO_MEM;
	req.iomem_base		= (unsigned char *) port->baddr;
	req.iomap_base		= port->baddr;
#ifndef CONFIG_ARCH_OMAP24XX
	if (cpu_is_omap1510()) {
		req.baud_base		= OMAP1510_BASE_BAUD;
		req.iomem_reg_shift	= 2;
	}
	if (cpu_is_omap16xx()) {
		req.baud_base		= OMAP1610_BASE_BAUD;
		req.iomem_reg_shift	= 2;
	}
	if (cpu_is_omap730()) {
		req.baud_base		= OMAP1610_BASE_BAUD;
		req.iomem_reg_shift	= 0;
	}
#else
	if (cpu_is_omap2420()) {
		req.baud_base		= OMAP2420_BASE_BAUD;
		req.iomem_reg_shift	= 2;
	}
#endif
        reset_port(&req);

	return register_serial(&req);
}

static int port_count = 0;
static int ports[MAX_PORTS];

static int __init omap_serial_init(void)
{
	int i;
	const struct omap_serial_port *port_table = NULL;
#ifdef CONFIG_ARCH_OMAP24XX	
	if (cpu_is_omap2420()) {
		port_count = 3;
		port_table = omap2420_serial_ports;
	}
#else
	if (cpu_is_omap1510() || cpu_is_omap16xx()) {
		//Michael:setting for wifi phone
		//port_count = 3;
		port_count = 1;
		port_table = omap1510_serial_ports;
	}
	if (cpu_is_omap730()) {
		port_count = 2;
		port_table = omap730_serial_ports;
	}
#endif
	for (i = 0; i < port_count; i++) {
		ports[i] = serial_register_onedev(&port_table[i]);
	}
	printk(KERN_INFO "OMAP serial: activated %d ports\n", i);

	return 0;
}

static void __exit omap_serial_exit(void)
{
	int i;

	for (i = 0; i < port_count; i++)
		unregister_serial(ports[i]);
}

MODULE_AUTHOR("Juha Yrjl");
MODULE_DESCRIPTION("OMAP 8250-compatible serial port driver");
MODULE_LICENSE("GPL");

module_init(omap_serial_init);
module_exit(omap_serial_exit);
