/*
 * BRIEF MODULE DESCRIPTION
 *
 *	uWire interface driver for the OMAP Platform
 *
 * Copyright 2003 MontaVista Software Inc.
 * Author: MontaVista Software, Inc.
 *	   source@mvista.com
 *
 * Ported to 2.6 uwire interface.
 * Copyright (C) 2004 Texas Instruments.
 *
 * Generalization patches by Juha Yrjl <juha.yrjola@nokia.com>
 *
 *  This program is free software; you can redistribute	 it and/or modify it
 *  under  the terms of	 the GNU General  Public License as published by the
 *  Free Software Foundation;  either version 2 of the	License, or (at your
 *  option) any later version.
 *
 *  THIS  SOFTWARE  IS PROVIDED	  ``AS	IS'' AND   ANY	EXPRESS OR IMPLIED
 *  WARRANTIES,	  INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO	EVENT  SHALL   THE AUTHOR  BE	 LIABLE FOR ANY	  DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED	  TO, PROCUREMENT OF  SUBSTITUTE GOODS	OR SERVICES; LOSS OF
 *  USE, DATA,	OR PROFITS; OR	BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN	 CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/delay.h>

#include <asm/system.h>
#include <asm/irq.h>
#include <asm/hardware.h>
#include <asm/io.h>
#include <asm/mach-types.h>
#include <asm/arch/mux.h>
#include <asm/semaphore.h>

/* uWire Registers: */

#define UWIRE_BASE    0xFFFB3000
#define UWIRE_IO_SIZE 0x20
#define UWIRE_TDR     UWIRE_BASE + 0x00
#define UWIRE_RDR     UWIRE_BASE + 0x00
#define UWIRE_CSR     UWIRE_BASE + 0x04
#define UWIRE_SR1     UWIRE_BASE + 0x08
#define UWIRE_SR2     UWIRE_BASE + 0x0C
#define UWIRE_SR3     UWIRE_BASE + 0x10
#define UWIRE_SR4     UWIRE_BASE + 0x14
#define UWIRE_SR5     UWIRE_BASE + 0x18

/* General OMAP uWire functions */

void
omap_uwire_configure_mode(u8 cs, u8 edge_rd, u8 edge_wr, u8 lvl, u8 frq, u8 chk)
{
	switch (cs) {
	case 0:
		omap_writew((omap_readw(UWIRE_SR1) & 0xFFC0) | (chk << 5) |
			    (frq << 3) | (lvl << 2) | (edge_wr << 1) | edge_rd,
			    UWIRE_SR1);
		break;
	case 1:
		omap_writew((omap_readw(UWIRE_SR1) & 0xF03F) | (chk << 11) |
			    (frq << 9) | (lvl << 8) | (edge_wr << 7) | (edge_rd
									<< 6),
			    UWIRE_SR1);
		break;
	case 2:
		omap_writew((omap_readw(UWIRE_SR2) & 0xFFC0) | (chk << 5) |
			    (frq << 3) | (lvl << 2) | (edge_wr << 1) | edge_rd,
			    UWIRE_SR2);
		break;
	case 3:
		omap_writew((omap_readw(UWIRE_SR2) & 0xF03F) | (chk << 11) |
			    (frq << 9) | (lvl << 8) | (edge_wr << 7) | (edge_rd
									<< 6),
			    UWIRE_SR2);
		break;
	default:
		BUG();
		break;
	}
}

static DECLARE_MUTEX(uwire_lock);

static int wait_uwire_csr_flag(u16 mask, u16 val, int might_not_catch)
{
	u16 w;
	int c = 0;
	unsigned long max_jiffies = jiffies + HZ;

	for (;;) {
		w = omap_readw(UWIRE_CSR);
		if ((w & mask) == val)
			break;
		if (time_after(jiffies, max_jiffies)) {
			printk(KERN_ERR
			       "%s: timeout. reg=%#06x mask=%#06x val=%#06x\n",
			       __FUNCTION__, w, mask, val);
			return -1;
		}
		c++;
		if (might_not_catch && c > 64)
			break;
	}
	return 0;
}

int
omap_uwire_data_transfer(u8 cs, u16 data, u8 trans_size, u8 rec_size,
			 u16 * rec_buf)
{
	u8 leavecs = cs & 0x80;
	u16 ret = -1, w;
	u16 mask;

	cs = cs & 3;

	if (rec_size && !rec_buf)
		goto exit;

	down(&uwire_lock);

	if (wait_uwire_csr_flag(1 << 14, 0, 0))
		goto exit;

	w = cs << 10;
	w |= 1 << 12;			/* CS_CMD : activate CS */
	omap_writew(w, UWIRE_CSR);

	omap_writew(data, UWIRE_TDR);	/* Place data in TX register */

	if (wait_uwire_csr_flag(1 << 14, 0, 0))
		goto exit;

	w = rec_size | (trans_size << 5) | (cs << 10);
	w |= (1 << 12) | (1 << 13);
	/* Start uWire read/write */
	omap_writew(w, UWIRE_CSR);

	/* Wait till read/write actually starts.
	 * This is needed at high (>=60MHz) MPU frequencies
	 * REVISIT: But occasionally we won't have time to catch it
	 */
	if (wait_uwire_csr_flag(1 << 14, 1 << 14, 1))
		goto exit;

	/* Wait for both transfers to be completed */
	mask = 1 << 14;		/* CSRB : reg busy */
	w = 0;
	if (rec_size) {
		mask |= 1 << 15;	/* RDRB : reg busy */
		w |= 1 << 15;
	}

	if (wait_uwire_csr_flag(mask, w, 0))
		goto exit;

	if (rec_size)
		*rec_buf = omap_readw(UWIRE_RDR);

	if (!leavecs)
		omap_writew(cs << 10, UWIRE_CSR);

	ret = 0;

      exit:
	up(&uwire_lock);
	return ret;
}

static int __init omap_uwire_init(void)
{
	omap_writew(1, UWIRE_SR3);
	if (machine_is_omap_h2() || machine_is_omap_osk()) {
		/* defaults: W21 SDO, U18 SDI, V19 SCL */
		omap_cfg_reg(N14_1610_UWIRE_CS0);
		omap_cfg_reg(N15_1610_UWIRE_CS1);
	}
	return 0;
}

static void __exit omap_uwire_exit(void)
{
}

subsys_initcall(omap_uwire_init);
module_exit(omap_uwire_exit);

EXPORT_SYMBOL(omap_uwire_configure_mode);
EXPORT_SYMBOL(omap_uwire_data_transfer);

MODULE_LICENSE("GPL");
