/*
 * drivers/video/omap24xxfb.c
 *
 * Framebuffer driver for OMAP24xx display controller.
 *
 * Author: Andy Lowe (source@mvista.com)
 *
 * Copyright (C) 2004 MontaVista Software, Inc.
 * Copyright (C) 2004 Texas Instruments.
 *
 * This file is licensed under the terms of the GNU General Public License 
 * version 2. This program is licensed "as is" without any warranty of any 
 * kind, whether express or implied.
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/tty.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/fb.h>
#include <linux/ioport.h>
#include <linux/types.h>
#include <linux/dma-mapping.h>
#include <linux/init.h>
#include <linux/smp_lock.h>
#include <linux/interrupt.h>
#include <asm/irq.h>

#include "omap24xxfb.h"

/******************************************************************************/
/* Board-specific customization for the framebuffer driver.
 * This stuff needs to be moved to board-specific files.
 */
#define CONFIG_H4
#ifdef CONFIG_H4

//#define H4_CLK_PERIOD (20000)  /* f_pico(1/50MHz)=20000 for H4 at present */
#define H4_CLK_PERIOD (26667)  /* f_pico(1/37.5MHz)=26667 for H4 at present */

#define H4_XRES 240
#define H4_YRES 320

/* omap24xxfb_gfx_clk_period() must return the period (in picoseconds) of the 
 * graphics timebase clock.  This clock is the input to the pixel clock 
 * divider.
 *
 * prototype:
 * unsigned long omap24xxfb_gfx_clk_period(void);
 */
#define omap24xxfb_gfx_clk_period() H4_CLK_PERIOD

/* omap24xxfb_fb_base() must return the physical base address of the 
 * framebuffer.  If the address is non-zero, then the framebuffer memory is 
 * ioremap'd.  If the address is zero, then the framebuffer memory is 
 * dynamically allocated by the driver.
 *
 * prototype:
 * unsigned long omap24xxfb_fb_base(void);
 */
#define omap24xxfb_fb_base() 0

/* omap24xxfb_fb_size() must return the size of the framebuffer in bytes.
 * The framebuffer is only ioremap'd (or kmalloc'd) at driver initialization 
 * time.  It does not change when the video mode (resolution and color depth) 
 * changes, so it must be large enough for all video modes that are to be 
 * supported.
 *
 * We're allocating a framebuffer 3 times the size of the physical display.  
 * This is to support triple buffering.  The panning ioctl can be used to 
 * switch between the three different framebuffer regions, so you effectively 
 * have an onscreen framebuffer and two offscreen framebuffers.
 *
 * prototype:
 * unsigned long omap24xxfb_fb_size(void);
 */
#define omap24xxfb_fb_size() (H4_XRES*H4_YRES*(16/8)*3) /* 450KB */

/* omap24xxfb_display_xres() must return the horizontal resolution of the 
 * display.
 *
 * prototype:
 * unsigned int omap24xxfb_display_xres(void);
 */
#define omap24xxfb_display_xres() (H4_XRES)

/* omap24xxfb_display_yres() must return the vertical resolution of the 
 * display.
 *
 * prototype:
 * unsigned int omap24xxfb_display_yres(void);
 */
#define omap24xxfb_display_yres() (H4_YRES)

/* omap24xxfb_default_var() must return a pointer to a default 
 * fb_var_screeninfo struct that will be used to set the initial video mode.  
 * If this video mode is invalid (as determined by omap24xxfb_check_var) then 
 * the driver will fail to initialize.
 *
 * We're intializing the virtual framebuffer dimensions to 
 * (H4_XRES, H4_YRES*3) in order to support triple buffering.  The 
 * onscreen framebuffer can be flipped via the panning ioctl by specifying 
 * offsets of (0, 0), (0, H4_YRES), or (0, 2*H4_YRES).
 *
 * prototype:
 * struct fb_var_screeninfo *omap24xxfb_default_var(void);
 */
static struct fb_var_screeninfo aptix_h4_var = {
	.xres		= H4_XRES,
	.yres		= H4_YRES,
	.xres_virtual	= H4_XRES,
	.yres_virtual	= H4_YRES*3,
	.xoffset	= 0,
	.yoffset	= 0,
	.bits_per_pixel	= 16,
	.grayscale	= 0,
	.red		= {11, 5, 0},
	.green		= { 5, 6, 0},
	.blue		= { 0, 5, 0},
	.transp		= { 0, 0, 0},
	.nonstd		= 0,
	.activate	= 0,
	.height		= -1,
	.width		= -1,
	.accel_flags	= 0,
	.pixclock	= H4_CLK_PERIOD*8, /* picoseconds */
	.left_margin	= 40,		/* pixclocks */
	.right_margin	= 4,		/* pixclocks */
	.upper_margin	= 8,		/* line clocks */
	.lower_margin	= 2,		/* line clocks */
	.hsync_len	= 4,		/* pixclocks */
	.vsync_len	= 2,		/* line clocks */
	.sync		= 0,
	.vmode		= FB_VMODE_NONINTERLACED,
	.rotate		= 0,
};
#define omap24xxfb_default_var() (&aptix_h4_var)

/* omap24xxfb_check_mode() must check the video mode specified in a 
 * fb_var_screeninfo struct and return 0 if the mode is supported and non-zero 
 * if the mode is not supported.  omap24xxfb_check_mode() is permitted to 
 * modify the var to make it conform to a supported mode.
 *
 * prototype:
 * int omap24xxfb_check_mode(struct fb_var_screeninfo *var);
 */
int omap24xxfb_check_mode(struct fb_var_screeninfo *var)
{
	if ((var->xres > H4_XRES) || (var->yres > H4_YRES))
		return -EINVAL;
	if (var->bits_per_pixel > 16)
		return -EINVAL;
	return 0;
}


/* omap24xxfb_change_mode() must make any changes to a shadow copy of the 
 * display controller registers that are required to change to the new video 
 * mode specified in var.  The values in var must not be modified.
 *
 * prototype:
 * void omap24xxfb_change_mode(const struct fb_var_screeninfo *var, 
 * 			       struct omap24xx_dispc_regs *dispc);
 */
void omap24xxfb_change_mode(const struct fb_var_screeninfo *var, 
			    struct omap24xx_dispc_regs *dispc)
{
	dispc->sysconfig = DISPC_SYSCONFIG_MIDLEMODE_NSTANDBY
			 | DISPC_SYSCONFIG_SIDLEMODE_NIDLE;
	dispc->control = DISPC_CONTROL_GPOUT1 | DISPC_CONTROL_GPOUT0 
			| DISPC_CONTROL_TFTDATALINES_OALSB16B 
			| DISPC_CONTROL_STNTFT 
			| DISPC_CONTROL_LCDENABLE;
	dispc->config = 0;
	dispc->gfx_attributes = DISPC_GFX_ATTRIBUTES_GFXBURSTSIZE_BURST8X32 
			      | DISPC_GFX_ATTRIBUTES_ENABLE;
	dispc->pol_freq = 0;
	dispc->gfx_fifo_threshold = (252 << DISPC_GFX_FIFO_THRESHOLD_HIGH_SHIFT)
				  | (192 << DISPC_GFX_FIFO_THRESHOLD_LOW_SHIFT);
	dispc->gfx_position = 0;
	dispc->gfx_window_skip = 0;
	return;
}

#endif /* ifdef CONFIG_H4 */
/******************************************************************************/

#define FB_NAME "omap24xxfb"		/* 16 chars max */

struct omap24xxfb_info {
	/* The fb_info structure must be first! */
	struct fb_info info;

	/* horizontal and vertical resolution of display */
	unsigned int display_xres;
	unsigned int display_yres;

	struct omap24xx_dispc_regs state;
	
	dma_addr_t mmio_base_phys;
	dma_addr_t fb_base_phys;
	unsigned long fb_size;
	unsigned long mmio_base;
	unsigned long fb_base;

	wait_queue_head_t vsync_wait;
	unsigned long vsync_cnt;

	u32 pseudo_palette[17];
	u32 *palette;
	dma_addr_t palette_phys;

	/* Period of the graphics clock in picoseconds.
	 * This is is the input to the pixel clock divider.
	 */
	unsigned long gfx_clk_period;
	
	unsigned int hsync;	/* horizontal sync frequency in Hz */
	unsigned int vsync;	/* vertical sync frequency in Hz */
	unsigned long timeout;	/* register update timeout period in ticks */

	int alloc_fb_mem;
	
	int asleep;
};

static struct omap24xxfb_info *saved_oinfo;

/*
 * display controller register I/O routines
 */

static __inline__ u32 
dispc_reg_in(const struct omap24xxfb_info *oinfo, u32 offset)
{
	return readl(oinfo->mmio_base + DISPC_REG_OFFSET + offset);
}

static __inline__ u32 
dispc_reg_out(const struct omap24xxfb_info *oinfo, u32 offset, u32 val)
{
	writel(val, oinfo->mmio_base + DISPC_REG_OFFSET + offset);
	return val;
}

static __inline__ u32 
dispc_reg_merge(const struct omap24xxfb_info *oinfo, u32 offset, 
		u32 val, u32 mask)
{
	u32 addr = oinfo->mmio_base + DISPC_REG_OFFSET + offset;
	u32 new_val = (readl(addr) & ~mask) | (val & mask);
	
	writel(new_val, addr);
	return new_val;
}

/* Bits-per-pixel and color depth aren't quite the same thing.  The OMAP24xx 
 * display controller supports color depths of 1, 2, 4, 8, 12, 16, and 24 bits.
 * Color depth and bits-per-pixel are the same for depths of 1, 2, 4, 8, and 
 * 16 bits, but for a color depth of 12 bits the pixel data is padded to 
 * 16 bits-per-pixel, and for a color depth of 24 bits the pixel data is padded 
 * to 32 bits-per-pixel.
 */
static inline int var_to_depth(const struct fb_var_screeninfo *var)
{
	switch (var->bits_per_pixel) {
		case 1:
		case 2:
		case 4:
		case 8:
		default:
			return var->bits_per_pixel;
		case 16:
			if ((var->red.length + var->green.length 
				+ var->blue.length) == 12)
			{
				return 12;
			}
			else
				return 16;
		case 32:
			return 24;
	}
}

/* Calculate the number of pixels sent to the display per pixel clock as 
 * (nom/den) pixels per clock.
 */
static void 
pixels_per_clock(const struct omap24xx_dispc_regs *dispc, 
			unsigned int *nom, unsigned int *den)
{
	if (dispc->control & DISPC_CONTROL_STNTFT) {
		/* active display (TFT) */
		if (dispc->control & DISPC_CONTROL_TDMENABLE) {
			/* TFT with TDM */
			switch (dispc->control & DISPC_CONTROL_TDMCYCLEFORMAT) {
				case DISPC_CONTROL_TDMCYCLEFORMAT_1CYCPERPIX:
					*nom = 1;
					*den = 1;
					break;
				case DISPC_CONTROL_TDMCYCLEFORMAT_2CYCPERPIX:
					*nom = 1;
					*den = 2;
					break;
				case DISPC_CONTROL_TDMCYCLEFORMAT_3CYCPERPIX:
					*nom = 1;
					*den = 3;
					break;
				case DISPC_CONTROL_TDMCYCLEFORMAT_3CYCPER2PIX:
					*nom = 2;
					*den = 3;
					break;
			}
		}
		else {
			/* TFT without TDM */
			*nom = 1;
			*den = 1;
		}
	}
	else {
		/* passive display (STN) */
		if (dispc->control & DISPC_CONTROL_MONOCOLOR) {
			/* STN mono */
			if (dispc->control & DISPC_CONTROL_M8B) {
				/* 8 pixels per pixclock */
				*nom = 8;
				*den = 1;
			}
			else {
				/* 4 pixels per pixclock */
				*nom = 4;
				*den = 1;
			}
		}
		else {
			/* STN color--8 pixels per 3 pixclocks */
			*nom = 8;
			*den = 3;
		}
	}
}

/* Calculate the horizontal sync frequency in Hertz 
 * with a resolution of 250Hz.
 */
static unsigned int
horizontal_sync_freq(const struct omap24xx_dispc_regs *dispc, 
			const struct fb_var_screeninfo *var)
{
	unsigned int hsf, hs, nom, den;
	unsigned int xres;

	/* Calculate the number of pixels per clock. */
	pixels_per_clock(dispc, &nom, &den);

	/* get the horizontal display resolution */
	xres = omap24xxfb_display_xres();

	hs = (xres*den)/nom
	 	+ (var->left_margin + var->right_margin + var->hsync_len);

	if ((var->pixclock > 0) && (hs > 0))
		hsf = (4000000000UL/(var->pixclock*hs))*250;
	else
		hsf = 0;

	return hsf;
}

/* Calculate the vertical sync frequency in Hertz. */
static unsigned int
vertical_sync_freq(const struct omap24xx_dispc_regs *dispc, 
			const struct fb_var_screeninfo *var)
{
	unsigned int vsf, vs;
	unsigned int yres;

	/* get the vertical display resolution */
	yres = omap24xxfb_display_yres();

	vs = yres + var->upper_margin + var->lower_margin + var->vsync_len;

	if (vs > 0)
		vsf = horizontal_sync_freq(dispc, var)/vs;
	else
		vsf = 0;

	return vsf;
}

/* Wait with a timeout for both of the GO bits in the control register to 
 * be cleared.  Returns 0 if the GO bits are clear on exit, or non-zero if 
 * the timeout occurs before the GO bits are clear.
 */
static u32
wait_for_go(const struct omap24xxfb_info *oinfo, unsigned long timeout_ticks)
{
	unsigned long timeout = jiffies + timeout_ticks;
	u32 mask = 0;
	u32 ctrl;
	
	ctrl = dispc_reg_in(oinfo, DISPC_CONTROL);
	if (ctrl & DISPC_CONTROL_DIGITALENABLE)
		mask |= DISPC_CONTROL_GODIGITAL;
	if (ctrl & DISPC_CONTROL_LCDENABLE)
		mask |= DISPC_CONTROL_GOLCD;
	while((ctrl & mask) && time_before(jiffies, timeout)) {
		if (!in_atomic()) {
			set_current_state(TASK_INTERRUPTIBLE);
			schedule_timeout(1);
		} else
			udelay(100);
		ctrl = dispc_reg_in(oinfo, DISPC_CONTROL);
	}

	if (ctrl & mask) {
		printk(KERN_WARNING FB_NAME 
			": timeout waiting for display controller update\n");
	}
	return (ctrl & mask);
}

/* Enable the display controller */
static void
omap24xxfb_enable(const struct omap24xxfb_info *oinfo, 
		const struct omap24xx_dispc_regs *dispc, 
		unsigned long timeout_ticks)
{
	unsigned long dispc_control;

	dispc_control = dispc->control;
	if (dispc_control & DISPC_CONTROL_DIGITALENABLE)
		dispc_control |= DISPC_CONTROL_GODIGITAL;
	if (dispc_control & DISPC_CONTROL_LCDENABLE)
		dispc_control |= DISPC_CONTROL_GOLCD;
	dispc_reg_out(oinfo, DISPC_CONTROL, dispc_control);

	wait_for_go(oinfo, timeout_ticks);
	
	return;
}

/* Disable the display controller */
static void
omap24xxfb_disable(const struct omap24xxfb_info *oinfo, 
		unsigned long timeout_ticks)
{
	unsigned long timeout;

	if (dispc_reg_in(oinfo, DISPC_CONTROL) 
		& (DISPC_CONTROL_DIGITALENABLE | DISPC_CONTROL_LCDENABLE))
	{
		/* disable the display controller */
		dispc_reg_merge(oinfo, DISPC_CONTROL, 0, 
			DISPC_CONTROL_DIGITALENABLE | DISPC_CONTROL_LCDENABLE);
		/* wait for any frame in progress to complete */
		dispc_reg_out(oinfo, DISPC_IRQSTATUS, 
				DISPC_IRQSTATUS_FRAMEDONE);
		timeout = jiffies + timeout_ticks;
		while(!(dispc_reg_in(oinfo, DISPC_IRQSTATUS) 
				& DISPC_IRQSTATUS_FRAMEDONE)
			&& time_before(jiffies, timeout))
		{
			if (!in_atomic()) {
				set_current_state(TASK_INTERRUPTIBLE);
				schedule_timeout(1);
			} else
				udelay(100);
		}
	}

	return;
}

/* Reset the display controller */
static void
omap24xxfb_reset(const struct omap24xxfb_info *oinfo, 
		unsigned long timeout_ticks)
{
	unsigned long timeout;

	/* disable the display controller */
	omap24xxfb_disable(oinfo, timeout_ticks);
	
	/* reset the display controller */
	dispc_reg_out(oinfo, DISPC_SYSCONFIG, DISPC_SYSCONFIG_SOFTRESET);
	/* wait for reset to complete */
	timeout = jiffies + timeout_ticks;
	while (!(dispc_reg_in(oinfo, DISPC_SYSSTATUS) 
			& DISPC_SYSSTATUS_RESETDONE)
		&& time_before(jiffies, timeout))
	{
		if (!in_atomic()) {
			set_current_state(TASK_INTERRUPTIBLE);
			schedule_timeout(1);
		} else
			udelay(100);
	}
	if (!(dispc_reg_in(oinfo, DISPC_SYSSTATUS) & DISPC_SYSSTATUS_RESETDONE))
	{
		printk(KERN_WARNING FB_NAME 
			": timeout waiting for display controller reset\n");
	}
	/* remove the reset */
	dispc_reg_merge(oinfo, DISPC_SYSCONFIG, 0, DISPC_SYSCONFIG_SOFTRESET);

	return;
}

static void 
omap24xxfb_write_state(struct omap24xxfb_info *oinfo, 
		struct omap24xx_dispc_regs *dispc, unsigned long timeout_ticks)
{
	/* set the state of bits that are dynamically managed by the driver */
	dispc->sysconfig &= ~DISPC_SYSCONFIG_SOFTRESET;
	dispc->control &= ~(DISPC_CONTROL_GODIGITAL | DISPC_CONTROL_GOLCD);
	dispc->config |= (DISPC_CONFIG_LOADMODE_PGTABUSETB 
				| DISPC_CONFIG_LOADMODE_FRDATLEFR);

	/* enable vsync interrupts */
	dispc->irqenable = DISPC_IRQENABLE_EVSYNC_ODD 
		| DISPC_IRQENABLE_EVSYNC_EVEN | DISPC_IRQENABLE_VSYNC;

	/* This driver doesn't currently support the video windows, so 
	 * we force the palette/gamma table to be a palette table and 
	 * force both video windows to be disabled.
	 */
	dispc->config &= ~DISPC_CONFIG_PALETTEGAMMATABLE;

	/* disable the display controller */
	omap24xxfb_disable(oinfo, timeout_ticks);	

	/*
	 * Update the state of the display controller.
	 */
	
	dispc_reg_out(oinfo, DISPC_SYSCONFIG, dispc->sysconfig);
	dispc_reg_out(oinfo, DISPC_IRQENABLE, dispc->irqenable);
	dispc_reg_out(oinfo, DISPC_CONFIG, dispc->config);
	dispc_reg_out(oinfo, DISPC_DEFAULT_COLOR0, dispc->default_color0);
	dispc_reg_out(oinfo, DISPC_DEFAULT_COLOR1, dispc->default_color1);
	dispc_reg_out(oinfo, DISPC_TRANS_COLOR0, dispc->trans_color0);
	dispc_reg_out(oinfo, DISPC_TRANS_COLOR1, dispc->trans_color1);
	dispc_reg_out(oinfo, DISPC_LINE_NUMBER, dispc->line_number);
	dispc_reg_out(oinfo, DISPC_DATA_CYCLE1, dispc->data_cycle1);
	dispc_reg_out(oinfo, DISPC_DATA_CYCLE2, dispc->data_cycle2);
	dispc_reg_out(oinfo, DISPC_DATA_CYCLE3, dispc->data_cycle3);
	dispc_reg_out(oinfo, DISPC_TIMING_H, dispc->timing_h);
	dispc_reg_out(oinfo, DISPC_TIMING_V, dispc->timing_v);
	dispc_reg_out(oinfo, DISPC_POL_FREQ, dispc->pol_freq);
	dispc_reg_out(oinfo, DISPC_DIVISOR, dispc->divisor);
	dispc_reg_out(oinfo, DISPC_SIZE_LCD, dispc->size_lcd);
	dispc_reg_out(oinfo, DISPC_SIZE_DIG, dispc->size_dig);
	dispc_reg_out(oinfo, DISPC_GFX_BA0, dispc->gfx_ba0);
	dispc_reg_out(oinfo, DISPC_GFX_BA1, dispc->gfx_ba1);
	dispc_reg_out(oinfo, DISPC_GFX_POSITION, dispc->gfx_position);
	dispc_reg_out(oinfo, DISPC_GFX_SIZE, dispc->gfx_size);
	dispc_reg_out(oinfo, DISPC_GFX_ATTRIBUTES, dispc->gfx_attributes);
	dispc_reg_out(oinfo, DISPC_GFX_FIFO_THRESHOLD, 
			dispc->gfx_fifo_threshold);
	dispc_reg_out(oinfo, DISPC_GFX_ROW_INC, dispc->gfx_row_inc);
	dispc_reg_out(oinfo, DISPC_GFX_PIXEL_INC, dispc->gfx_pixel_inc);
	dispc_reg_out(oinfo, DISPC_GFX_WINDOW_SKIP, dispc->gfx_window_skip);
	dispc_reg_out(oinfo, DISPC_GFX_TABLE_BA, dispc->gfx_table_ba);
	dispc_reg_out(oinfo, DISPC_VID1_BA0, dispc->vid1_ba0);
	dispc_reg_out(oinfo, DISPC_VID1_BA1, dispc->vid1_ba1);
	dispc_reg_out(oinfo, DISPC_VID2_BA0, dispc->vid2_ba0);
	dispc_reg_out(oinfo, DISPC_VID2_BA1, dispc->vid2_ba1);
	dispc_reg_out(oinfo, DISPC_VID1_POSITION, dispc->vid1_position);
	dispc_reg_out(oinfo, DISPC_VID2_POSITION, dispc->vid2_position);
	dispc_reg_out(oinfo, DISPC_VID1_SIZE, dispc->vid1_size);
	dispc_reg_out(oinfo, DISPC_VID2_SIZE, dispc->vid2_size);
	dispc_reg_out(oinfo, DISPC_VID1_ATTRIBUTES, dispc->vid1_attributes);
	dispc_reg_out(oinfo, DISPC_VID2_ATTRIBUTES, dispc->vid2_attributes);
	dispc_reg_out(oinfo, DISPC_VID1_FIFO_THRESHOLD, 
			dispc->vid1_fifo_threshold);
	dispc_reg_out(oinfo, DISPC_VID2_FIFO_THRESHOLD, 
			dispc->vid2_fifo_threshold);
	dispc_reg_out(oinfo, DISPC_VID1_ROW_INC, dispc->vid1_row_inc);
	dispc_reg_out(oinfo, DISPC_VID2_ROW_INC, dispc->vid2_row_inc);
	dispc_reg_out(oinfo, DISPC_VID1_PIXEL_INC, dispc->vid1_pixel_inc);
	dispc_reg_out(oinfo, DISPC_VID2_PIXEL_INC, dispc->vid2_pixel_inc);
	dispc_reg_out(oinfo, DISPC_VID1_FIR, dispc->vid1_fir);
	dispc_reg_out(oinfo, DISPC_VID2_FIR, dispc->vid2_fir);
	dispc_reg_out(oinfo, DISPC_VID1_ACCU0, dispc->vid1_accu0);
	dispc_reg_out(oinfo, DISPC_VID1_ACCU1, dispc->vid1_accu1);
	dispc_reg_out(oinfo, DISPC_VID2_ACCU0, dispc->vid2_accu0);
	dispc_reg_out(oinfo, DISPC_VID2_ACCU1, dispc->vid2_accu1);
	dispc_reg_out(oinfo, DISPC_VID1_PICTURE_SIZE, 
			dispc->vid1_picture_size);
	dispc_reg_out(oinfo, DISPC_VID2_PICTURE_SIZE, 
			dispc->vid2_picture_size);
	dispc_reg_out(oinfo, DISPC_VID1_FIR_COEF_H0, dispc->vid1_fir_coef_h0);
	dispc_reg_out(oinfo, DISPC_VID1_FIR_COEF_H1, dispc->vid1_fir_coef_h1);
	dispc_reg_out(oinfo, DISPC_VID1_FIR_COEF_H2, dispc->vid1_fir_coef_h2);
	dispc_reg_out(oinfo, DISPC_VID1_FIR_COEF_H3, dispc->vid1_fir_coef_h3);
	dispc_reg_out(oinfo, DISPC_VID1_FIR_COEF_H4, dispc->vid1_fir_coef_h4);
	dispc_reg_out(oinfo, DISPC_VID1_FIR_COEF_H5, dispc->vid1_fir_coef_h5);
	dispc_reg_out(oinfo, DISPC_VID1_FIR_COEF_H6, dispc->vid1_fir_coef_h6);
	dispc_reg_out(oinfo, DISPC_VID1_FIR_COEF_H7, dispc->vid1_fir_coef_h7);
	dispc_reg_out(oinfo, DISPC_VID2_FIR_COEF_H0, dispc->vid2_fir_coef_h0);
	dispc_reg_out(oinfo, DISPC_VID2_FIR_COEF_H1, dispc->vid2_fir_coef_h1);
	dispc_reg_out(oinfo, DISPC_VID2_FIR_COEF_H2, dispc->vid2_fir_coef_h2);
	dispc_reg_out(oinfo, DISPC_VID2_FIR_COEF_H3, dispc->vid2_fir_coef_h3);
	dispc_reg_out(oinfo, DISPC_VID2_FIR_COEF_H4, dispc->vid2_fir_coef_h4);
	dispc_reg_out(oinfo, DISPC_VID2_FIR_COEF_H5, dispc->vid2_fir_coef_h5);
	dispc_reg_out(oinfo, DISPC_VID2_FIR_COEF_H6, dispc->vid2_fir_coef_h6);
	dispc_reg_out(oinfo, DISPC_VID2_FIR_COEF_H7, dispc->vid2_fir_coef_h7);
	dispc_reg_out(oinfo, DISPC_VID1_FIR_COEF_HV0, dispc->vid1_fir_coef_hv0);
	dispc_reg_out(oinfo, DISPC_VID1_FIR_COEF_HV1, dispc->vid1_fir_coef_hv1);
	dispc_reg_out(oinfo, DISPC_VID1_FIR_COEF_HV2, dispc->vid1_fir_coef_hv2);
	dispc_reg_out(oinfo, DISPC_VID1_FIR_COEF_HV3, dispc->vid1_fir_coef_hv3);
	dispc_reg_out(oinfo, DISPC_VID1_FIR_COEF_HV4, dispc->vid1_fir_coef_hv4);
	dispc_reg_out(oinfo, DISPC_VID1_FIR_COEF_HV5, dispc->vid1_fir_coef_hv5);
	dispc_reg_out(oinfo, DISPC_VID1_FIR_COEF_HV6, dispc->vid1_fir_coef_hv6);
	dispc_reg_out(oinfo, DISPC_VID1_FIR_COEF_HV7, dispc->vid1_fir_coef_hv7);
	dispc_reg_out(oinfo, DISPC_VID2_FIR_COEF_HV0, dispc->vid2_fir_coef_hv0);
	dispc_reg_out(oinfo, DISPC_VID2_FIR_COEF_HV1, dispc->vid2_fir_coef_hv1);
	dispc_reg_out(oinfo, DISPC_VID2_FIR_COEF_HV2, dispc->vid2_fir_coef_hv2);
	dispc_reg_out(oinfo, DISPC_VID2_FIR_COEF_HV3, dispc->vid2_fir_coef_hv3);
	dispc_reg_out(oinfo, DISPC_VID2_FIR_COEF_HV4, dispc->vid2_fir_coef_hv4);
	dispc_reg_out(oinfo, DISPC_VID2_FIR_COEF_HV5, dispc->vid2_fir_coef_hv5);
	dispc_reg_out(oinfo, DISPC_VID2_FIR_COEF_HV6, dispc->vid2_fir_coef_hv6);
	dispc_reg_out(oinfo, DISPC_VID2_FIR_COEF_HV7, dispc->vid2_fir_coef_hv7);
	dispc_reg_out(oinfo, DISPC_VID1_CONV_COEF0, dispc->vid1_conv_coef0);
	dispc_reg_out(oinfo, DISPC_VID2_CONV_COEF0, dispc->vid2_conv_coef0);
	dispc_reg_out(oinfo, DISPC_VID1_CONV_COEF1, dispc->vid1_conv_coef1);
	dispc_reg_out(oinfo, DISPC_VID2_CONV_COEF1, dispc->vid2_conv_coef1);
	dispc_reg_out(oinfo, DISPC_VID1_CONV_COEF2, dispc->vid1_conv_coef2);
	dispc_reg_out(oinfo, DISPC_VID2_CONV_COEF2, dispc->vid2_conv_coef2);
	dispc_reg_out(oinfo, DISPC_VID1_CONV_COEF3, dispc->vid1_conv_coef3);
	dispc_reg_out(oinfo, DISPC_VID2_CONV_COEF3, dispc->vid2_conv_coef3);
	dispc_reg_out(oinfo, DISPC_VID1_CONV_COEF4, dispc->vid1_conv_coef4);
	dispc_reg_out(oinfo, DISPC_VID2_CONV_COEF4, dispc->vid2_conv_coef4);

	/* enable the display controller */
	omap24xxfb_enable(oinfo, dispc, timeout_ticks);

	return;
}

static void 
omap24xxfb_save_state(struct omap24xxfb_info *oinfo, 
			struct omap24xx_dispc_regs *dispc)
{
	dispc->revision = dispc_reg_in(oinfo, DISPC_REVISION);
	dispc->sysconfig = dispc_reg_in(oinfo, DISPC_SYSCONFIG);
	dispc->sysstatus = dispc_reg_in(oinfo, DISPC_SYSSTATUS);
	dispc->irqstatus = dispc_reg_in(oinfo, DISPC_IRQSTATUS);
	dispc->irqenable = dispc_reg_in(oinfo, DISPC_IRQENABLE);
	dispc->control = dispc_reg_in(oinfo, DISPC_CONTROL);
	dispc->config = dispc_reg_in(oinfo, DISPC_CONFIG);
	dispc->capable = dispc_reg_in(oinfo, DISPC_CAPABLE);
	dispc->default_color0 = dispc_reg_in(oinfo, DISPC_DEFAULT_COLOR0);
	dispc->default_color1 = dispc_reg_in(oinfo, DISPC_DEFAULT_COLOR1);
	dispc->trans_color0 = dispc_reg_in(oinfo, DISPC_TRANS_COLOR0);
	dispc->trans_color1 = dispc_reg_in(oinfo, DISPC_TRANS_COLOR1);
	dispc->line_status = dispc_reg_in(oinfo, DISPC_LINE_STATUS);
	dispc->line_number = dispc_reg_in(oinfo, DISPC_LINE_NUMBER);
	dispc->data_cycle1 = dispc_reg_in(oinfo, DISPC_DATA_CYCLE1);
	dispc->data_cycle2 = dispc_reg_in(oinfo, DISPC_DATA_CYCLE2);
	dispc->data_cycle3 = dispc_reg_in(oinfo, DISPC_DATA_CYCLE3);
	dispc->timing_h = dispc_reg_in(oinfo, DISPC_TIMING_H);
	dispc->timing_v = dispc_reg_in(oinfo, DISPC_TIMING_V);
	dispc->pol_freq = dispc_reg_in(oinfo, DISPC_POL_FREQ);
	dispc->divisor = dispc_reg_in(oinfo, DISPC_DIVISOR);
	dispc->size_lcd = dispc_reg_in(oinfo, DISPC_SIZE_LCD);
	dispc->size_dig = dispc_reg_in(oinfo, DISPC_SIZE_DIG);
	dispc->gfx_ba0 = dispc_reg_in(oinfo, DISPC_GFX_BA0);
	dispc->gfx_ba1 = dispc_reg_in(oinfo, DISPC_GFX_BA1);
	dispc->gfx_position = dispc_reg_in(oinfo, DISPC_GFX_POSITION);
	dispc->gfx_size = dispc_reg_in(oinfo, DISPC_GFX_SIZE);
	dispc->gfx_attributes = dispc_reg_in(oinfo, DISPC_GFX_ATTRIBUTES);
	dispc->gfx_fifo_size = dispc_reg_in(oinfo, DISPC_GFX_FIFO_SIZE);
	dispc->gfx_fifo_threshold = 
		dispc_reg_in(oinfo, DISPC_GFX_FIFO_THRESHOLD);
	dispc->gfx_row_inc = dispc_reg_in(oinfo, DISPC_GFX_ROW_INC);
	dispc->gfx_pixel_inc = dispc_reg_in(oinfo, DISPC_GFX_PIXEL_INC);
	dispc->gfx_window_skip = dispc_reg_in(oinfo, DISPC_GFX_WINDOW_SKIP);
	dispc->gfx_table_ba = dispc_reg_in(oinfo, DISPC_GFX_TABLE_BA);
	dispc->vid1_ba0 = dispc_reg_in(oinfo, DISPC_VID1_BA0);
	dispc->vid1_ba1 = dispc_reg_in(oinfo, DISPC_VID1_BA1);
	dispc->vid2_ba0 = dispc_reg_in(oinfo, DISPC_VID2_BA0);
	dispc->vid2_ba1 = dispc_reg_in(oinfo, DISPC_VID2_BA1);
	dispc->vid1_position = dispc_reg_in(oinfo, DISPC_VID1_POSITION);
	dispc->vid2_position = dispc_reg_in(oinfo, DISPC_VID2_POSITION);
	dispc->vid1_size = dispc_reg_in(oinfo, DISPC_VID1_SIZE);
	dispc->vid2_size = dispc_reg_in(oinfo, DISPC_VID2_SIZE);
	dispc->vid1_attributes = dispc_reg_in(oinfo, DISPC_VID1_ATTRIBUTES);
	dispc->vid2_attributes = dispc_reg_in(oinfo, DISPC_VID2_ATTRIBUTES);
	dispc->vid1_fifo_size = dispc_reg_in(oinfo, DISPC_VID1_FIFO_SIZE);
	dispc->vid2_fifo_size = dispc_reg_in(oinfo, DISPC_VID2_FIFO_SIZE);
	dispc->vid1_fifo_threshold = 
		dispc_reg_in(oinfo, DISPC_VID1_FIFO_THRESHOLD);
	dispc->vid2_fifo_threshold = 
		dispc_reg_in(oinfo, DISPC_VID2_FIFO_THRESHOLD);
	dispc->vid1_row_inc = dispc_reg_in(oinfo, DISPC_VID1_ROW_INC);
	dispc->vid2_row_inc = dispc_reg_in(oinfo, DISPC_VID2_ROW_INC);
	dispc->vid1_pixel_inc = dispc_reg_in(oinfo, DISPC_VID1_PIXEL_INC);
	dispc->vid2_pixel_inc = dispc_reg_in(oinfo, DISPC_VID2_PIXEL_INC);
	dispc->vid1_fir = dispc_reg_in(oinfo, DISPC_VID1_FIR);
	dispc->vid2_fir = dispc_reg_in(oinfo, DISPC_VID2_FIR);
	dispc->vid1_accu0 = dispc_reg_in(oinfo, DISPC_VID1_ACCU0);
	dispc->vid1_accu1 = dispc_reg_in(oinfo, DISPC_VID1_ACCU1);
	dispc->vid2_accu0 = dispc_reg_in(oinfo, DISPC_VID2_ACCU0);
	dispc->vid2_accu1 = dispc_reg_in(oinfo, DISPC_VID2_ACCU1);
	dispc->vid1_picture_size = dispc_reg_in(oinfo, DISPC_VID1_PICTURE_SIZE);
	dispc->vid2_picture_size = dispc_reg_in(oinfo, DISPC_VID2_PICTURE_SIZE);
	dispc->vid1_fir_coef_h0 = dispc_reg_in(oinfo, DISPC_VID1_FIR_COEF_H0);
	dispc->vid1_fir_coef_h1 = dispc_reg_in(oinfo, DISPC_VID1_FIR_COEF_H1);
	dispc->vid1_fir_coef_h2 = dispc_reg_in(oinfo, DISPC_VID1_FIR_COEF_H2);
	dispc->vid1_fir_coef_h3 = dispc_reg_in(oinfo, DISPC_VID1_FIR_COEF_H3);
	dispc->vid1_fir_coef_h4 = dispc_reg_in(oinfo, DISPC_VID1_FIR_COEF_H4);
	dispc->vid1_fir_coef_h5 = dispc_reg_in(oinfo, DISPC_VID1_FIR_COEF_H5);
	dispc->vid1_fir_coef_h6 = dispc_reg_in(oinfo, DISPC_VID1_FIR_COEF_H6);
	dispc->vid1_fir_coef_h7 = dispc_reg_in(oinfo, DISPC_VID1_FIR_COEF_H7);
	dispc->vid2_fir_coef_h0 = dispc_reg_in(oinfo, DISPC_VID2_FIR_COEF_H0);
	dispc->vid2_fir_coef_h1 = dispc_reg_in(oinfo, DISPC_VID2_FIR_COEF_H1);
	dispc->vid2_fir_coef_h2 = dispc_reg_in(oinfo, DISPC_VID2_FIR_COEF_H2);
	dispc->vid2_fir_coef_h3 = dispc_reg_in(oinfo, DISPC_VID2_FIR_COEF_H3);
	dispc->vid2_fir_coef_h4 = dispc_reg_in(oinfo, DISPC_VID2_FIR_COEF_H4);
	dispc->vid2_fir_coef_h5 = dispc_reg_in(oinfo, DISPC_VID2_FIR_COEF_H5);
	dispc->vid2_fir_coef_h6 = dispc_reg_in(oinfo, DISPC_VID2_FIR_COEF_H6);
	dispc->vid2_fir_coef_h7 = dispc_reg_in(oinfo, DISPC_VID2_FIR_COEF_H7);
	dispc->vid1_fir_coef_hv0 = dispc_reg_in(oinfo, DISPC_VID1_FIR_COEF_HV0);
	dispc->vid1_fir_coef_hv1 = dispc_reg_in(oinfo, DISPC_VID1_FIR_COEF_HV1);
	dispc->vid1_fir_coef_hv2 = dispc_reg_in(oinfo, DISPC_VID1_FIR_COEF_HV2);
	dispc->vid1_fir_coef_hv3 = dispc_reg_in(oinfo, DISPC_VID1_FIR_COEF_HV3);
	dispc->vid1_fir_coef_hv4 = dispc_reg_in(oinfo, DISPC_VID1_FIR_COEF_HV4);
	dispc->vid1_fir_coef_hv5 = dispc_reg_in(oinfo, DISPC_VID1_FIR_COEF_HV5);
	dispc->vid1_fir_coef_hv6 = dispc_reg_in(oinfo, DISPC_VID1_FIR_COEF_HV6);
	dispc->vid1_fir_coef_hv7 = dispc_reg_in(oinfo, DISPC_VID1_FIR_COEF_HV7);
	dispc->vid2_fir_coef_hv0 = dispc_reg_in(oinfo, DISPC_VID2_FIR_COEF_HV0);
	dispc->vid2_fir_coef_hv1 = dispc_reg_in(oinfo, DISPC_VID2_FIR_COEF_HV1);
	dispc->vid2_fir_coef_hv2 = dispc_reg_in(oinfo, DISPC_VID2_FIR_COEF_HV2);
	dispc->vid2_fir_coef_hv3 = dispc_reg_in(oinfo, DISPC_VID2_FIR_COEF_HV3);
	dispc->vid2_fir_coef_hv4 = dispc_reg_in(oinfo, DISPC_VID2_FIR_COEF_HV4);
	dispc->vid2_fir_coef_hv5 = dispc_reg_in(oinfo, DISPC_VID2_FIR_COEF_HV5);
	dispc->vid2_fir_coef_hv6 = dispc_reg_in(oinfo, DISPC_VID2_FIR_COEF_HV6);
	dispc->vid2_fir_coef_hv7 = dispc_reg_in(oinfo, DISPC_VID2_FIR_COEF_HV7);
	dispc->vid1_conv_coef0 = dispc_reg_in(oinfo, DISPC_VID1_CONV_COEF0);
	dispc->vid2_conv_coef0 = dispc_reg_in(oinfo, DISPC_VID2_CONV_COEF0);
	dispc->vid1_conv_coef1 = dispc_reg_in(oinfo, DISPC_VID1_CONV_COEF1);
	dispc->vid2_conv_coef1 = dispc_reg_in(oinfo, DISPC_VID2_CONV_COEF1);
	dispc->vid1_conv_coef2 = dispc_reg_in(oinfo, DISPC_VID1_CONV_COEF2);
	dispc->vid2_conv_coef2 = dispc_reg_in(oinfo, DISPC_VID2_CONV_COEF2);
	dispc->vid1_conv_coef3 = dispc_reg_in(oinfo, DISPC_VID1_CONV_COEF3);
	dispc->vid2_conv_coef3 = dispc_reg_in(oinfo, DISPC_VID2_CONV_COEF3);
	dispc->vid1_conv_coef4 = dispc_reg_in(oinfo, DISPC_VID1_CONV_COEF4);
	dispc->vid2_conv_coef4 = dispc_reg_in(oinfo, DISPC_VID2_CONV_COEF4);

	return;
}

/* Wait for a vsync interrupt.  This routine sleeps so it can only be called 
 * from process context.
 */
static int
omap24xxfb_wait_for_vsync(struct omap24xxfb_info *oinfo)
{
	wait_queue_t wq;
	unsigned long cnt;
	int ret;

	init_waitqueue_entry(&wq, current);

	cnt = oinfo->vsync_cnt;
	ret = wait_event_interruptible_timeout(oinfo->vsync_wait, 
		cnt != oinfo->vsync_cnt, oinfo->timeout);
	if (ret < 0)
		return ret;
	if (ret == 0)
		return -ETIMEDOUT;

	return 0;
}

/* Interrupt service routine. */
static irqreturn_t
omap24xxfb_isr(int irq, void *arg, struct pt_regs *regs)
{
	struct omap24xxfb_info *oinfo = (struct omap24xxfb_info *) arg;
	unsigned long dispc_irqstatus;

	dispc_irqstatus = dispc_reg_in(oinfo, DISPC_IRQSTATUS);

	/* ack the interrupt */
	dispc_reg_out(oinfo, DISPC_IRQSTATUS, dispc_irqstatus);

	if (dispc_irqstatus & 
		(DISPC_IRQSTATUS_EVSYNC_ODD | DISPC_IRQSTATUS_EVSYNC_EVEN 
		| DISPC_IRQSTATUS_VSYNC))
	{
		++oinfo->vsync_cnt;
		wake_up_interruptible(&oinfo->vsync_wait);
	}
	return IRQ_HANDLED;
}


	/*
	 * fbops functions
	 */

/*
 *	omap24xxfb_check_var - Validates a var passed in. 
 *	@var: frame buffer variable screen structure
 *	@info: frame buffer structure that represents a single frame buffer 
 *
 *	Returns negative errno on error, or zero on success.  The var info on 
 *	exit may be different than the var info on entry.
 *
 *	This function accepts any bits-per-pixel value in the range 0 to 32 
 *	(0 to 8 for monochrome displays).  If the specified number of 
 *	bits-per-pixel isn't supported, then the next greater number of 
 *	bits-per-pixel will be substituted.  This function differentiates 
 *	between color depths of 12 and 16 (which both have 16 bits-per-pixel) 
 *	by checking to see if the sum of the lengths of the RGB fields is 12, 
 *	in which case the color depth is assumed to be 12.  Except for 
 *	differentiating between 12-bit and 16-bit color depths, the values 
 *	passed in var->red, var->green, and var->blue are ignored and replaced 
 *	with the correct values for the specified color depth.
 *
 *	The xres/yres variables in the var screeninfo specify the size of the 
 *	graphics window.  The graphics window size must not be larger than the 
 *	physical display size nor larger than the size of the framebuffer.
 *	The xres_virtual/yres_virtual values in the var screeninfo specify the 
 *	size of the framebuffer in memory.   The framebuffer must not be smaller
 *	than the size of the graphics window.  The display_xres/display_yres 
 *	values in the private info structure specify the size of the display.  
 *	The display must not be smaller than the graphics window.  The display 
 *	size is a compile-time (or perhaps initialization time) configuration 
 *	option and cannot be changed by the user.  The xoffset/yoffset variables 
 *	in the var screeninfo specify the offset of the graphics window within 
 *	the framebuffer.  There is no means for the user to specify the offset 
 *	of the graphics window on the display, so that offset will always be 
 *	zero unless the board-specific mode changing function implements some 
 *	other behavior, such as centering.
 */
static int
omap24xxfb_check_var(struct fb_var_screeninfo *var, struct fb_info *info)
{
	const struct omap24xxfb_info *oinfo = 
				(const struct omap24xxfb_info *) info->par;
	struct fb_var_screeninfo v;
	u32 clkdiv, pixclock, hbp, hfp, hsw, vbp, vfp, vsw;

	memcpy(&v, var, sizeof(v));

	/* do board-specific checks on the var */
	if (omap24xxfb_check_mode(&v))
		return -EINVAL;

	switch (v.bits_per_pixel) {
		case 0 ... 1:
			v.bits_per_pixel = 1;
			break;
		case 2:
			v.bits_per_pixel = 2;
			break;
		case 3 ... 4:
			v.bits_per_pixel = 4;
			break;
		case 5 ... 8:
			v.bits_per_pixel = 8;
			break;
		case 9 ... 16:
			if (v.grayscale)
				return -EINVAL;
			v.bits_per_pixel = 16;
			break;
		case 17 ... 32:
			if (v.grayscale)
				return -EINVAL;
			v.bits_per_pixel = 32;
			break;
		default:
			return -EINVAL;
	}

	switch (var_to_depth(&v)) {
		case 1:
			v.red.offset = v.green.offset = v.blue.offset = 0;
			v.red.length = v.green.length = v.blue.length = 1;
			v.transp.offset = v.transp.length = 0;
			break;
		case 2:
			v.red.offset = v.green.offset = v.blue.offset = 0;
			v.red.length = v.green.length = v.blue.length = 2;
			v.transp.offset = v.transp.length = 0;
			break;
		case 4:
			v.red.offset = v.green.offset = v.blue.offset = 0;
			v.red.length = v.green.length = v.blue.length = 4;
			v.transp.offset = v.transp.length = 0;
			break;
		case 8:
			v.red.offset = v.green.offset = v.blue.offset = 0;
			v.red.length = v.green.length = v.blue.length = 8;
			v.transp.offset = v.transp.length = 0;
			break;
		case 12:
			v.red.offset = 8;
			v.green.offset = 4;
			v.blue.offset = 0;
			v.red.length = v.green.length = v.blue.length = 4;
			v.transp.offset = v.transp.length = 0;
			break;
		case 16:
			v.red.offset = 11;
			v.green.offset = 5;
			v.blue.offset = 0;
			v.red.length = 5;
			v.green.length = 6;
			v.blue.length = 5;
			v.transp.offset = v.transp.length = 0;
			break;
		case 24:
			v.red.offset = 16;
			v.green.offset = 8;
			v.blue.offset = 0;
			v.red.length = v.blue.length = v.green.length = 8;
			v.transp.offset = v.transp.length = 0;
			break;
		default:
			return -EINVAL;
	}

	/* 2048x2048 is max res supported by OMAP24xx display controller */
	if (oinfo->display_xres > 2048 || oinfo->display_yres > 2048)
		return -EINVAL;

	if (oinfo->display_xres == 0 || oinfo->display_yres == 0)
		return -EINVAL;

	if (v.xres_virtual < v.xres || v.yres_virtual < v.yres)
		return -EINVAL;

	if (oinfo->display_xres < v.xres || oinfo->display_yres < v.yres)
		return -EINVAL;

	if (v.xoffset > v.xres_virtual - v.xres)
		return -EINVAL;
	if (v.yoffset > v.yres_virtual - v.yres)
		return -EINVAL;

	if ((v.bits_per_pixel < 8) && (((v.xres_virtual % 8) != 0) 
		|| ((v.xres % 8) != 0) || ((v.xoffset % 8) != 0)))
	{
		return -EINVAL;
	}

	/* check if we have enough video memory to support this framebuffer */
	if (((v.xres_virtual*v.yres_virtual*v.bits_per_pixel)/8) 
		> info->fix.smem_len)
	{
		return -EINVAL;
	}

	/* Convert the timing parameters in the var info to the values needed 
	 * to program into the display controller timing registers, and then 
	 * convert the register values back to the var info values.
	 */
	pixclock = (v.pixclock > 0) ? v.pixclock : (oinfo->gfx_clk_period * 8);
	clkdiv = pixclock / oinfo->gfx_clk_period;
	if (clkdiv < 2)		/* minimum divisor is 2 */
		clkdiv = 2;
	else if (clkdiv > 255)
		clkdiv = 255;	/* maximum divisor is 255 */
	pixclock = oinfo->gfx_clk_period * clkdiv;
	v.pixclock = pixclock;

	/* calculate horizontal timing parameters in pixclocks */
	hbp = (v.left_margin > 0) ? (v.left_margin - 1) : 0;
	if (hbp > 255)
		hbp = 255;
	hfp = (v.right_margin > 0) ? (v.right_margin - 1) : 0;
	if (hfp > 255)
		hfp = 255;
	hsw = (v.hsync_len > 0) ? (v.hsync_len - 1) : 0;
	if (hsw > 63)
		hsw = 63;
	v.left_margin = hbp + 1;
	v.right_margin = hfp + 1;
	v.hsync_len = hsw + 1;
	
	/* calculate vertical timing parameters in line clocks */
	vbp = v.upper_margin;
	if (vbp > 255)
		vbp = 255;
	vfp = v.lower_margin;
	if (vfp > 255)
		vfp = 255;
	vsw = (v.vsync_len > 0) ? (v.vsync_len - 1) : v.vsync_len;
	if (vsw > 63)
		vsw = 63;
	v.upper_margin = vbp;
	v.lower_margin = vfp;
	v.vsync_len = vsw + 1;

	v.red.msb_right = v.green.msb_right = v.blue.msb_right 
		= v.transp.msb_right = 0;

	v.nonstd = 0;
	v.accel_flags = 0;

	memcpy(var, &v, sizeof(v));

	return 0;
}

/*
 *	omap24xxfb_set_par - Alters the hardware state.
 *	@info: frame buffer structure that represents a single frame buffer
 *
 *	Using the fb_var_screeninfo in fb_info we set the resolution of
 *	this particular framebuffer. This function alters the par AND the
 *	fb_fix_screeninfo stored in fb_info. It does not alter var in 
 *	fb_info since we are using that data. This means we depend on the
 *	data in var inside fb_info to be supported by the hardware. 
 *	omap24xxfb_check_var is always called before omap24xxfb_set_par to 
 *	ensure this.
 *
 */
static int
omap24xxfb_set_par(struct fb_info *info)
{
	struct omap24xxfb_info *oinfo = (struct omap24xxfb_info *) info->par;
	struct fb_var_screeninfo *var = &info->var;
	struct omap24xx_dispc_regs *dispc = &oinfo->state;
	u32 gfx_format;
	u32 clkdiv, pixclock, hbp, hfp, hsw, vbp, vfp, vsw;

	/* update the fix screeninfo */
	info->fix.line_length = (var->xres_virtual*var->bits_per_pixel)/8;
	info->fix.visual = ((var_to_depth(var) <= 8) ? 
				FB_VISUAL_PSEUDOCOLOR : FB_VISUAL_TRUECOLOR);
	
	/* Call a board-specific function to configure 
	 * the display controller registers for this mode.
	 * This function may modify values in dispc but must not modify 
	 * values in var.
	 */
	omap24xxfb_change_mode(var, dispc);

	/* Disable RFBI mode, which is not currently supported. */
	dispc->control &= ~DISPC_CONTROL_RFBIMODE;

	/* Convert the timing parameters in the var info to the values needed 
	 * to program into the display controller timing registers.
	 */
	pixclock = (var->pixclock > 0) ? var->pixclock : 
			(oinfo->gfx_clk_period * 8);
	clkdiv = pixclock / oinfo->gfx_clk_period;
	if (clkdiv < 2)		/* minimum divisor is 2 */
		clkdiv = 2;
	else if (clkdiv > 255)
		clkdiv = 255;	/* maximum divisor is 255 */

	dispc->divisor = (1 << DISPC_DIVISOR_LCD_SHIFT) 
		       | (clkdiv << DISPC_DIVISOR_PCD_SHIFT);

	/* calculate horizontal timing parameters in pixclocks */
	hbp = (var->left_margin > 0) ? (var->left_margin - 1) : 0;
	if (hbp > 255)
		hbp = 255;
	hfp = (var->right_margin > 0) ? (var->right_margin - 1) : 0;
	if (hfp > 255)
		hfp = 255;
	hsw = (var->hsync_len > 0) ? (var->hsync_len - 1) : 0;
	if (hsw > 63)
		hsw = 63;
	dispc->timing_h = (hbp << DISPC_TIMING_H_HBP_SHIFT) 
			| (hfp << DISPC_TIMING_H_HFP_SHIFT)
			| (hsw << DISPC_TIMING_H_HSW_SHIFT);
	
	/* calculate vertical timing parameters in line clocks */
	vbp = var->upper_margin;
	if (vbp > 255)
		vbp = 255;
	vfp = var->lower_margin;
	if (vfp > 255)
		vfp = 255;
	vsw = (var->vsync_len > 0) ? (var->vsync_len - 1) : var->vsync_len;
	if (vsw > 63)
		vsw = 63;
	dispc->timing_v = (vbp << DISPC_TIMING_V_VBP_SHIFT) 
			| (vfp << DISPC_TIMING_V_VFP_SHIFT)
			| (vsw << DISPC_TIMING_V_VSW_SHIFT);

	oinfo->hsync = horizontal_sync_freq(dispc, var);
	oinfo->vsync = vertical_sync_freq(dispc, var);
	if (oinfo->vsync > 0)
		oinfo->timeout = ((HZ + oinfo->vsync - 1)/oinfo->vsync)*2;
	else
		oinfo->timeout = HZ/5;	/* 200ms default timeout */

	/* set the sync polarity */
	if (var->sync & FB_SYNC_HOR_HIGH_ACT)
		dispc->pol_freq |=  DISPC_POL_FREQ_IHS;
	else
		dispc->pol_freq &= ~DISPC_POL_FREQ_IHS;
	if (var->sync & FB_SYNC_VERT_HIGH_ACT)
		dispc->pol_freq |=  DISPC_POL_FREQ_IVS;
	else
		dispc->pol_freq &= ~DISPC_POL_FREQ_IVS;

	/* set the display size */
	dispc->size_lcd = (((oinfo->display_xres - 1) 
			    << DISPC_SIZE_LCD_PPL_SHIFT) & DISPC_SIZE_LCD_PPL)
			| (((oinfo->display_yres - 1) 
			    << DISPC_SIZE_LCD_LPP_SHIFT) & DISPC_SIZE_LCD_LPP);
	dispc->size_dig = dispc->size_lcd;

	/* set the graphics window size */
	dispc->gfx_size = (((var->xres - 1) << DISPC_GFX_SIZE_GFXSIZEX_SHIFT) 
				& DISPC_GFX_SIZE_GFXSIZEX)
			| (((var->yres - 1) << DISPC_GFX_SIZE_GFXSIZEY_SHIFT)
				& DISPC_GFX_SIZE_GFXSIZEY);

	/* set the offset of the graphics window within the framebuffer */
	dispc->gfx_ba0 = info->fix.smem_start 
		+ var->yoffset*info->fix.line_length 
		+ (var->xoffset*var->bits_per_pixel)/8;
	dispc->gfx_ba1 = dispc->gfx_ba0;
	dispc->gfx_row_inc = 1 + info->fix.line_length 
		- (var->xres*var->bits_per_pixel)/8;
	dispc->gfx_pixel_inc = 1;

	/* Set the base address of the palette */
	dispc->gfx_table_ba = oinfo->palette_phys;

	/* set the graphics format */
	switch(var_to_depth(var)) {
		case 1:
			gfx_format = DISPC_GFX_ATTRIBUTES_GFXFORMAT_BITMAP1; 
			break;
		case 2:
			gfx_format = DISPC_GFX_ATTRIBUTES_GFXFORMAT_BITMAP2; 
			break;
		case 4:
			gfx_format = DISPC_GFX_ATTRIBUTES_GFXFORMAT_BITMAP4; 
			break;
		case 8:
			gfx_format = DISPC_GFX_ATTRIBUTES_GFXFORMAT_BITMAP8; 
			break;
		case 12:
			gfx_format = DISPC_GFX_ATTRIBUTES_GFXFORMAT_RGB12; 
			break;
		case 16:
			gfx_format = DISPC_GFX_ATTRIBUTES_GFXFORMAT_RGB16; 
			break;
		case 24:
			gfx_format = DISPC_GFX_ATTRIBUTES_GFXFORMAT_RGB24; 
			break;
		default:
			gfx_format = dispc->gfx_attributes 
					& DISPC_GFX_ATTRIBUTES_GFXFORMAT;
			break;
	}
	dispc->gfx_attributes &= ~DISPC_GFX_ATTRIBUTES_GFXFORMAT;
	dispc->gfx_attributes |=  gfx_format;

	/* disable the graphics window if its size is zero */
	if (var->xres == 0 || var->yres == 0)
		dispc->gfx_attributes &= ~DISPC_GFX_ATTRIBUTES_ENABLE;
	else
		dispc->gfx_attributes |=  DISPC_GFX_ATTRIBUTES_ENABLE;

	if (!oinfo->asleep)
		omap24xxfb_write_state(oinfo, dispc, oinfo->timeout);

	return 0;	
}

/*
 *  	omap24xxfb_setcolreg - Sets a color register.
 *	@regno: Which register in the CLUT we are programming 
 *	@red: The red value which can be up to 16 bits wide 
 *	@green: The green value which can be up to 16 bits wide 
 *	@blue:  The blue value which can be up to 16 bits wide.
 *	@transp: If supported the alpha value which can be up to 16 bits wide.	
 *	@info: frame buffer info structure
 * 
 *	Returns non-zero on error, or zero on success.
 */
static int
omap24xxfb_setcolreg(unsigned regno, unsigned red, unsigned green, 
		     unsigned blue, unsigned transp, struct fb_info *info)
{
	struct omap24xxfb_info *oinfo = (struct omap24xxfb_info *) info->par;

	if (regno >= 256)  /* maximum number of palette entries */
		return 1;

	if (info->var.grayscale) {
		/* grayscale = 0.30*R + 0.59*G + 0.11*B */
		red = green = blue = (red * 77 + green * 151 + blue * 28) >> 8;
	}

	/* Truecolor has hardware-independent 16-entry pseudo-palette */
	if (info->fix.visual == FB_VISUAL_TRUECOLOR) {
		u32 v;
	
		if (regno >= 16)
			return 1;

		red >>= (16 - info->var.red.length);
		green >>= (16 - info->var.green.length);
		blue >>= (16 - info->var.blue.length);

		v = (red << info->var.red.offset) |
		    (green << info->var.green.offset) |
		    (blue << info->var.blue.offset);

		switch (info->var.bits_per_pixel) {
			case 16:
		   		((u16*)(info->pseudo_palette))[regno] = v;
				break;
			case 32:	
		   		((u32*)(info->pseudo_palette))[regno] = v;
				break;
			default:
				return 1;
		}
		return 0;
	}

	red >>= 8;
	green >>= 8;
	blue >>= 8;
	(oinfo->palette)[regno] = (red << 16) | (green << 8) | (blue << 0);

	if (!oinfo->asleep) {
		/* tell the display controller to reload the palette table */
		dispc_reg_out(oinfo, DISPC_CONFIG, oinfo->state.config);
	}

	return 0;
}

/*
 *	omap24xxfb_pan_display - Pans the display.
 *	@var: frame buffer variable screen structure
 *	@info: frame buffer structure that represents a single frame buffer
 *
 *	Pan the display using the `xoffset' and `yoffset' fields of the `var' 
 *	structure.  We don't support wrapping and ignore the FB_VMODE_YWRAP 
 *	flag.
 *
 *  	If the values don't fit, return -EINVAL.
 *
 *	Returns negative errno on error, or zero on success.
 */
static int
omap24xxfb_pan_display(struct fb_var_screeninfo *var, struct fb_info *info)
{
	struct omap24xxfb_info *oinfo = (struct omap24xxfb_info *) info->par;
	struct omap24xx_dispc_regs *dispc = &oinfo->state;
	u32 dispc_control;

	if (var->xoffset > info->var.xres_virtual - info->var.xres)
		return -EINVAL;
	if (var->yoffset > info->var.yres_virtual - info->var.yres)
		return -EINVAL;
	if ((info->var.bits_per_pixel < 8) && ((var->xoffset % 8) != 0))
	{
		return -EINVAL;
	}

	/* set the offset of the graphics window within the framebuffer */
	dispc->gfx_ba0 = info->fix.smem_start 
		+ var->yoffset*info->fix.line_length 
		+ (var->xoffset*info->var.bits_per_pixel)/8;
	dispc->gfx_ba1 = dispc->gfx_ba0;
	dispc->gfx_row_inc = 1 + info->fix.line_length 
		- (info->var.xres*info->var.bits_per_pixel)/8;
	dispc->gfx_pixel_inc = 1;

	if (oinfo->asleep)
		return 0;

	/* write our updated register values to the display controller */
	dispc_reg_out(oinfo, DISPC_GFX_BA0, dispc->gfx_ba0);
	dispc_reg_out(oinfo, DISPC_GFX_BA1, dispc->gfx_ba1);
	dispc_reg_out(oinfo, DISPC_GFX_ROW_INC, dispc->gfx_row_inc);
	dispc_reg_out(oinfo, DISPC_GFX_PIXEL_INC, dispc->gfx_pixel_inc);

	/*
	 * Inform the display controller that our register updates are complete.
	 */
	dispc_control = dispc->control & 
		~(DISPC_CONTROL_DIGITALENABLE | DISPC_CONTROL_LCDENABLE);
	dispc_control |= dispc_reg_in(oinfo, DISPC_CONTROL) 
		& (DISPC_CONTROL_DIGITALENABLE | DISPC_CONTROL_LCDENABLE);
	if (dispc_control & DISPC_CONTROL_DIGITALENABLE)
		dispc_control |= DISPC_CONTROL_GODIGITAL;
	if (dispc_control & DISPC_CONTROL_LCDENABLE)
		dispc_control |= DISPC_CONTROL_GOLCD;
	dispc_reg_out(oinfo, DISPC_CONTROL, dispc_control);

	return 0;
}

/**
 *	  omap24xxfb_blank - NOT a required function. Blanks the display.
 *	  @blank_mode: the blank mode we want. 
 *	  @info: frame buffer structure that represents a single frame buffer
 *
 *	  Blank the screen if blank_mode != 0, else unblank. Return 0 if
 *	  blanking succeeded, != 0 if un-/blanking failed due to e.g. a 
 *	  video mode which doesn't support it. Implements VESA suspend
 *	  and powerdown modes on hardware that supports disabling hsync/vsync:
 *	  blank_mode == 2: suspend vsync
 *	  blank_mode == 3: suspend hsync
 *	  blank_mode == 4: powerdown
 *
 *	  Returns negative errno on error, or zero on success.
 *
 */
static int
omap24xxfb_blank(int blank_mode, struct fb_info *info)
{
	struct omap24xxfb_info *oinfo = (struct omap24xxfb_info *) info->par;

	if (oinfo->asleep)
		return 0;

	if (blank_mode) {
		/* blank--disable display controller */
		omap24xxfb_disable(oinfo, oinfo->timeout);
	}
	else {
		/* unblank--enable display controller */
		omap24xxfb_enable(oinfo, &oinfo->state, oinfo->timeout);
	}

	return 0;
}

/**
 *	omap24xxfb_sync - NOT a required function. Normally the accel engine 
 *			  for a graphics card take a specific amount of time.
 *			  Often we have to wait for the accelerator to finish
 *			  its operation before we can write to the framebuffer
 *			  so we can have consistent display output. 
 *
 *	@info: frame buffer structure that represents a single frame buffer
 */
static int
omap24xxfb_sync(struct fb_info *info)
{
	return 0;
}

/**
 *	omap24xxfb_ioctl - handler for private ioctls.
 */
static int
omap24xxfb_ioctl(struct inode *inode, struct file *file, unsigned int cmd, 
	unsigned long arg, struct fb_info *info)
{
	struct omap24xxfb_info *oinfo = (struct omap24xxfb_info *) info->par;

	switch (cmd) {
		case FBIO_WAITFORVSYNC:
			/* This ioctl accepts an integer argument to specify a 
			 * display.  We only support one display, so we will 
			 * simply ignore the argument.
			 */
			return omap24xxfb_wait_for_vsync(oinfo);
	}
	return -EINVAL;
}


	/*
	 *  Frame buffer operations
	 */

static struct fb_ops omap24xxfb_ops = {
	.owner		= THIS_MODULE,
	.fb_check_var	= omap24xxfb_check_var,
	.fb_set_par	= omap24xxfb_set_par,	
	.fb_setcolreg	= omap24xxfb_setcolreg,
	.fb_blank	= omap24xxfb_blank,
	.fb_pan_display	= omap24xxfb_pan_display,	
	.fb_fillrect	= cfb_fillrect, 
	.fb_copyarea	= cfb_copyarea, 
	.fb_imageblit	= cfb_imageblit,
	.fb_cursor	= soft_cursor,
	.fb_rotate	= NULL,
	.fb_sync	= omap24xxfb_sync,
	.fb_ioctl	= omap24xxfb_ioctl,
};

int __init omap24xxfb_setup(char *options)
{
	/* no command-line options to parse for now */	
	return 0;
}

	/*
	 *  Initialization
	 */

int __init omap24xxfb_init(void)
{
	struct omap24xxfb_info *oinfo;
	struct fb_info *info;

#ifndef MODULE
	{
		char *option;
		
		if (fb_get_options("omap24xxfb", &option)) {
			return -ENODEV;
		}
		omap24xxfb_setup(option);
	}
#endif
	oinfo = kmalloc(sizeof(struct omap24xxfb_info), GFP_KERNEL);
	if (!oinfo) {
		printk(KERN_ERR FB_NAME ": could not allocate memory\n");
		return -ENODEV;
	}
	memset(oinfo, 0, sizeof(struct omap24xxfb_info));
	saved_oinfo = oinfo;
	
	/* set the period (in picoseconds) of the graphics timebase */
	oinfo->gfx_clk_period = omap24xxfb_gfx_clk_period();

	/* initialize the display resolution */
	oinfo->display_xres = omap24xxfb_display_xres();
	oinfo->display_yres = omap24xxfb_display_yres();
	
	/* Set base addrs */
	oinfo->fb_base_phys = omap24xxfb_fb_base();
	oinfo->fb_size = omap24xxfb_fb_size();
	oinfo->mmio_base_phys = DSS_REG_BASE;

	/* A null base address indicates that the framebuffer memory should be 
	 * dynamically allocated.
	 */
	if (!oinfo->fb_base_phys)
		oinfo->alloc_fb_mem = 1;

	/* request the mem regions */
	if (!request_mem_region(oinfo->mmio_base_phys, DSS_REG_SIZE, FB_NAME))
	{
		printk(KERN_ERR FB_NAME ": cannot reserve MMIO region\n");
		goto free_par;
	}
	
	if (!oinfo->alloc_fb_mem) {
		if (!request_mem_region(oinfo->fb_base_phys, 
					oinfo->fb_size, FB_NAME))
		{
			printk(KERN_ERR FB_NAME ": cannot reserve FB region\n");
			goto release_mmio;
		}
	}

	/* map the regions */
	oinfo->mmio_base = 
		(unsigned long) ioremap(oinfo->mmio_base_phys, DSS_REG_SIZE);
	if (!oinfo->mmio_base) {
		printk(KERN_ERR FB_NAME ": cannot map MMIO\n");
		goto release_fb;
	}

	if (!oinfo->alloc_fb_mem) {
		oinfo->fb_base = 
			(unsigned long) ioremap(oinfo->fb_base_phys, 
						oinfo->fb_size);
		if (!oinfo->fb_base) {
			printk(KERN_ERR FB_NAME ": cannot map framebuffer\n");
			goto unmap_mmio;
		}
	}
	else {
		/* allocate coherent memory for the framebuffer */
		oinfo->fb_base = (unsigned long) dma_alloc_coherent(NULL, 
				oinfo->fb_size, &oinfo->fb_base_phys, 
				GFP_KERNEL | GFP_DMA);
		if (!oinfo->fb_base) {
			printk(KERN_ERR FB_NAME ": cannot allocate "
				"framebuffer\n");
			goto unmap_mmio;
		}
	}
	memset((void *)oinfo->fb_base, 0, oinfo->fb_size);

	/* allocate coherent memory for the palette */
	oinfo->palette = (u32 *) dma_alloc_coherent(NULL, sizeof(u32)*256, 
				&oinfo->palette_phys, GFP_KERNEL | GFP_DMA);
	if (!oinfo->palette) {
		printk(KERN_ERR FB_NAME ": cannot allocate palette memory\n");
		goto free_fb;
	}
	memset(oinfo->palette, 0, sizeof(u32)*256);

	/* reset the display controller with a timeout of 200ms */
#ifndef OMAP_24XX_VIRTIO
	/* Right now there is a problem with softrest on virtio */
	omap24xxfb_reset(oinfo, HZ/5);
#endif

	/* capture the initial state of the display controller */
	omap24xxfb_save_state(oinfo, &oinfo->state);

	/* initialize the fb_info structure */
	info = &oinfo->info;
	info->flags = FBINFO_DEFAULT;
	info->fbops = &omap24xxfb_ops;
	info->screen_base = (char *)(oinfo->fb_base);
	info->currcon = -1;
	info->pseudo_palette = oinfo->pseudo_palette;
	info->par = oinfo;

	/* Initialize variable screeninfo.
	 * The variable screeninfo can be directly specified by the user 
	 * via an ioctl.
	 */
	memcpy(&info->var, omap24xxfb_default_var(), sizeof(info->var));
	info->var.activate = FB_ACTIVATE_NOW;

	/* Initialize fixed screeninfo.
	 * The fixed screeninfo cannot be directly specified by the user, but 
	 * it may change to reflect changes to the var info.
	 */
	strlcpy(info->fix.id, FB_NAME, sizeof(info->fix.id));
	info->fix.smem_start = oinfo->fb_base_phys;
	info->fix.smem_len = oinfo->fb_size;
	info->fix.type = FB_TYPE_PACKED_PIXELS;
	info->fix.visual = ((var_to_depth(&info->var) <= 8) ? 
				FB_VISUAL_PSEUDOCOLOR: FB_VISUAL_TRUECOLOR);
	info->fix.xpanstep = 0;
	info->fix.ypanstep = 0;
	info->fix.ywrapstep = 0;
	info->fix.line_length = 
		(info->var.xres_virtual*info->var.bits_per_pixel)/8;
	info->fix.type_aux = 0;
	info->fix.mmio_start = oinfo->mmio_base_phys;
	info->fix.mmio_len = DSS_REG_SIZE;
	info->fix.accel = FB_ACCEL_NONE;

	/* Allocate the color map to the maximum size.
	 * Color depths greater than 8 bits are true color and use a 16-bit 
	 * pseudo-palette instead of an actual palette, so the maximum color 
	 * map size is 256 colors.
	 */
	if (fb_alloc_cmap(&info->cmap, 256, 0) < 0) {
		printk(KERN_ERR FB_NAME ": cannot allocate color map\n");
		goto free_palette;
	}

	if (omap24xxfb_check_var(&info->var, info)) {
		printk(KERN_ERR FB_NAME ": invalid default video mode\n");
		goto dealloc_cmap;
	}

	/* initialize board-specific display controller registers */
	omap24xxfb_change_mode(&info->var, &oinfo->state);
	
	oinfo->hsync = horizontal_sync_freq(&oinfo->state, &info->var);
	oinfo->vsync = vertical_sync_freq(&oinfo->state, &info->var);
	if (oinfo->vsync > 0)
		oinfo->timeout = ((HZ + oinfo->vsync - 1)/oinfo->vsync)*2;
	else
		oinfo->timeout = HZ/5;	/* 200ms default timeout */

	/* initialize the vsync wait queue */
	init_waitqueue_head(&oinfo->vsync_wait);

	/* install our interrupt service routine */
	if (request_irq(INT_DSS_IRQ, omap24xxfb_isr, SA_SHIRQ, FB_NAME, oinfo))
	{
		printk(KERN_ERR FB_NAME 
			": could not install interrupt service routine\n");
		goto dealloc_cmap;
	}

	if (register_framebuffer((struct fb_info *) oinfo) < 0) {
		printk(KERN_ERR FB_NAME ": could not register framebuffer\n");
		goto uninstall_isr;
	}

	printk(KERN_INFO FB_NAME ": fb%d frame buffer device\n", info->node);
	printk(KERN_INFO FB_NAME 
		": OMAP24xx display controller hardware rev %d.%d\n",
		(oinfo->state.revision & DISPC_REVISION_MAJOR) 
			>> DISPC_REVISION_MAJOR_SHIFT,
		(oinfo->state.revision & DISPC_REVISION_MINOR) 
			>> DISPC_REVISION_MINOR_SHIFT);
	printk(KERN_INFO FB_NAME ": display mode %dx%dx%d", 
		oinfo->display_xres, oinfo->display_yres, 
		var_to_depth(&info->var));
	if (oinfo->state.control & DISPC_CONTROL_LCDENABLE) {
		printk(" hsync %dkHz vsync %dHz", 
			oinfo->hsync/1000, oinfo->vsync);
	}
	printk("\n");

	return 0;

uninstall_isr:
	free_irq(INT_DSS_IRQ, oinfo);

dealloc_cmap:
	fb_dealloc_cmap(&info->cmap);

free_palette:
	dma_free_coherent(NULL, sizeof(u32)*256, oinfo->palette, 
				oinfo->palette_phys);

free_fb:
	if (!oinfo->alloc_fb_mem)
		iounmap((void*)oinfo->fb_base);
	else
		dma_free_coherent(NULL, oinfo->fb_size, (void *)oinfo->fb_base, 
					oinfo->fb_base_phys);

unmap_mmio:
	iounmap((void*)oinfo->mmio_base);

release_fb:
	if (!oinfo->alloc_fb_mem)
		release_mem_region(oinfo->fb_base_phys, oinfo->fb_size);

release_mmio:
	release_mem_region(oinfo->mmio_base_phys, DSS_REG_SIZE);

free_par:
	kfree(oinfo);
	saved_oinfo = NULL;

	return -ENODEV;
}

	/*
	 *  Cleanup
	 */

static void __exit omap24xxfb_cleanup(void)
{
	struct omap24xxfb_info *oinfo = saved_oinfo;

	if (!oinfo)
		return;

	unregister_framebuffer((struct fb_info *) oinfo);

	/* uninstall isr */
 	free_irq(INT_DSS_IRQ, oinfo);

	/* reset the video controller with a timeout of 200ms */
#ifndef OMAP_24XX_VIRTIO
	/* Right now there is a problem with softrest on virtio */
	omap24xxfb_reset(oinfo, HZ/5);
#endif

	fb_dealloc_cmap(&oinfo->info.cmap);
	
	dma_free_coherent(NULL, sizeof(u32)*256, oinfo->palette, 
			  oinfo->palette_phys);

	if (!oinfo->alloc_fb_mem)
		iounmap((void*)oinfo->fb_base);
	else
		dma_free_coherent(NULL, oinfo->fb_size, 
				(void *)oinfo->fb_base, oinfo->fb_base_phys);
	iounmap((void*)oinfo->mmio_base);

	if (!oinfo->alloc_fb_mem)
		release_mem_region(oinfo->fb_base_phys, oinfo->fb_size);
	release_mem_region(oinfo->mmio_base_phys, DSS_REG_SIZE);

	kfree(oinfo);
	saved_oinfo = NULL;
}

	/*
	 *  Modularization
	 */

module_init(omap24xxfb_init);
module_exit(omap24xxfb_cleanup);

MODULE_LICENSE("GPL");
