/**
*  @file   kmfg.h
*  @brief  The module provides access functions of mfg-information block in flash.
*
*  Copyright (C)
*       2006 Accton Technology Corporation
*
*  @author William Chiang (william_chiang@accton.com.tw)
*
*  References:
*    if needed ( code referenced, data book, ...)
*
*  History:
*    05/01/2005 William, Init draft spec.
&    09/02/2005 William, define each block contents and size (for block-id 1)
*
*  Notes:
*    1. This component (in kernel layer) shall check the signature and checksum,
*       it's internal checking item.
*    2. The key is in percentage, a linear distribution over the physical step.
*       eg. 25% means 1/4, lower layer shall use calibration data mapping to
*       physical val.
*    3. Caller may use get_value function to retrieve value with exactly matched or
*       leaner mapping. eg. to map non-calibrated data -
*       if (kmfg_get_data (KMFG_BLOCK_ID_CONTRAST, &buffer, &buf_len)>0)
*       {
*           kmfg_get_value_with_key (&buffer, KMFG_MATCH_LINEAR_MAPPING, 37, &value);
*       }
*       or caller maps itself -
*       if (kmfg_get_data (KMFG_BLOCK_ID_CONTRAST, &buffer, &buf_len)>0)
*       {
*           kmfg_locate_closer_keys (&buffer, 37, &pre_key, &post_key);
*           kmfg_get_value_with_key (&buffer, KMFG_MATCH_EXACTLY_MATCHED, pre_key, &pre_value);
*           kmfg_get_value_with_key (&buffer, KMFG_MATCH_EXACTLY_MATCHED, post_key, &post_value);
*           calculate by caller .....
*       }
*
**/

#ifndef __KMFG_H__
#define __KMFG_H__

/*** Include file section
 ***      each include file shall self-contain, does not include other header file.
 ***/
#include "dcl_constant.h"

/*** Constant definition section
 ***      define the constants and structures used for KMFG module.
 ***/

/*   Constant used in KMFG
 *        If any constant should refer other module,
 *        the constant should move to/undefine in this module.
 */
/** HW Configuration ID **/
#define  KMFG_HW_CONFIG_ID_RD                0x0000
#define  KMFG_HW_CONFIG_ID_WIFI              0x0100
#define  KMFG_HW_CONFIG_ID_LAN               0x0200
/** FW Family **/
#define  KMFG_FW_FAMILY_RD                   0x0000
#define  KMFG_FW_FAMILY_SK                   0x0100
#define  KMFG_FW_FAMILY_SIP                  0x0200
/** Calibration constants */
#define  KMFG_CALIBRATE_POT_NO               12  //need to confirm
#define  KMFG_WIFI_DATA_LEN                  1018
#define  KMFG_RESERVE_SYSTEM_LEN             16
#define  KMFG_RESERVE_CONTRAST_LEN           6
#define  KMFG_RESERVE_BRIGHTNESS_LEN         6
#define  KMFG_RESERVE_BATTERY_LEN            6
#define  KMFG_RESERVE_VOLUME_LEN             10

/**  Return Value Definition  **/
#define  DCL_ERR_MFG_BLOCK_ID_UNDEFINED         (DCL_ERR_MFG | 0x11)
#define  DCL_ERR_MFG_BUFFER_TOO_SMALL           (DCL_ERR_MFG | 0x12)
#define  DCL_ERR_MFG_OVER_BLOCK_SIZE            (DCL_ERR_MFG | 0x13)
#define  DCL_ERR_MFG_CHECKSUM_ERROR             (DCL_ERR_MFG | 0x14)
#define  DCL_ERR_MFG_INVALID_DATA               (DCL_ERR_MFG | 0x15)
#define  DCL_ERR_MFG_INVALID_KEY                (DCL_ERR_MFG | 0x16)
#define  DCL_ERR_MFG_KEY_NOT_FOUND              (DCL_ERR_MFG | 0x17)
#define  DCL_ERR_MFG_BLOCK_WRITE_ERROR          (DCL_ERR_MFG | 0x18)
#define  DCL_ERR_MFG_BLOCK_READ_ERROR           (DCL_ERR_MFG | 0x19)

/*** Structure / type definition section
 ***      define the enumeric block ID.
 ***      for each ID, it cannot be removed
 ***      even the block is not used for the product.
 ***/
typedef enum KMFG_BLOCK_ID_E
{
    KMFG_BLOCK_ID_SYSTEM = 1,
    KMFG_BLOCK_ID_CONTRAST,
    KMFG_BLOCK_ID_BRIGHTNESS,
    KMFG_BLOCK_ID_BATTERY,
    //KMFG_BLOCK_ID_VOLUME,
    KMFG_BLOCK_ID_HND_VOLUME,
    KMFG_BLOCK_ID_HED_VOLUME,
    KMFG_BLOCK_ID_HND_GAIN,
    KMFG_BLOCK_ID_HED_GAIN,
    KMFG_BLOCK_ID_WIFI
}   KMFG_BLOCK_ID_T;

typedef enum KMFG_MATCH_OPTION_E
{
    KMFG_MATCH_EXACTLY_MATCHED = 1,
    KMFG_MATCH_LINEAR_MAPPING,
    KMFG_MATCH_ALIGNMENT             // default option
}   KMFG_MATCH_OPTION_T;

#pragma pack(2)     // align to 2 bytes boundary

typedef struct KMFG_SYSTEM_BLK_S
{
    char   mac_address[6];
    char   hw_revision[4];

    //William add
    char   hw_minor_update[2];
	
    char   mfg_date[8];
    char   serial_no[16];
    char   product_id[16];
    char   product_name[32];

    //William add
    char   country_code[4];
	
    char   vendor_product_name[32];

    //William add
    //short  hw_config_id;
    //short  fw_family;
    unsigned short  hw_config_id;
    unsigned short  fw_family;

    char   reserved[KMFG_RESERVE_SYSTEM_LEN];
}   __attribute__ ((packed)) KMFG_SYSTEM_BLK_T;

typedef struct KMFG_2_2_REC_S
{
    short  key;   /* 0-100 for 0-100% */
    short  val;   /* step value */
}   __attribute__ ((packed)) KMFG_2_2_REC_T;

typedef struct KMFG_CALIBRATE_BLK_S
{
    short               key_no;       /* total key used for calibration,
                                         must be less than KMFG_CALIBRATION_POT_NO */
    short               default_key;  /* default key for system initalized */
    KMFG_2_2_REC_T   rec[KMFG_CALIBRATE_POT_NO];
}   __attribute__ ((packed)) KMFG_CALIBRATE_BLK_T;

typedef struct KMFG_CONTRAST_BLK_S
{
    KMFG_CALIBRATE_BLK_T contrast;
    char                    reserved[KMFG_RESERVE_CONTRAST_LEN];
}   __attribute__ ((packed)) KMFG_CONTRAST_BLK_T;

typedef struct KMFG_BRIGHTNESS_BLK_S
{
    KMFG_CALIBRATE_BLK_T brightness;
    char                    reserved[KMFG_RESERVE_BRIGHTNESS_LEN];
}   __attribute__ ((packed)) KMFG_BRIGHTNESS_BLK_T;

typedef struct KMFG_BATTERY_BLK_S
{
    KMFG_CALIBRATE_BLK_T battery;
    char                    reserved[KMFG_RESERVE_BATTERY_LEN];
}   __attribute__ ((packed)) KMFG_BATTERY_BLK_T;

typedef struct KMFG_VOLUME_BLK_S
{
    KMFG_CALIBRATE_BLK_T hmic;   /* headset mic */
    KMFG_CALIBRATE_BLK_T hspk;   /* headset speaker */
    KMFG_CALIBRATE_BLK_T nmic;   /* handset mic */
    KMFG_CALIBRATE_BLK_T nspk;   /* handset speaker */
    char                    reserved[KMFG_RESERVE_VOLUME_LEN];
}   __attribute__ ((packed)) KMFG_VOLUME_BLK_T;

typedef struct KMFG_WIFI_BLK_S
{
    short   key_no;
    short   default_key;
    char    wifi_data[KMFG_WIFI_DATA_LEN];
}   __attribute__ ((packed)) KMFG_WIFI_BLK_T;

#pragma pack()    // use default alignment


#define KMFG_BLOCK
#ifdef KMFG_BLOCK
typedef struct kmfg_block_s {
	KMFG_BLOCK_ID_T block_id;
	unsigned int block_size;
	char *block;
}  __attribute__ ((packed)) kmfg_block_T;;
#endif

// don't put variable instance here

/*** Export functions section
 ***      define the export functions for applications
 ***/


int omap_flash_read_mac(u8* buf);

/**   Name:     get_data
 **   Function: retrieve the information saved in mfg-info block.
 **   Input:
 **   @param    block_id - the identifier of block
 **   Output:
 **   @param    buffer - buffer for retrieved data
 **   @param    buf_len - the buffer for receving
 **   @retval   > 0 - the length of data returned.
 **   @retval   DCL_ERR_MFG_BLOCK_ID_UNDEFINED - invalid block ID.
 **   @retval   DCL_ERR_MFG_BUFFER_TOO_SMALL - the specified buffer is too small for the block information
 **   @retval   DCL_ERR_MFG_CHECKSUM_ERROR - the specified block is corrupted.
 **   Notes:
 **     1. If buffer is NULL, buf_len return the buffer size caller should be prepared.
 **/
int kmfg_get_data (KMFG_BLOCK_ID_T block_id, char *buffer, int *buf_len);


/**   Name:     kmfg_set_data
 **   Function: save the mfg-info block to flash
 **   Input:
 **   @param    block_id - the identifier of block
 **   @param    data_len - the data size to be saved.
 **   Output:
 **   @param    data - the whole block of data
 **   @return   > 0 - the length of data returned
 **   @retval   DCL_ERR_MFG_BLOCK_ID_UNDEFINED - invalid block ID.
 **   @retval   DCL_ERR_MFG_OVER_BLOCK_SIZE - the specified data over the boundary of the block.
 **   Notes:
 **
 **/
int kmfg_set_data (KMFG_BLOCK_ID_T block_id, char *data, int data_len);


/**   Name:     kmfg_get_value_with_key
 **   Function: Use specific key to retrieve related calibrating value.
 **   Input:
 **   @param    buffer - which stored calibration data in KMFG_CALIBRATE_BLK_T format
 **   @param    option - KMFG_MATCH_EXACTLY_MATCHED; must match the request key
 **                      KMFG_MATCH_LINEAR_MAPPING; linear mapping if cannot match the key
 **                      KMFG_MATCH_ALIGNMENT; return the value with closer key
 **   @param    request_key - the request key
 **   Output:
 **   @param    value - the calibrated value for the request key.
 **   @return   DCL_OK - successfully retrieves the value as option indicated.
 **   @retval   DCL_ERR_MFG_INVALID_DATA - not a valid calibration block
 **   @retval   DCL_ERR_MFG_INVALID_KEY - the key out of range (0-100)
 **   @retval   DCL_ERR_MFG_KEY_NOT_FOUND - can not exactly match when option is KMFG_EXACTLY_MATCHED
 **   Notes:
 **
 **/
int kmfg_get_value_with_key (char *buffer, KMFG_MATCH_OPTION_T option, short request_key, short *value);


/**   Name:     kmfg_locate_closer_keys
 **   Function: Use specific key to retrieve related calibrating value.
 **   Input:
 **   @param    buffer - which stored calibration data in KMFG_CALIBRATE_BLK_T format
 **   @param    request_key - the request key
 **   Output:
 **   @param    pre_key - the key before the request key or the key itself if matched.
 **   @param    post_key - the key after the request key or the key itself if matched.
 **   @return   DCL_OK - successfully return closer key
 **   @retval   DCL_ERR_MFG_INVALID_DATA - not a valid calibration block
 **   @retval   DCL_ERR_MFG_INVALID_KEY - the key out of range (0-100)
 **   Notes:
 **
 **/
int kmfg_locate_closer_keys (char *buffer, short request_key, short *pre_key, short *post_key);

#endif  /* end of __KMFG_H__ */

