/**
*  @file   kmsg.c
*  @brief  This module defines the functions for error message or 
*          debug message or information. All applications should
*          use these functions to show message.
*
*  Copyright (C)
*       2006 Accton Technology Corporation
*
*  @author Sean Hsieh (sean_hsieh@accton.com.tw)
*
*  References:
*    if needed ( code referenced, data book, ...)
*
*  History:
*    02/06/2006 Sean Hsieh,  define whole mechanism and code
*    mm/dd/yyyy author modified code descriptions
*
*
*  TO DO:
*    Linux kernel has a built-in function "snprintf()", but we should have the 
*    time handling functions in kernel. Otherwise, we may skip the time information
*    for kernel.
*    For multi-threading support, we should implement the two macros MUTEX_LOCK
*    and MUTEX_UNLOCK for kernel.
*
**/


/*** Include file section
 ***      each include file shall self-contain, does not include other header file.
 ***/
/*
 
#ifdef PMSG_SUPPORT_THREAD
#include <pthread.h>
#endif // PMSG_SUPPORT_THREAD

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>
#include <time.h>
*/

#include <linux/module.h>

#include <linux/types.h>

/*** Constant definition section
 ***      define the constants and structures used for DCL_SYS module.
 ***/

//#define PRINTF	printf
#define PRINTF(x)	printk(x)

/*
#ifdef PMSG_SUPPORT_THREAD


static pthread_mutex_t msgMutex = PTHREAD_MUTEX_INITIALIZER;
#define MUTEX_LOCK	pthread_mutex_lock( &msgMutex );
#define MUTEX_UNLOCK	pthread_mutex_unlock( &msgMutex );
*/
#define __MUTEX_LOCK \
	//spin_lock_irqsave(&kmsgLock, kmsgFlags);

#define __MUTEX_UNLOCK \
	//spin_unlock_irqrestore(&kmsgLock, kmsgFlags);



#define MUTEX_LOCK	((in_atomic()) ? NULL : __MUTEX_LOCK)

#define MUTEX_UNLOCK	\
	((in_atomic())? 	\
	NULL				\
	: __MUTEX_UNLOCK)
/*
#else

#define MUTEX_LOCK
#define MUTEX_UNLOCK
*/
//#endif // PMSG_SUPPORT_THREAD

/*** Structure / type definition / Classes section
 ***      define the structures used for all DCL modules
 ***/
typedef struct {
    int prefix: 1;
    int time: 1;
    int file: 1;
    int func: 1;
    int line: 1;
    int msg: 1;
} INFO_ITEMS;

typedef union {
    INFO_ITEMS items;
    int val;
} OUTPUT_FIELDS;

/*** Local variable declaration section
 ***/
static char *fmtPrefix = "[%s]";
static char *fmtTime = " %lu";
static char *fmtFile = " [%s]";
static char *fmtFunc = " %s()";
static char *fmtLine = " %d";
static char *fmtSep = ", ";
//static char *fmtErr = " : %s\n";

static int debugLevel = 2;	// define the debug info's level
static spinlock_t kmsgLock;


static OUTPUT_FIELDS outputFields = { 
    items: { prefix: 1,  time: 1, file: 1, func: 1, line: 1, msg: 1 }
};				// define the detail level of output message



/*
** Local function declaration section
 ***/
static int assembleInfo( char *buf, size_t size, const char *prefix, const char *file, const char *func, int line );

/*** Local function body section
 ***/

/**   Name:     assembleInfo
 **   Assemble the output message to a specified string buffer.
 **   Input:
 **   @param    buf
 **                 A pointer of string buffer for output string.
 **                 output any message.
 **             size
 **                 The size of buf, 
 **             prefix
 **                 User can pass a prefix string for the output message.
 **             file
 **                 The source code file name, usually use the complier default resvered word. Like
 **                 using __FILE__ while using GCC.
 **             func
 **                 The function name, usually use the complier default resvered word. Like
 **                 using __FUNCTION__ while using GCC.
 **             line
 **                 The line number in source file, usually use the complier default resvered word. Like
 **                 using __LINE__ while using GCC.
 **/
static int assembleInfo( char *buf, size_t size, const char *prefix, const char *file, const char *func, int line )
{
    int sl;
    //time_t t;
    //struct tm *tm;

    sl = 0;
    if ( outputFields.items.prefix ) {
	sl = snprintf( buf, size, fmtPrefix, prefix );
    }

/*
    if ( outputFields.items.date || outputFields.items.time ) {
	time( &t );
	tm = localtime( &t );
	if ( outputFields.items.date ) {
	    sl += snprintf( &buf[sl], size - sl, fmtDate, tm->tm_year + 1900, tm->tm_mon + 1, tm->tm_mday );
	}
	if ( outputFields.items.time ) {
	    sl += snprintf( &buf[sl], size - sl, fmtTime, tm->tm_hour + 1, tm->tm_min, tm->tm_sec );
	}
    }
*/

	if ( outputFields.items.time ) {
	    sl += snprintf( &buf[sl], size - sl, fmtTime, jiffies);
	}

    if ( outputFields.items.file ) {
	sl += snprintf( &buf[sl], size - sl, fmtFile, file );
    }

    if ( outputFields.items.func ) {
	sl += snprintf( &buf[sl], size - sl, fmtFunc, func );
    }

    if ( outputFields.items.line ) {
	sl += snprintf( &buf[sl], size - sl, fmtLine, line );
    }

    if ( outputFields.val ) {
	sl += snprintf( &buf[sl], size - sl, "%s", fmtSep );
    }
    
 
    return sl;
}

/*** Export function body section
 ***/

/**   Name:     PMSG_setDebugLevel
 **   Set the debug output level to control the debug message output.
 **   Input:
 **   @param    level
 **                 Set debug output level.
 **/
void __KMSG_setDebugLevel( int level )
{
    debugLevel = level;
}

int __KMSG_getDebugLevel(void)
{
    return debugLevel;
}



/**   Name:     PMSG_setOutputFields
 **   Set which field should be printed.
 **   Input:
 **   @param    prefix
 **                 Set 0 to disable printing the prefix field, 1 to enable printing
 **		    the prefix field.
 **             date
 **		    Set 0 to disable printing the date field, 1 to enable printing
 **		    the date field.
 **             time
 **		    Set 0 to disable printing the time field, 1 to enable printing
 **		    the time field.
 **             file
 **                 Set 0 to disable printing the file field, 1 to enable printing
 **		    the file field.
 **             func
 **                 Set 0 to disable printing the func field, 1 to enable printing
 **		    the func field.
 **             line
 **                 Set 0 to disable printing the line field, 1 to enable printing
 **		    the line field.
 **             msg
 **                 Set 0 to disable printing the msg field, 1 to enable printing
 **		    the msg field.
 **/
void KMSG_setOutputFields( int prefix,  int time, int file, int func, int line, int msg )
{
    outputFields.items.prefix = (prefix != 0);
    outputFields.items.time = (time != 0);
    outputFields.items.file = (file != 0);
    outputFields.items.func = (func != 0);
    outputFields.items.line = (line != 0);
    outputFields.items.msg = (msg != 0);
}



/**   Name:     PMSG_printMsg
 **   Print the debug message.
 **   Input:
 **   @param    buglevel
 **                 If buglevel larger than debugLevel, this function will not 
 **                 output any message.
 **             prefix
 **                 User can pass a prefix string for the output message.
 **             file
 **                 The source code file name, usually use the complier default resvered word. Like
 **                 using __FILE__ while using GCC.
 **             func
 **                 The function name, usually use the complier default resvered word. Like
 **                 using __FUNCTION__ while using GCC.
 **             line
 **                 The line number in source file, usually use the complier default resvered word. Like
 **                 using __LINE__ while using GCC.
 **             fmt
 **                 The string format of output message. The syntax is the same as printf.
 **/
void KMSG_printMsg( int buglevel, const char *prefix, const char *file, const char *func, int line, const char *fmt, ... )
{

	char buf[404];
	int sl;
	va_list ap;
	unsigned long kmsgFlags;

	if ( buglevel > debugLevel )
		return;

	//protect critical section espectually called by isr
	spin_lock_irqsave(&kmsgLock, kmsgFlags);

	sl = assembleInfo( buf, 400, prefix, file, func, line );
   	va_start( ap, fmt );
    	sl += vsnprintf( &buf[sl], 400 - sl, fmt, ap );
    	va_end( ap );

	spin_unlock_irqrestore(&kmsgLock, kmsgFlags);

	PRINTF( buf );	

}



