/*
 * linux/sound/oss/omap-audio-tsc2101.c
 *
 * Glue driver for TSC2101 for OMAP processors
 *
 * Copyright (C) 2004 Texas Instruments, Inc.
 *
 * This package is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * THIS PACKAGE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * History:
 *  -------
 *  2004-08-12 Nishanth Menon - Modified to integrate Audio requirements on 1610,1710
 *               and 2420 platforms.
 *  2004-09-14 Sriram Kannan - Added /proc support for asynchronous starting/stopping the codec
 *		(without affecting the normal driver flow).
 *  2004-10-20 Nishanth Menon - Fixed issue of Mixer operations in 2420.
 *  2004-11-04 Nishanth Menon - Support for power management
 *  2004-11-07 Nishanth Menon - Support for Common TSC access b/w Touchscreen and audio drivers
 */

/***************************** INCLUDES ************************************/

#include <linux/module.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/fs.h>
#include <linux/delay.h>
#include <linux/pm.h>
#include <linux/errno.h>
#include <linux/sound.h>
#include <linux/soundcard.h>
#include <linux/apm_bios.h>


#include <asm/semaphore.h>
#include <asm/uaccess.h>
#include <asm/hardware.h>
#include <asm/arch/dma.h>
#include <asm/io.h>
#include <asm/hardware.h>

#include <asm/arch/mux.h>
#include <asm/arch/io.h>

#include "omap-audio.h"
#include "omap-audio-dma-intfc.h"
#include <../drivers/ssi/omap-mcbsp.h>
#if CONFIG_ARCH_OMAP16XX
#include <asm/arch/clocks.h>
#include <../drivers/ssi/omap-uwire.h>
#elif CONFIG_ARCH_OMAP24XX
#include <../drivers/ssi/omap-mcspi.h>
#endif

#include <asm/hardware/tsc2101.h>
#include <../drivers/ssi/omap-tsc2101.h>
#include <linux/interrupt.h>
#include <asm/arch/gpio.h>
/***************************** MACROS ************************************/

#define PROC_SUPPORT

#ifdef PROC_SUPPORT
#include <linux/proc_fs.h>
#define PROC_START_FILE "driver/tsc2101-audio-start"
#define PROC_STOP_FILE  "driver/tsc2101-audio-stop"
#endif

#define CODEC_NAME		 "TSC2101"

#if CONFIG_ARCH_OMAP16XX
#define PLATFORM_NAME "OMAP16XX"
#define DMA_TX  OMAP_DMA_MCBSP1_TX
#define DMA_RX  OMAP_DMA_MCBSP1_RX
#elif CONFIG_ARCH_OMAP24XX
#define PLATFORM_NAME "OMAP24xx"
#define DMA_TX  eAudioTx
#define DMA_RX  eAudioRx
#endif

/* Define to set the tsc as the master w.r.t McBSP */
#define TSC_MASTER

/*
 * AUDIO related MACROS
 */
#define DEFAULT_BITPERSAMPLE          16
#define AUDIO_RATE_DEFAULT	          44100
#define PAGE2_AUDIO_CODEC_REGISTERS   (2)
#define LEAVE_CS 			          0x80

/* Select the McBSP For Audio */
#if CONFIG_ARCH_OMAP16XX
#define AUDIO_MCBSP                   OMAP_MCBSP1
#elif CONFIG_ARCH_OMAP24XX
#define AUDIO_MCBSP                   OMAP_MCBSP2
#else
#error "UnSupported Configuration"
#endif

#define REC_MASK 			          (SOUND_MASK_LINE | SOUND_MASK_MIC)
#define DEV_MASK 			          (REC_MASK | SOUND_MASK_VOLUME)

#define SET_VOLUME 			          1
#define SET_LINE   			          2
#define SET_ADPGA				   3

//#define DEFAULT_VOLUME                93
#define DEFAULT_VOLUME                99
#define DEFAULT_INPUT_VOLUME          0	/* 0 ==> mute line in */

/* Tsc Audio Specific */
#define NUMBER_SAMPLE_RATES_SUPPORTED 16
#define OUTPUT_VOLUME_MIN 0x7F
//Alan Modify
//#define OUTPUT_VOLUME_MAX 0x32
#define OUTPUT_VOLUME_MAX 0x03
#define OUTPUT_VOLUME_RANGE           (OUTPUT_VOLUME_MIN - OUTPUT_VOLUME_MAX)
#define OUTPUT_VOLUME_MASK            OUTPUT_VOLUME_MIN
#define DEFAULT_VOLUME_LEVEL          OUTPUT_VOLUME_MAX

/* use input vol of 75 for 0dB gain */
#define INPUT_VOLUME_MIN 		      0x0
#define INPUT_VOLUME_MAX 		      LIV_MAX
#define INPUT_VOLUME_RANGE 		      (INPUT_VOLUME_MAX - INPUT_VOLUME_MIN)
#define INPUT_VOLUME_MASK 		      INPUT_VOLUME_MAX

//For Headset detection
#define OMAP_1710_GPIO_NUM 7




/*********** Debug Macros ********/
/* To Generate a rather shrill tone -test the entire path */
//#define TONE_GEN
/* To Generate a tone for each keyclick - test the tsc,spi paths*/
//#define TEST_KEYCLICK
/* To dump the tsc registers for debug */
//#define TSC_DUMP_REGISTERS

#ifdef DPRINTK
#undef DPRINTK
#endif
#undef DEBUG

//#define DEBUG
#ifdef DEBUG
#define DPRINTK(ARGS...)  printk(KERN_INFO "<%s>: ",__FUNCTION__);printk(ARGS)
#define FN_IN printk(KERN_INFO "[%s]: start\n", __FUNCTION__)
#define FN_OUT(n) printk(KERN_INFO "[%s]: end(%u)\n",__FUNCTION__, n)
#else
#define DPRINTK( x... )
#define FN_IN
#define FN_OUT(n)
#endif

#define WIFI_PHONE_PM
//#define DEBUG_PM
#ifdef DEBUG_PM
#define info(format, arg...) printk(KERN_INFO "omap-audio-tsc2101" ": " format "\n",  ## arg )
#else
#define info(format, arg...)
#endif

#ifdef WIFI_PHONE_PM
struct pm_dev *pm_dev_tsc2101 = NULL;
#endif


/***************************** Data Structures **********************************/

static audio_stream_t output_stream = {
	.id               = "TSC2101 out",
	.dma_dev          = DMA_TX,
	.input_or_output  = FMODE_WRITE
};

static audio_stream_t input_stream = {
	.id               = "TSC2101 in",
	.dma_dev          = DMA_RX,
	.input_or_output  = FMODE_READ
};

static int audio_dev_id, mixer_dev_id;

static struct tsc_local_info {
	u8  volume;
	u16 volume_reg;
	u8  line;
	u8  mic;
	u16 input_volume_reg;
	int mod_cnt;
} tsc2101_local;

struct sample_rate_reg_info {
	u16 sample_rate;
	u8  divisor;
	u8  fs_44kHz;		/* if 0 48 khz, if 1 44.1 khz fsref */
};

/* To Store the default sample rate */
static long audio_samplerate = AUDIO_RATE_DEFAULT;

static const struct sample_rate_reg_info
 reg_info[NUMBER_SAMPLE_RATES_SUPPORTED] = {
	/* Div 1 */
	{48000, 0, 0},
	{44100, 0, 1},
	/* Div 1.5 */
	{32000, 1, 0},
	{29400, 1, 1},
	/* Div 2 */
	{24000, 2, 0},
	{22050, 2, 1},
	/* Div 3 */
	{16000, 3, 0},
	{14700, 3, 1},
	/* Div 4 */
	{12000, 4, 0},
	{11025, 4, 1},
	/* Div 5 */
	{9600, 5, 0},
	{8820, 5, 1},
	/* Div 5.5 */
	{8727, 6, 0},
	{8018, 6, 1},
	/* Div 6 */
	{8000, 7, 0},
	{7350, 7, 1},
};

/***************************** MODULES SPECIFIC FUNCTION PROTOTYPES ********************/

static void omap_tsc2101_initialize(void *dummy);

static void omap_tsc2101_shutdown(void *dummy);

static int  omap_tsc2101_ioctl(struct inode *inode, struct file *file,
			       uint cmd, ulong arg);

static int  omap_tsc2101_probe(void);

static void omap_tsc2101_remove(void);

static int  omap_tsc2101_suspend(void);

static int  omap_tsc2101_resume(void);

static inline void tsc2101_configure(void);

static int  mixer_open(struct inode *inode, struct file *file);

static int  mixer_release(struct inode *inode, struct file *file);

static int  mixer_ioctl(struct inode *inode, struct file *file, uint cmd,
		        ulong arg);

#ifdef TEST_KEYCLICK
void tsc2101_testkeyclick(void);
#endif

#ifdef TONE_GEN
void toneGen(void);
#endif

#ifdef TSC_DUMP_REGISTERS
static void tsc2101_dumpRegisters(void);
#endif

#ifdef PROC_SUPPORT
static int codec_start(char *buf, char **start, off_t offset, int count,
		       int *eof, void *data);

static int codec_stop(char *buf, char **start, off_t offset, int count,
		      int *eof, void *data);

static void tsc2101_start(void);
#endif

//For Headset/Handset detection
static void omap_tsc2101_detect_tasklet(unsigned long);
DECLARE_TASKLET(omap_tsc2101_detect, omap_tsc2101_detect_tasklet, 0);
char audio_target=1;//target=0 -Headset; target=1 -Handset
static inline void omap_tsc2101_change_audioroute(char audio_target); //target=0 -Headset; target=1 -Handset

/******************** DATA STRUCTURES USING FUNCTION POINTERS **************************/

/* File Op structure for mixer */
static struct file_operations omap_mixer_fops = {
	.open           = mixer_open,
	.release        = mixer_release,
	.ioctl          = mixer_ioctl,
	.owner          = THIS_MODULE
};

/* To store characteristic info regarding the codec for the audio driver */
static audio_state_t tsc2101_state = {
	.output_stream  = &output_stream,
	.input_stream   = &input_stream,
/*	.need_tx_for_rx = 1, //Once the Full Duplex works */
	.need_tx_for_rx = 0,
	.hw_init        = omap_tsc2101_initialize,
	.hw_shutdown    = omap_tsc2101_shutdown,
	.client_ioctl   = omap_tsc2101_ioctl,
	.hw_probe       = omap_tsc2101_probe,
	.hw_remove      = omap_tsc2101_remove,
	.hw_suspend     = omap_tsc2101_suspend,
	.hw_resume      = omap_tsc2101_resume,

	.sem            = __MUTEX_INITIALIZER(tsc2101_state.sem),
};

/* This will be defined in the Audio.h */
static struct file_operations *omap_audio_fops;

/***************************** MODULES SPECIFIC FUNCTIONs *******************************/

/*********************************************************************************
 *
 * Simplified write for tsc Audio
 *
 *********************************************************************************/
static __inline__ void audio_tsc2101_write(u8 address, u16 data)
{
	omap_tsc2101_write(PAGE2_AUDIO_CODEC_REGISTERS, address, data);
}

/*********************************************************************************
 *
 * Simplified read for tsc  Audio
 *
 *********************************************************************************/
static __inline__ u16 audio_tsc2101_read(u8 address)
{
	return (omap_tsc2101_read(PAGE2_AUDIO_CODEC_REGISTERS, address));
}

/*********************************************************************************
 *
 * tsc2101_update()
 * Volume Adj etc
 *
 ********************************************************************************/
static int tsc2101_update(int flag, int val)
{
	u16 volume;
	u16 data;

	FN_IN;
	switch (flag) {
	case SET_VOLUME:
		if (val < 0 || val > 100) {
			printk(KERN_ERR "Trying a bad volume value(%d)!\n",
			       val);
			return -EPERM;
		}
		/* Convert 0 -> 100 volume to 0x7F(min) -> y(max) volume range */
		volume =
		    ((val * OUTPUT_VOLUME_RANGE) / 100) + OUTPUT_VOLUME_MAX;
		/* invert the value for getting the proper range 0 min and 100 max */
		volume = OUTPUT_VOLUME_MIN - volume;
		data = audio_tsc2101_read(TSC2101_DAC_GAIN_CTRL);
		data &=
		    ~(DGC_DALVL(OUTPUT_VOLUME_MIN) |
		      DGC_DARVL(OUTPUT_VOLUME_MIN));
		data |= DGC_DALVL(volume) | DGC_DARVL(volume);
		audio_tsc2101_write(TSC2101_DAC_GAIN_CTRL, data);
		data = audio_tsc2101_read(TSC2101_DAC_GAIN_CTRL);
		printk("DAC Gain value %04x\r\n", data);

		break;

	case SET_LINE:
		/* Not implemented yet */
		break;

	case SET_ADPGA:
		/* Michael implemented*/
		//val:0~127
		//Headset
		if( audio_target == 0)
		{
			audio_tsc2101_write(TSC2101_HEADSET_GAIN_CTRL, HGC_ADPGA_HED(val));
			printk("Michael:Headset SET_ADPGA=%d\n", val);
		}
		else
		{
			audio_tsc2101_write(TSC2101_HANDSET_GAIN_CTRL, HNGC_ADPGA_HND(val));
			printk("Michael:Handset SET_ADPGA=%d\n", val);
		}
	}
	FN_OUT(0);
	return 0;
}

/*********************************************************************************
 *
 * mixer_open()
 *
 ********************************************************************************/
static int mixer_open(struct inode *inode, struct file *file)
{
	/* Any mixer specific initialization */

	/* Initalize the tsc2101 */
	omap_tsc2101_enable();

	return 0;
}

/*********************************************************************************
 *
 * mixer_release()
 *
 ********************************************************************************/
static int mixer_release(struct inode *inode, struct file *file)
{
	/* Any mixer specific Un-initialization */
	omap_tsc2101_disable();

	return 0;
}

/*********************************************************************************
 *
 * mixer_ioctl()
 *
 ********************************************************************************/
static int
mixer_ioctl(struct inode *inode, struct file *file, uint cmd, ulong arg)
{
	int val;
	int gain;
	int ret = 0;
	int nr = _IOC_NR(cmd);

	/*
	 * We only accept mixer (type 'M') ioctls.
	 */
	FN_IN;
	printk("Call mixer_ioctl %04x\r\n", cmd);
	if (_IOC_TYPE(cmd) != 'M')
		return -EINVAL;

	DPRINTK(" 0x%08x\n", cmd);

	if (cmd == SOUND_MIXER_INFO) {
		struct mixer_info mi;

		strncpy(mi.id, "TSC2101", sizeof(mi.id));
		strncpy(mi.name, "TI TSC2101", sizeof(mi.name));
		mi.modify_counter = tsc2101_local.mod_cnt;
		FN_OUT(1);
		return copy_to_user((void *)arg, &mi, sizeof(mi));
	}

	if (_IOC_DIR(cmd) & _IOC_WRITE) {
		ret = get_user(val, (int *)arg);
		if (ret)
			goto out;

		/* Ignore separate left/right channel for now,
		 * even the codec does support it.
		 */
		gain = val & 255;

		switch (nr) {
		case SOUND_MIXER_VOLUME:
			tsc2101_local.volume = val;
			tsc2101_local.mod_cnt++;
			ret = tsc2101_update(SET_VOLUME, gain);
			break;

		case SOUND_MIXER_LINE:
			tsc2101_local.line = val;
			tsc2101_local.mod_cnt++;
			ret = tsc2101_update(SET_LINE, gain);
			break;

		case SOUND_MIXER_MIC:
			tsc2101_local.mic = val;
			tsc2101_local.mod_cnt++;
			ret = tsc2101_update(SET_LINE, gain);
			break;

		case SOUND_MIXER_RECSRC:
			break;

		default:
			ret = -EINVAL;
		}
	}

	if (ret == 0 && _IOC_DIR(cmd) & _IOC_READ) {
		ret = 0;

		switch (nr) {
		case SOUND_MIXER_VOLUME:
			val = tsc2101_local.volume;
			break;
		case SOUND_MIXER_LINE:
			val = tsc2101_local.line;
			break;
		case SOUND_MIXER_MIC:
			val = tsc2101_local.mic;
			break;
		case SOUND_MIXER_RECSRC:
			val = REC_MASK;
			break;
		case SOUND_MIXER_RECMASK:
			val = REC_MASK;
			break;
		case SOUND_MIXER_DEVMASK:
			val = DEV_MASK;
			break;
		case SOUND_MIXER_CAPS:
			val = 0;
			break;
		case SOUND_MIXER_STEREODEVS:
			val = 0;
			break;
		default:
			val = 0;
			ret = -EINVAL;
			break;
		}

		if (ret == 0)
			ret = put_user(val, (int *)arg);
	}
      out:
	FN_OUT(0);
	return ret;

}

/*********************************************************************************
 *
 * omap_set_samplerate()
 *
 ********************************************************************************/
int omap_set_samplerate(long sample_rate)
{
	u8 count = 0;
	u16 data = 0;
	/* wait for any frame to complete */
	udelay(125);

	/* Search for the right sample rate */
	while ((reg_info[count].sample_rate != sample_rate) &&
	       (count < NUMBER_SAMPLE_RATES_SUPPORTED)) {
		count++;
	}
	if (count == NUMBER_SAMPLE_RATES_SUPPORTED) {
		printk(KERN_ERR "Invalid Sample Rate %d requested\n",
		       (int)sample_rate);
		return -EPERM;
	}

	/* Set AC1 */
	data = audio_tsc2101_read(TSC2101_AUDIO_CTRL_1);
	/*Clear prev settings */
	data &= ~(AC1_DACFS(0x07) | AC1_ADCFS(0x07));
	data |=
	    AC1_DACFS(reg_info[count].divisor) | AC1_ADCFS(reg_info[count].
							   divisor);
	audio_tsc2101_write(TSC2101_AUDIO_CTRL_1, data);

printk(KERN_ALERT  "Michael: omap_set_samplerate: TSC2101_AUDIO_CTRL_1 val= 0x%x\n", audio_tsc2101_read(TSC2101_AUDIO_CTRL_1));

	/* Set the AC3 */
	data = audio_tsc2101_read(TSC2101_AUDIO_CTRL_3);
	/*Clear prev settings */
	data &= ~(AC3_REFFS | AC3_SLVMS);
	data |= (reg_info[count].fs_44kHz) ? AC3_REFFS : 0;
#ifdef TSC_MASTER
	data |= AC3_SLVMS;
#endif				/* #ifdef TSC_MASTER */
	audio_tsc2101_write(TSC2101_AUDIO_CTRL_3, data);

	/* program the PLLs */
	if (reg_info[count].fs_44kHz) {
		/* 44.1 khz - 12 MHz Mclk */
		audio_tsc2101_write(TSC2101_PLL_PROG_1, PLL1_PLLSEL | PLL1_PVAL(1) | PLL1_I_VAL(7));	/* PVAL 1; I_VAL 7 */
		audio_tsc2101_write(TSC2101_PLL_PROG_2, PLL2_D_VAL(0x1490));	/* D_VAL 5264 */
	} else {
		/* 48 khz - 12 Mhz Mclk */
		audio_tsc2101_write(TSC2101_PLL_PROG_1, PLL1_PLLSEL | PLL1_PVAL(1) | PLL1_I_VAL(8));	/* PVAL 1; I_VAL 8 */
		audio_tsc2101_write(TSC2101_PLL_PROG_2, PLL2_D_VAL(0x780));	/* D_VAL 1920 */
	}

	audio_samplerate = sample_rate;

	/* Call Mcbsp Sample rate function */
	omap_mcbsp_set_rate(AUDIO_MCBSP, sample_rate, DEFAULT_BITPERSAMPLE);

	return 0;
}

/*********************************************************************************
 *
 * omap_tsc2101_initialize() [hw_init() ]
 *
 ********************************************************************************/
static void omap_tsc2101_initialize(void *dummy)
{
	DPRINTK("omap_tsc2101_initialize entry\n");

	//printk(KERN_ALERT "Michael:omap_tsc2101_initialize entry\n");

	/* initialize with default sample rate */
	audio_samplerate = AUDIO_RATE_DEFAULT;

	omap_mcbsp_begin(AUDIO_MCBSP);
	omap_tsc2101_enable();
	tsc2101_configure();

#ifdef TEST_KEYCLICK
	tsc2101_testkeyclick();
#endif

#ifdef TONE_GEN
	toneGen();
#endif

	DPRINTK("omap_tsc2101_initialize exit\n");
}

/*********************************************************************************
 *
 * omap_tsc2101_shutdown() [hw_shutdown() ]
 *
 ********************************************************************************/
static void omap_tsc2101_shutdown(void *dummy)
{
	/*
	   Turn off codec after it is done.
	   Can't do it immediately, since it may still have
	   buffered data.

	   Wait 20ms (arbitrary value) and then turn it off.
	 */
	//printk(KERN_ALERT  "Michael: omap_tsc2101_shutdown\n");
	FN_IN;
	set_current_state(TASK_INTERRUPTIBLE);
	schedule_timeout(2);

	omap_mcbsp_end(AUDIO_MCBSP);

	audio_tsc2101_write(TSC2101_CODEC_POWER_CTRL,
			    ~(CPC_SP1PWDN | CPC_SP2PWDN | CPC_BASSBC));

	omap_tsc2101_disable();

	FN_OUT(0);
}



/*********************************************************************************
 *
 * omap_tsc2101_change_audioroute():change audio route path
 *
 ********************************************************************************/


static inline void omap_tsc2101_change_audioroute(char audio_target)
{
	
	//Headset
	if( audio_target == 0)
	{
		printk(KERN_ALERT  "Michael: Plugged, Voice to Headset\n");

		audio_tsc2101_write(TSC2101_CODEC_POWER_CTRL, 0x0000);

		//Michael:Handset Mute
		audio_tsc2101_write(TSC2101_HANDSET_GAIN_CTRL, HNGC_ADMUT_HND);

		//Michael:Headset
		audio_tsc2101_write(TSC2101_HEADSET_GAIN_CTRL, HGC_ADPGA_HED(0x28));

	
		/*Mute Analog Sidetone */
		/*Select MIC_INHED input for headset */
		/*Cell Phone In not connected */
		//Michael:MICIN_HED
		audio_tsc2101_write(TSC2101_MIXER_PGA_CTRL, MPC_ASTMU | MPC_ASTG(0x40) |MPC_MICSEL(0)| MPC_MICADC);



		/* ADC, DAC, Analog Sidetone, cellphone, buzzer softstepping enabled */
		/* 1dB AGC hysteresis */
		/* MICes bias 2V */
		//Michael:Headset
		//audio_tsc2101_write(TSC2101_AUDIO_CTRL_4, AC4_MB_HED(3));
		audio_tsc2101_write(TSC2101_AUDIO_CTRL_4, AC4_MB_HED(0));


	audio_tsc2101_write(TSC2101_AUDIO_CTRL_6,
			    AC6_MUTLSPK | AC6_MUTSPK2 | AC6_LDSCPTC |
			    AC6_VGNDSCPTC);



	}
	//Handset
	else
	{
		printk(KERN_ALERT  "Michael: Unplugged, Voice to Handset\n");

		//Michael:MICBIAS_HND->UP;MICBIAS_HED->DOWN
		audio_tsc2101_write(TSC2101_CODEC_POWER_CTRL, 0x5880);

		//Michael:Headset Mute
		audio_tsc2101_write(TSC2101_HEADSET_GAIN_CTRL, HGC_ADMUT_HED);

		//Michael:Handset Gain 70
		audio_tsc2101_write(TSC2101_HANDSET_GAIN_CTRL, HNGC_ADPGA_HND(0x46));

		//Michael:MICIN_HND, Mute Analog Sidetone, Select MICIN_HND input for headset (MICSEL: 0x01 << 5), Cell Phone In not connected
		audio_tsc2101_write(TSC2101_MIXER_PGA_CTRL, MPC_ASTMU | MPC_ASTG(0x40) |((0x01)<<5)| MPC_MICADC);

		//Michael:Hanset, MICes bias 2V
		audio_tsc2101_write(TSC2101_AUDIO_CTRL_4, AC4_MB_HND);


	//For loudSpeaker
	audio_tsc2101_write(TSC2101_AUDIO_CTRL_6, 0x9070);






	}

	return;
}



/*********************************************************************************
*
*     omap_tsc2101_ts_tasklet(unsigned long nothing) : Tasklet Function 
*
**********************************************************************************/
static void omap_tsc2101_detect_tasklet(unsigned long nothing)
{
	
	u16 data = audio_tsc2101_read(TSC2101_AUDIO_CTRL_7);
	//u16 data2 = audio_tsc2101_read(TSC2101_AUDIO_CTRL_7);
	//printk(KERN_ALERT "Michael: Before Head/Hand interrupt detect! 0x%x\n",audio_tsc2101_read(TSC2101_AUDIO_CTRL_7));

	//Headset
	if( data &= 0x1000)
	{
		audio_target=0;
		apm_queue_event(WIFI_EVENT_HEADSET);
	}
	//Handset
	else
	{	
		audio_target=1;
		apm_queue_event(WIFI_EVENT_HANDSET);
	}

	omap_tsc2101_change_audioroute(audio_target);
	//data2 &= 0x7FFF; //clear AC7_DETECT
	//audio_tsc2101_write(TSC2101_AUDIO_CTRL_7,  data2);
	//printk(KERN_ALERT "Michael: ERASE Headset interrupt detect!0x%x\n", audio_tsc2101_read(TSC2101_AUDIO_CTRL_7));
	//audio_tsc2101_write(TSC2101_AUDIO_CTRL_7,  AC7_DETECT | AC7_HDDEBNPG(3) | AC7_DGPIO2);
	//printk(KERN_ALERT "Michael: After Head/Hand set interrupt detect!0x%x\n\n",audio_tsc2101_read(TSC2101_AUDIO_CTRL_7));

	return;
	
}

/*********************************************************************************
*
*    omap_tsc2101_ts_handler(int irq, void *dev_id,
*                                     struct pt_regs *regs)  : Interrupt Handler
*
***********************************************************************************/
static irqreturn_t omap_tsc2101_detect_handler(int irq, void *dev_id, struct pt_regs *regs)
{

	/* Schedule Takslet */
	tasklet_schedule(&omap_tsc2101_detect);


	return IRQ_HANDLED;
}
/*********************************************************************************
 *
 * tsc2101_configure
 *
 ********************************************************************************/
 
static inline void tsc2101_configure()
{
	FN_IN;

//printk(KERN_ALERT "Michael:tsc2101_configure\n");
//printk(KERN_ALERT  "Michael: tsc2101_configure ori reg val= 0x%x\n", audio_tsc2101_read(TSC2101_AUDIO_CTRL_7));




	omap_tsc2101_change_audioroute(audio_target);

	/* Set codec output volume */
	audio_tsc2101_write(TSC2101_DAC_GAIN_CTRL, 0x0000);

	/* DAC left and right routed to SPK2 */
	/* SPK1/2 unmuted */
	audio_tsc2101_write(TSC2101_AUDIO_CTRL_5,
			    AC5_DAC2SPK1(3) | AC5_AST2SPK1 | AC5_KCL2SPK1 |
			    AC5_DAC2SPK2(3) | AC5_AST2SPK2 | AC5_KCL2SPK2 |
			    AC5_HDSCPTC);

	/* Headset detection control */
	//Do in init
	//audio_tsc2101_write(TSC2101_AUDIO_CTRL_7, AC7_DETECT | AC7_HDDEBNPG(3) |AC7_DGPIO2);

//printk(KERN_ALERT  "Michael: tsc2101_configure new reg val= 0x%x\n", audio_tsc2101_read(TSC2101_AUDIO_CTRL_7));



	/* Headset/Hook switch detect disabled */
//	audio_tsc2101_write(TSC2101_AUDIO_CTRL_7, 0x0000);

	/* Initialize the AIC23 internal state */
	/*
	   The AIC23 uses 9 bits for register control.  The
	   extra bit gets placed in the LSB of the subregister
	   address, and the address is shifted by one.
	   the volume control is only for earphone not line out
	   line input volume can be controlled but not in following code
	   which pick the default value 0dB
	 */

	/* Left line input volume control */
	tsc2101_local.line = DEFAULT_INPUT_VOLUME;
	tsc2101_local.mic = DEFAULT_INPUT_VOLUME;
	tsc2101_update(SET_LINE, DEFAULT_INPUT_VOLUME);

	/* Left/Right headphone channel volume control */
	/* Zero-cross detect on */
	tsc2101_local.volume = DEFAULT_VOLUME;
	tsc2101_update(SET_VOLUME, DEFAULT_VOLUME);

	/* clock configuration */
	omap_set_samplerate(audio_samplerate);

#ifdef TSC_DUMP_REGISTERS
	tsc2101_dumpRegisters();
#endif

	FN_OUT(0);
}


#ifdef PROC_SUPPORT
static void tsc2101_start()
{
	FN_IN;

#ifdef CONFIG_RECEIVER
	audio_tsc2101_write(TSC2101_CODEC_POWER_CTRL, 0x0000);
#else
	//For loudSpeaker
	audio_tsc2101_write(TSC2101_CODEC_POWER_CTRL, 0x9880);
#endif

	/*Headset Input not muted */
	/*AGC for Headset In off */
	audio_tsc2101_write(TSC2101_HEADSET_GAIN_CTRL, HGC_ADPGA_HED(0x7D));

	/*Mute Analog Sidetone */
	/*Select MIC_INHED input for headset */
	/*Cell Phone In not connected */
	audio_tsc2101_write(TSC2101_MIXER_PGA_CTRL,
			    MPC_ASTMU | MPC_ASTG(0x40) | MPC_MICADC);

	/* ADC, DAC, Analog Sidetone, cellphone, buzzer softstepping enabled */
	/* 1dB AGC hysteresis */
	/* MICes bias 2V */
	audio_tsc2101_write(TSC2101_AUDIO_CTRL_4, AC4_MB_HED(0));

	/* Set codec output volume */
	audio_tsc2101_write(TSC2101_DAC_GAIN_CTRL, 0x0000);

	/* DAC left and right routed to SPK2 */
	/* SPK1/2 unmuted */
	audio_tsc2101_write(TSC2101_AUDIO_CTRL_5,
			    AC5_DAC2SPK1(3) | AC5_AST2SPK1 | AC5_KCL2SPK1 |
			    AC5_DAC2SPK2(3) | AC5_AST2SPK2 | AC5_KCL2SPK2 |
			    AC5_HDSCPTC);



#ifdef CONFIG_RECEIVER
	/* OUT8P/N muted, CPOUT muted */

	audio_tsc2101_write(TSC2101_AUDIO_CTRL_6,
			    AC6_MUTLSPK | AC6_MUTSPK2 | AC6_LDSCPTC |
			    AC6_VGNDSCPTC);
#else
	//For loudSpeaker
	audio_tsc2101_write(TSC2101_AUDIO_CTRL_6, 0x9070);
#endif

	/* Headset/Hook switch detect disabled */
	//audio_tsc2101_write(TSC2101_AUDIO_CTRL_7, 0x0000);

	/* Initialize the AIC23 internal state */
	/*
	   The AIC23 uses 9 bits for register control.  The
	   extra bit gets placed in the LSB of the subregister
	   address, and the address is shifted by one.
	   the volume control is only for earphone not line out
	   line input volume can be controlled but not in following code
	   which pick the default value 0dB
	 */

	/* Left line input volume control */
	tsc2101_local.line = DEFAULT_INPUT_VOLUME;
	tsc2101_local.mic = DEFAULT_INPUT_VOLUME;
	tsc2101_update(SET_LINE, DEFAULT_INPUT_VOLUME);

	/* Left/Right headphone channel volume control */
	/* Zero-cross detect on */
	tsc2101_local.volume = DEFAULT_VOLUME;
	tsc2101_update(SET_VOLUME, DEFAULT_VOLUME);

	FN_OUT(0);

}
#endif

/******************************************************************************************
 *
 * All generic ioctl's are handled by audio_ioctl() [File: omap-audio.c]. This
 * routine handles some platform specific ioctl's
 *
 ******************************************************************************************/
static int
omap_tsc2101_ioctl(struct inode *inode, struct file *file, uint cmd, ulong arg)
{
	long val;
	int ret = 0;

	DPRINTK(" 0x%08x\n", cmd);

	/*
	 * These are platform dependent ioctls which are not handled by the
	 * generic omap-audio module.
	 */
	switch (cmd) {
	case SNDCTL_DSP_STEREO:
		ret = get_user(val, (int *)arg);
		if (ret)
			return ret;
		/* the AIC23 is stereo only */
		ret = (val == 0) ? -EINVAL : 1;
		FN_OUT(1);
		return put_user(ret, (int *)arg);

	case SNDCTL_DSP_CHANNELS:
	case SOUND_PCM_READ_CHANNELS:
		/* the AIC23 is stereo only */
		FN_OUT(2);
		return put_user(2, (long *)arg);

	case SNDCTL_DSP_SPEED:
		ret = get_user(val, (long *)arg);
		if (ret)
			break;
		ret = omap_set_samplerate(val);
		if (ret)
			break;
		/* fall through */

	case SOUND_PCM_READ_RATE:
		FN_OUT(3);
		return put_user(audio_samplerate, (long *)arg);

	case SOUND_PCM_READ_BITS:
	case SNDCTL_DSP_SETFMT:

		ret = get_user(val, (long *)arg);
		ret = tsc2101_update(SET_ADPGA, val);
		
	case SNDCTL_DSP_GETFMTS:
		/* we can do 16-bit only */
		FN_OUT(4);
		return put_user(AFMT_S16_LE, (long *)arg);

	//Add by Alan
	case SOUND_MIXER_MUTE:
	{
		int gain;

		ret = get_user(val, (int *)arg);
		gain = audio_tsc2101_read(TSC2101_DAC_GAIN_CTRL);
		if ( val ) {
			gain |= 0x00008080;
		}
		else {
			gain &= ~(0x00008080);
		}
		audio_tsc2101_write(TSC2101_DAC_GAIN_CTRL, gain);
		val = audio_tsc2101_read(TSC2101_DAC_GAIN_CTRL);
		printk("DAC Gain value %04x\r\n", val);
	}
		break;
	default:
		/* Maybe this is meant for the mixer (As per OSS Docs) */
		FN_OUT(5);
		return mixer_ioctl(inode, file, cmd, arg);
	}

	FN_OUT(0);
	return ret;
}

/*********************************************************************************
 *
 * module_probe for TSC2101
 *
 ********************************************************************************/
static int omap_tsc2101_probe(void)
{
	FN_IN;

	/* Get the fops from audio oss driver */
	if (!(omap_audio_fops = audio_get_fops())) {
		printk(KERN_ERR "Unable to Get the FOPs of Audio OSS driver\n");
		audio_unregister_codec(&tsc2101_state);
		return -EPERM;
	}

	tsc2101_local.volume = DEFAULT_VOLUME;

	/* register devices */
	audio_dev_id = register_sound_dsp(omap_audio_fops, -1);
	mixer_dev_id = register_sound_mixer(&omap_mixer_fops, -1);

#ifdef PROC_SUPPORT
	create_proc_read_entry(PROC_START_FILE, 0 /* default mode */ ,
			       NULL /* parent dir */ ,
			       codec_start, NULL /* client data */ );

	create_proc_read_entry(PROC_STOP_FILE, 0 /* default mode */ ,
			       NULL /* parent dir */ ,
			       codec_stop, NULL /* client data */ );
#endif

	/* Announcement Time */
	printk(KERN_INFO PLATFORM_NAME " " CODEC_NAME
	       " Audio support initialized\n");

	FN_OUT(0);
	return 0;
}

/*********************************************************************************
 *
 * Module Remove for TSC2101
 *
 ********************************************************************************/
static void omap_tsc2101_remove(void)
{
	FN_IN;
	/* Un-Register the codec with the audio driver */
	unregister_sound_dsp(audio_dev_id);
	unregister_sound_mixer(mixer_dev_id);

#ifdef PROC_SUPPORT
	remove_proc_entry(PROC_START_FILE, NULL);
	remove_proc_entry(PROC_STOP_FILE, NULL);
#endif
	FN_OUT(0);

}

#ifdef WIFI_PHONE_PM
static int tsc2101_pm_callback(struct pm_dev *pmdev, pm_request_t pmreq,  void *pmdata)
{
	switch(pmreq) {
		case PM_SUSPEND:
			omap_tsc2101_suspend();
			break;
		case PM_RESUME:
			omap_tsc2101_resume();
			break;
	}


	return 0;
}
#endif
/*********************************************************************************
 *
 * Module Suspend for TSC2101
 *
 ********************************************************************************/
static int omap_tsc2101_suspend(void)
{
#ifdef WIFI_PHONE_PM		//ron_lee 2005-11-18
	info("tsc2101 suspend");
	audio_tsc2101_write(TSC2101_CODEC_POWER_CTRL, 0xfffd);
#endif	//WIFI_PHONE_PM
	FN_OUT(0);
	return 0;
}

/*********************************************************************************
 *
 * Module Resume for TSC2101
 *
 ********************************************************************************/
static int omap_tsc2101_resume(void)
{
#ifdef WIFI_PHONE_PM
	info("tsc2101 resume");
	audio_tsc2101_write(TSC2101_CODEC_POWER_CTRL, 0x0000);
#endif
	FN_OUT(0);
	return 0;
}

/*********************************************************************************
 *
 * module_init for TSC2101
 *
 ********************************************************************************/
static int __init audio_tsc2101_init(void)
{

	powerOntsc2101();
	int err = 0;
	FN_IN;
	/* register the codec with the audio driver */
	if ((err = audio_register_codec(&tsc2101_state))) {
		printk(KERN_ERR
		       "Failed to register TSC driver with Audio OSS Driver\n");
	}
	else
	{

		//printk(KERN_ALERT  "Michael: ori reg val= 0x%x\n", audio_tsc2101_read(TSC2101_AUDIO_CTRL_7));
		audio_tsc2101_write(TSC2101_AUDIO_CTRL_7,  AC7_DETECT | AC7_HDDEBNPG(0) | AC7_DGPIO2);
		//audio_tsc2101_write(TSC2101_GPIO_CTRL,  GC_GPO2EN);

		int detect_irq = OMAP_GPIO_IRQ(OMAP_1710_GPIO_NUM);
		/* Configure GPIO for Detect IRQ */
		omap_cfg_reg(V9_1710_GPIO7);
		if (omap_request_gpio(OMAP_1710_GPIO_NUM) != 0) {
			printk(KERN_ALERT "omap-audio-tsc2101.c: Could not reserve GPIO7!\n");
			return -EINVAL;
		}
		else
		{	
			//printk(KERN_ALERT "Michael:Reserve GPIO7 O.K.!\n");
		}

		omap_set_gpio_direction(OMAP_1710_GPIO_NUM, 1);
		//omap_set_gpio_edge_ctrl(OMAP_1710_GPIO_NUM, OMAP_GPIO_RISING_EDGE);

	/* Request irq */
	if (request_irq(detect_irq, omap_tsc2101_detect_handler,
			SA_INTERRUPT,
			"omap_tsc2101_detect", NULL)) {
		printk(KERN_ALERT  "omap-audio-tsc2101.c: Could not allocate detect IRQ!\n");
	}
	else
		{	
			//printk(KERN_ALERT "Michael:Allocate detect IRQ GPIO7 O.K.!\n");
			//printk(KERN_ALERT  "Michael: new reg val= 0x%x\n", audio_tsc2101_read(TSC2101_AUDIO_CTRL_7));
		}
	}


	//Make MIC detection work during init.
	u16 data = audio_tsc2101_read(TSC2101_AUDIO_CTRL_7);
	//Headset
	if( data &= 0x1000)
	{
		audio_target=0;
	}
	//Handset
	else
	{	
		audio_target=1;
	}
#ifdef WIFI_PHONE_PM
	if( !(pm_dev_tsc2101 = pm_register(PM_SYS_UNKNOWN, (unsigned long) 0x41/*A*/, tsc2101_pm_callback)))
		info("Fail to register pm callback funtion");
#endif
	
	FN_OUT(err);
	return err;
}

/*********************************************************************************
 *
 * module_exit for TSC2101
 *
 ********************************************************************************/
static void __exit audio_tsc2101_exit(void)
{

	FN_IN;
	(void)audio_unregister_codec(&tsc2101_state);
	FN_OUT(0);
	return;
}

/**************************** DEBUG FUNCTIONS ***********************************/

/*********************************************************************************
 * TEST_KEYCLICK:
 * This is a test to generate various keyclick sound on tsc.
 * verifies if the tsc and the spi interfaces are operational.
 *
 ********************************************************************************/
#ifdef TEST_KEYCLICK
void tsc2101_testkeyclick(void)
{
	u8 freq = 0;
	u16 old_reg_val, reg_val;
	u32 uDummyVal = 0;
	u32 uTryVal = 0;

	old_reg_val = audio_tsc2101_read(TSC2101_AUDIO_CTRL_2);

	/* Keyclick active, max amplitude and longest key click len(32 period) */
	printk(KERN_INFO " TESTING KEYCLICK\n Listen carefully NOW....\n");
	printk(KERN_INFO " OLD REG VAL=0x%x\n", old_reg_val);
	/* try all frequencies */
	for (; freq < 8; freq++) {
		/* Keyclick active, max amplitude and longest key click len(32 period) */
		reg_val = old_reg_val | AC2_KCLAC(0x7) | AC2_KCLLN(0xF);
		uDummyVal = 0;
		uTryVal = 0;
		printk(KERN_INFO "\n\nTrying frequency %d reg val= 0x%x\n",
		       freq, reg_val | AC2_KCLFRQ(freq) | AC2_KCLEN);
		audio_tsc2101_write(TSC2101_AUDIO_CTRL_2,
				    reg_val | AC2_KCLFRQ(freq) | AC2_KCLEN);
		printk("DONE. Wait 10 ms ...\n");
		/* wait till the kclk bit is auto cleared! time out also to be considered. */
		while (audio_tsc2101_read(TSC2101_AUDIO_CTRL_2) & AC2_KCLEN) {
			udelay(3);
			uTryVal++;
			if (uTryVal > 2000) {
				printk(KERN_ERR
				       "KEYCLICK TIMED OUT! freq val=%d, POSSIBLE ERROR!\n",
				       freq);
				printk(KERN_INFO
				       "uTryVal == %d: Read back new reg val= 0x%x\n",
				       uTryVal,
				       audio_tsc2101_read
				       (TSC2101_AUDIO_CTRL_2));
				/* clear */
				audio_tsc2101_write(TSC2101_AUDIO_CTRL_2, 0x00);
				break;
			}
		}
	}
	/* put the old value back */
	audio_tsc2101_write(TSC2101_AUDIO_CTRL_2, old_reg_val);
	printk(KERN_INFO " KEYCLICK TEST COMPLETE\n");

}				/* End of tsc2101_testkeyclick */

#endif				/* TEST_KEYCLICK */

/*********************************************************************************
 * TONEGEN:
 * This is a test to generate a rather unpleasant sound..
 * verifies if the mcbsp is active (requires MCBSP_DIRECT_RW to be active on McBSP)
 *
 ********************************************************************************/
#ifdef TONE_GEN
/* Generates a shrill tone */
u16 tone[] = {
	0x0ce4, 0x0ce4, 0x1985, 0x1985, 0x25A1, 0x25A1, 0x30FD, 0x30FE,
	0x3B56, 0x3B55, 0x447A, 0x447A, 0x4C3B, 0x4C3C, 0x526D, 0x526C,
	0x56F1, 0x56F1, 0x59B1, 0x59B1, 0x5A9E, 0x5A9D, 0x59B1, 0x59B2,
	0x56F3, 0x56F2, 0x526D, 0x526D, 0x4C3B, 0x4C3B, 0x447C, 0x447C,
	0x3B5A, 0x3B59, 0x30FE, 0x30FE, 0x25A5, 0x25A6, 0x1989, 0x198A,
	0x0CE5, 0x0CE3, 0x0000, 0x0000, 0xF31C, 0xF31C, 0xE677, 0xE676,
	0xDA5B, 0xDA5B, 0xCF03, 0xCF03, 0xC4AA, 0xC4AA, 0xBB83, 0xBB83,
	0xB3C5, 0xB3C5, 0xAD94, 0xAD94, 0xA90D, 0xA90E, 0xA64F, 0xA64E,
	0xA562, 0xA563, 0xA64F, 0xA64F, 0xA910, 0xA90F, 0xAD93, 0xAD94,
	0xB3C4, 0xB3C4, 0xBB87, 0xBB86, 0xC4AB, 0xC4AB, 0xCF03, 0xCF03,
	0xDA5B, 0xDA5A, 0xE67B, 0xE67B, 0xF31B, 0xF3AC, 0x0000, 0x0000,
	0x0CE4, 0x0CE4, 0x1985, 0x1985, 0x25A1, 0x25A1, 0x30FD, 0x30FE,
	0x3B56, 0x3B55, 0x447A, 0x447A, 0x4C3B, 0x4C3C, 0x526D, 0x526C,
	0x56F1, 0x56F1, 0x59B1, 0x59B1, 0x5A9E, 0x5A9D, 0x59B1, 0x59B2,
	0x56F3, 0x56F2, 0x526D, 0x526D, 0x4C3B, 0x4C3B, 0x447C, 0x447C,
	0x3B5A, 0x3B59, 0x30FE, 0x30FE, 0x25A5, 0x25A6, 0x1989, 0x198A,
	0x0CE5, 0x0CE3, 0x0000, 0x0000, 0xF31C, 0xF31C, 0xE677, 0xE676,
	0xDA5B, 0xDA5B, 0xCF03, 0xCF03, 0xC4AA, 0xC4AA, 0xBB83, 0xBB83,
	0xB3C5, 0xB3C5, 0xAD94, 0xAD94, 0xA90D, 0xA90E, 0xA64F, 0xA64E,
	0xA562, 0xA563, 0xA64F, 0xA64F, 0xA910, 0xA90F, 0xAD93, 0xAD94,
	0xB3C4, 0xB3C4, 0xBB87, 0xBB86, 0xC4AB, 0xC4AB, 0xCF03, 0xCF03,
	0xDA5B, 0xDA5A, 0xE67B, 0xE67B, 0xF31B, 0xF3AC, 0x0000, 0x0000,
	0x0CE4, 0x0CE4, 0x1985, 0x1985, 0x25A1, 0x25A1, 0x30FD, 0x30FE,
	0x3B56, 0x3B55, 0x447A, 0x447A, 0x4C3B, 0x4C3C, 0x526D, 0x526C,
	0x56F1, 0x56F1, 0x59B1, 0x59B1, 0x5A9E, 0x5A9D, 0x59B1, 0x59B2,
	0x56F3, 0x56F2, 0x526D, 0x526D, 0x4C3B, 0x4C3B, 0x447C, 0x447C,
	0x3B5A, 0x3B59, 0x30FE, 0x30FE, 0x25A5, 0x25A6, 0x1989, 0x198A,
	0x0CE5, 0x0CE3, 0x0000, 0x0000, 0xF31C, 0xF31C, 0xE677, 0xE676,
	0xDA5B, 0xDA5B, 0xCF03, 0xCF03, 0xC4AA, 0xC4AA, 0xBB83, 0xBB83,
	0xB3C5, 0xB3C5, 0xAD94, 0xAD94, 0xA90D, 0xA90E, 0xA64F, 0xA64E,
	0xA562, 0xA563, 0xA64F, 0xA64F, 0xA910, 0xA90F, 0xAD93, 0xAD94,
	0xB3C4, 0xB3C4, 0xBB87, 0xBB86, 0xC4AB, 0xC4AB, 0xCF03, 0xCF03,
	0xDA5B, 0xDA5A, 0xE67B, 0xE67B, 0xF31B, 0xF3AC, 0x0000, 0x0000
};

void toneGen(void)
{
	int count = 0;
	int ret = 0;
	printk(KERN_INFO "TONE GEN TEST :");

	for (count = 0; count < 5000; count++) {
		int bytes;
		for (bytes = 0; bytes < sizeof(tone) / 2; bytes++) {
			ret = omap_mcbsp_write(AUDIO_MCBSP, tone[bytes]);
			if (ret == -1) {
				/* retry */
				bytes--;
			} else if (ret == -2) {
				printk(KERN_INFO "ERROR:bytes=%d\n", bytes);
				return;
			}
		}
	}
	printk(KERN_INFO "SUCCESS\n");
}

#endif				/* End of TONE_GEN */

/*********************************************************************************
 *
 * TSC_DUMP_REGISTERS:
 * This will dump the entire register set of Page 2 tsc2101. 
 * Useful for major goof ups
 *
 ********************************************************************************/
#ifdef TSC_DUMP_REGISTERS
static void tsc2101_dumpRegisters(void)
{
	int i = 0;
	u16 data = 0;
	printk("TSC 2101 Register dump for Page 2 \n");
	for (i = 0; i < 0x27; i++) {
		data = audio_tsc2101_read(i);
		printk(KERN_INFO "Register[%x]=0x%04x\n", i, data);

	}
}
#endif				/* End of #ifdef TSC_DUMP_REGISTERS */

#ifdef PROC_SUPPORT
static int codec_start(char *buf, char **start, off_t offset, int count,
		       int *eof, void *data)
{
	omap_tsc2101_enable();
	tsc2101_start();
	printk("Codec initialization done.\n");
	return 0;
}
static int codec_stop(char *buf, char **start, off_t offset, int count,
		      int *eof, void *data)
{

	omap_tsc2101_disable();
	audio_tsc2101_write(TSC2101_CODEC_POWER_CTRL,
			    ~(CPC_SP1PWDN | CPC_SP2PWDN | CPC_BASSBC));
	printk("Codec shutdown.\n");
	return 0;
}
#endif

/*********************************************************************************
 *
 * Other misc management, registration etc
 *
 ********************************************************************************/
module_init(audio_tsc2101_init);
module_exit(audio_tsc2101_exit);

MODULE_AUTHOR("Texas Instruments");
MODULE_DESCRIPTION
    ("Glue audio driver for the TI OMAP1610/OMAP1710/2420 TSC2101 codec.");
MODULE_LICENSE("GPL");
